/*
    saa7146_core.c - core-functions + i2c driver for the saa7146 by
    Philips Semiconductors.
    
    Copyright (C) 1998,1999 Michael Hunold <michael@mihu.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <linux/module.h>	/* for module-version */
#include <linux/delay.h>	/* for delay-stuff */
#include <linux/slab.h>		/* for kmalloc/kfree */
#include <linux/pci.h>		/* for pci-config-stuff, vendor ids etc. */
#include <linux/wrapper.h>	/* for mem_map_reserve */
#if LINUX_VERSION_CODE >= 0x020300
#include <linux/init.h>
#endif
#include <asm/io.h>		/* for accessing the pci-device */
#include <linux/vmalloc.h>	/* for module-version */
#include <linux/highmem.h>

#include "saa7146_defs.h"
#include "saa7146_core.h"

#ifdef MODULE
MODULE_PARM(mode,"i");
MODULE_PARM(debug,"i");
MODULE_PARM(buffers,"i");
#endif

/* insmod parameter: here you can specify the number of video-buffers
   to be allocated. for simple capturing 2 buffers (double-buffering)
   should suffice. but if you plan to do 25fps grabbing, you should
   set this to 4(=maximum), in order to be able to catch up from
   temporarily delays */
static int buffers = 2;

/* insmod parameter: some programs (e.g. vic) do not allow to
   specify the used video-mode, so you have to tell this to the
   modules by hand, 0 = PAL, 1 = NTSC  */
static int mode = 0;

/* debug levels: 0 -- no debugging outputs
		 1 -- prints out entering (and exiting if useful) of functions
		 2 -- prints out very, very detailed informations of what is going on
		 3 -- both of the above */
int debug = 0;	/* insmod parameter */

#define dprintk		if (debug&1) printk
#define hprintk		if (debug&2) printk

/* ---------------------------------------------*/
/* memory functions - taken from bttv.c		*/
/* ---------------------------------------------*/

#define pte_page_old(pte) \
((unsigned long) __va(pte_val(pte) & PAGE_MASK))

static inline unsigned long uvirt_to_kva(pgd_t *pgd, unsigned long adr)
{
        unsigned long ret = 0UL;
	pmd_t *pmd;
	pte_t *ptep, pte;
  
	if (!pgd_none(*pgd)) {
                pmd = pmd_offset(pgd, adr);
                if (!pmd_none(*pmd)) {
                        ptep = pte_offset(pmd, adr);
                        pte = *ptep;
                        if(pte_present(pte))
                                ret = (pte_page_old(pte)|(adr&(PAGE_SIZE-1)));
                }
        }
	return ret;
}

static inline unsigned long kvirt_to_pa(unsigned long adr) 
{
        unsigned long va, kva, ret;

        va = VMALLOC_VMADDR(adr);
        kva = uvirt_to_kva(pgd_offset_k(va), va);
	ret = __pa(kva);
        return ret;
}



static inline unsigned long uvirt_to_phys(unsigned long adr)
{
	pgd_t *pgd;
	pmd_t *pmd;
	pte_t *ptep, pte;
  
	pgd = pgd_offset(current->mm, adr);
	if (pgd_none(*pgd))
		return 0;
	pmd = pmd_offset(pgd, adr);
	if (pmd_none(*pmd))
		return 0;
	ptep = pte_offset(pmd, adr);
	pte = *ptep;
	if(pte_present(pte))
		return 
			virt_to_phys((void *)(pte_page_old(pte)|(adr&(PAGE_SIZE-1))));
	return 0;
}

static inline unsigned long uvirt_to_bus(unsigned long adr) 
{
	return virt_to_bus(phys_to_virt(uvirt_to_phys(adr)));
}



#define SAA7146_RESCHED \
        do { \
          if (current->need_resched) \
            schedule(); \
        } while (0)

static int saa7146_num = 0;
static struct saa7146		saa7146_adap[SAA7146_MAX_DEVICES];

static int saa7146_extension_count = 0;
static struct saa7146_extension* saa7146_ext[SAA7146_MAX_EXTENSIONS];

/* i2c-timeout-value in ms */
#define SAA7146_I2C_TIMEOUT 		100
/* how many times shall we retry an i2c-operation? */
#define SAA7146_I2C_RETRIES 		3

static u32 SAA7146_I2C_BBR = SAA7146_I2C_BUS_BIT_RATE_3200;

#define	__COMPILE_SAA7146_I2C__
#define	__COMPILE_SAA7146_DEBI__
#include "saa7146.c"
#undef	__COMPILE_SAA7146_I2C__

/* ---------------------------------------------*/
/* memory functions designed for saa7146	*/
/* ---------------------------------------------*/

/* rvmalloc allocates the memory and builds up
   the page-tables for quant-number of buffers */
static void* rvmalloc(int quant, u32* pt[])
{
	void* mem;

	unsigned long adr = 0;
	unsigned long count = 0;

	u32* ptp = 0;
	int i = 0, j = 0;

	dprintk(KERN_ERR "saa7146: rvmalloc called, quant:%d\n",quant);

	if(!quant) 
		return NULL;

	/* get grabbing memory */
	mem = vmalloc_32(quant*GRABBING_MEM_SIZE); 

	if(!mem) 
		return NULL;

	dprintk(KERN_ERR "saa7146: alloc page tables\n");

	/* alloc one page for a page-table for quant buffers */
	for(i = 0; i < quant; i++) {
		pt[i] = (u32*)kmalloc(PAGE_SIZE,GFP_KERNEL);

		/* error: memory could not be allocated */
		if(!pt[i]) {
			dprintk(KERN_ERR "saa7146: failed, free tables\n");
			for(j = (i-1); j >= 0; j--) 
				kfree(pt[j]);
			dprintk(KERN_ERR "saa7146: free buffer memory\n");
			vfree(mem);
			dprintk(KERN_ERR "saa7146: return 0 address for buffer\n");
			return NULL;
		}
		memset(pt[i], 0x00, PAGE_SIZE);
	}

	dprintk(KERN_ERR "saa7146: clear RAM\n");

	/* clear the ram out, no junk to the user
	   note: 0x7f gives a nice grey field
	   in RGB and YUV as well */
	memset(mem, 0x7f, quant*GRABBING_MEM_SIZE); 

	dprintk(KERN_ERR "saa7146: build page tables\n");
	adr = (unsigned long)mem;
	/* walk through the grabbing-memory and build up the page-tables */
	for(i = 0; i < quant; i++) {

	        for (count=0; count<GRABBING_MEM_SIZE; count+=PAGE_SIZE) 
#if LINUX_VERSION_CODE >= 0x020400
		        mem_map_reserve(virt_to_page(__va(kvirt_to_pa(adr+count))));
#else
		        mem_map_reserve(MAP_NR(phys_to_virt(kvirt_to_pa(adr+count))));
#endif
		/* separate loop for SAA MMU, PAGE_SIZE can be !=4096 */
		ptp = pt[i];
		for (count=0; count<GRABBING_MEM_SIZE; count+=4096, adr+=4096)
 	                *(ptp++) = cpu_to_le32(kvirt_to_pa(adr));
	}
	dprintk(KERN_ERR "saa7146: page tables built\n");
	return mem;
}

static void rvfree(void* mem, int quant, u32* pt[])
{
        unsigned long adr, page;
	unsigned long size = 0;

	int i = 0;

	dprintk(KERN_ERR "saa7146: rvfree called\n");

	if (!quant) 
		return;
	
	if (mem) {
		adr = (unsigned long)mem;
		size = quant * GRABBING_MEM_SIZE;
		
		while (size > 0) {
			page = kvirt_to_pa(adr);
#if LINUX_VERSION_CODE >= 0x020400
			mem_map_unreserve(virt_to_page(__va(page)));
#else

			mem_map_unreserve(MAP_NR(phys_to_virt(page)));
#endif
			adr	+= PAGE_SIZE;
			size	-= PAGE_SIZE;
		}
		
		/* release the grabbing memory */
		vfree(mem);
	}
	/* free the page tables */
	for(i = 0; i < quant; i++) {
		kfree(pt[i]);
	}
}


/* ---------------------------------------------*/
/* i2c-functions				*/
/* ---------------------------------------------*/

int master_xfer(struct i2c_adapter *adap,
	struct i2c_msg msgs[], int num)
{
	struct saa7146 *a = (struct saa7146*)(adap->data);

	int result, count;
	int retries = adap->retries;
	int i = 0;
	
	dprintk(KERN_ERR "saa7146_core.o: master_xfer called, num:%d\n",num);

	/* prepare the message(s), get number of u32s to transfer */
	count = prepare(msgs, num, a->i2c);
	if ( 0 > count ) {
		hprintk(KERN_ERR "saa7146_core.o: master_xfer: could not prepare i2c-message\n");
		return -1;
	}

	/* loop through number of retries ... */
	while ( 0 < retries ) {
	
		retries--;
	
		/* reset the i2c-device if necessary */
		result = i2c_reset( a );
		if ( 0 > result ) {
			hprintk(KERN_ERR "saa7146_core.o: master_xfer: could not reset i2c-bus\n");
			return -1;
		}	

		/* see how many u32 have to be transferred; if there is only 1,
		   we do not start the whole rps1-engine... */

		for(i = 0; i < count; i++) {
			result = i2c_write_out( a, &a->i2c[i], adap->timeout );
			if ( 0 != result) {
				/* if address-error occured, dont retry */
				if ( -2 == result ) {
					hprintk(KERN_ERR "saa7146_core.o: master_xfer: error in address phase\n");
					return -1;
				}
				hprintk(KERN_ERR "saa7146_core.o: master_xfer: error tranferring,trying again\n");
				break;
			}
		}
		
		/* see if an error occured & the last retry failed */
		if( (0 != result) && (0 == retries) ) {
			hprintk(KERN_ERR "saa7146_core.o: master_xfer: could not transfer i2c-message\n");
			return -1;
		}
		
		if( 0 == result )
			break;
	}
	
	/* if any things had to be read, get the results */
	result = clean_up(msgs, num, a->i2c);
	if ( 0 > result ) {
		hprintk(KERN_ERR "saa7146_core.o: master_xfer: could not cleanup\n");
		return -1;
	}

	/* return the number of delivered messages */
	return num;	
}

int algo_control(struct i2c_adapter *adapter, 
	unsigned int cmd, unsigned long arg)
{
	dprintk("saa7146_core.o: algo_control called\n");
	return -ENOIOCTLCMD;
}

u32 functionality(struct i2c_adapter *adapter)
{
	/* fixme: there should be a serious return-value here */
	return 0xffffffff;
}

struct i2c_algorithm saa7146_algo = {
	"saa7146 i2c-algorithms",
	I2C_ALGO_SAA7146,
	master_xfer,
	NULL,			/* smbus_xfer */
	NULL,			/* slave_send */
	NULL,			/* slave_recv */
	algo_control,
	functionality,
};

/* registering functions to load algorithms at runtime */
int i2c_saa7146_add_bus(struct saa7146 *adap)
{
	struct i2c_adapter *i2c_adap;

	i2c_adap = kmalloc(sizeof(struct i2c_adapter), GFP_KERNEL);

	if (i2c_adap == NULL)
		return -ENOMEM;

	adap->i2c_adap = i2c_adap;

	/* register new adapter to i2c module... */
	memset(i2c_adap,0,sizeof(struct i2c_adapter));
	strcpy(i2c_adap->name,adap->name);
	i2c_adap->id 	  = saa7146_algo.id;	/* unused */
	i2c_adap->algo	  = &saa7146_algo;
	i2c_adap->data	  = adap;
	i2c_adap->timeout = SAA7146_I2C_TIMEOUT;
	i2c_adap->retries = SAA7146_I2C_RETRIES;

	i2c_add_adapter(i2c_adap);

	/* enable i2c-port pins */
	saa7146_write(adap->mem,MC1, (MASK_08 | MASK_24));

	dprintk("saa7146_core.o: i2c-routines for %s registered.\n",adap->name);

	return 0;
}

int i2c_saa7146_del_bus(struct saa7146 *adap)
{
	i2c_del_adapter(adap->i2c_adap);
	kfree(adap->i2c_adap);
	
	dprintk("saa7146_core.o: adapter unregistered: %s\n",adap->name);

	return 0;
}

/* ---------------------------------------------*/
/* debug-helper function: dump-registers	*/
/* ---------------------------------------------*/

void	dump_registers(unsigned char* mem) {	
	
	u16 j = 0;
	
	for( j = 0x0; j < 0x1fe; j+=0x4 ) {
	printk("0x%03x: 0x%08x\n",j,saa7146_read(mem,j));
	}

}

/* -----------------------------------------------------*/
/* dispatcher-function for handling external commands	*/
/* -----------------------------------------------------*/

static int saa7146_core_command(struct i2c_adapter *adap, unsigned int cmd, void *arg)
{
	int i = 0, result = -ENOIOCTLCMD;
	struct saa7146* saa = adap->data;

	dprintk("saa7146_core.o: ==> saa7146_core_command\n");

	if( NULL == saa)
		return -EINVAL;

	/* first let the extensions handle the command */
	for (i = 0; i < SAA7146_MAX_EXTENSIONS; i++) {
		if (NULL != saa7146_ext[i]) {
			if( -ENOIOCTLCMD != (result = saa7146_ext[i]->command(saa, saa->data[i], cmd, arg))) {
				break;
			}
		}
	}

	/* if command has not been handled by an extension, handle it now */
	if( result == -ENOIOCTLCMD ) {
		
		switch(cmd) {
	        case SAA7146_DUMP_REGISTERS:
		{
			dump_registers(saa->mem);
                        break;
		}
		case SAA7146_SET_DD1:
		{
			u32 *i = arg;

			dprintk(KERN_ERR "saa7146_core.o: SAA7146_SET_DD1 to 0x%08x\n",*i);

			/* set dd1 port register */
			saa7146_write(saa->mem, DD1_INIT, *i);

			/* write out init-values */
			saa7146_write(saa->mem,MC2, (MASK_09 | MASK_10 | MASK_26 | MASK_26));

			break;
		}
		case SAA7146_DO_MMAP: {
			struct saa7146_mmap_struct* mms = arg;

			const char *adr = mms->adr;
			unsigned long size = mms->size;
			unsigned long start=(unsigned long) adr;
			unsigned long page,pos;

			dprintk(KERN_ERR "saa7146_core.o: SAA7146_DO_MMAP.\n");
			
			if (size > saa->buffers * GRABBING_MEM_SIZE)
			        return -EINVAL;

                        if ( NULL == saa->grabbing )
                                return -EINVAL;

			pos=(unsigned long)saa->grabbing;

			while (size > 0) 
			{
			        page = kvirt_to_pa(pos);
				if (remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED))
					return -EAGAIN;
				start	+= PAGE_SIZE;
				pos	+= PAGE_SIZE;
				size	-= PAGE_SIZE;    
			}
				
			break;
		}
		case SAA7146_DEBI_TRANSFER: {
		
			struct saa7146_debi_transfer *dt = arg;

			printk("saa7146_core.o: SAA7146_DEBI_TRANSFER\n");
			printk("saa7146_core.o: timeout:%d, swap:%d, slave16:%d, increment:%d, intel:%d, tien:%d\n", dt->timeout, dt->swap, dt->slave16, dt->increment, dt->intel, dt->tien);
			printk("saa7146_core.o: address:0x%04x, num_bytes:%d, direction:%d, mem:0x%08x\n",dt->address,dt->address,dt->direction,dt->mem);						

			debi_transfer(saa, dt);
			break;
		}
		
		default: {
			return -ENOIOCTLCMD;
		}
		}
	}
	
	return 0;
}

/* -----------------------------------------------------*/
/* dispatcher-function for handling irq-events		*/
/* -----------------------------------------------------*/

/* irq-handler function */
static void saa7146_irq(int irq, void *dev_id, struct pt_regs * regs)
{
	struct saa7146 *saa = (struct saa7146 *)dev_id;
	u32 isr = 0;
	int i;
	int count = 0;

	/* process all interrupts */
	while (1) {

		/* read out the primary status register */
		isr = saa7146_read(saa->mem, ISR);
		/* clear all IRQs */
		saa7146_write(saa->mem, ISR, isr);
	
		/* is anything to do? */
		if ( 0 == isr )
			return;

		dprintk("%s: irq-call: isr:0x%08x\n",saa->name,isr);
		
		/* first let the extensions handle the interrupt */
		for (i = 0; i < SAA7146_MAX_EXTENSIONS; i++) 
			if (saa7146_ext[i] && 
			    (isr&saa7146_ext[i]->handles_irqs)) {
			  saa7146_ext[i]->irq_handler(saa, isr, saa->data[i]);
			  //saa7146_write(saa->mem, ISR, saa7146_ext[i]->handles_irqs);
			}
		
		//printk(KERN_ERR "%s: unhandled interrupt: 0x%08x\n", saa->name, isr);
		
		/* see if we are in a hard interrupt loop */
		++count;
		if (count > 10)
			printk (KERN_WARNING "%s: irq loop %d\n", saa->name, count);
		if (count > 20) {
			saa7146_write(saa->mem, IER, 0x00000000);
			printk(KERN_ERR "%s: IRQ lockup, cleared int mask\n", saa->name);
			break;
		}
	}
}

/* -----------------------------------------------------
   functions for finding any saa7146s in the system,
   inserting/removing module for kernel, etc.
   -----------------------------------------------------*/

int configure_saa7146(struct pci_dev *dev, int i)
{
	u32 rev = 0;
	int result = 0;
	
	hprintk("saa7146_core.o: ==> configure_saa7146\n");

	/* check for maximum number of saa7146-devices */
	if ( i > SAA7146_MAX_DEVICES ) {
		printk(KERN_ERR "saa7146_core.o: maximum number of saa7146-devices reached!\n");
		return -1;
	}

	/* check module-parameters for sanity */

	/* check if wanted number of video-buffers is valid, otherwise fix it */
	//if (buffers < 2)
	//	buffers = 2;

	if ( buffers > SAA7146_MAX_BUF )
		buffers = SAA7146_MAX_BUF;

	/* check if mode is supported */
	switch( mode ) {
		/* 0 = pal, 1 = ntsc */
		case 0:
		case 1:
		{
			break;
		}
		/* default to pal */
		default:
		{
			mode = 0;
			break;
		}
	}

	/* get chip-revision; this is needed to enable bug-fixes */
	if( 0 > pci_read_config_dword(dev, 0x08, &rev)) {
		printk(KERN_ERR "saa7146_core.o: cannot read from pci-device!\n");
		return -1;
	}
	saa7146_adap[i].revision = (rev & 0xf);
	
	sprintf(saa7146_adap[i].name, "saa7146(%d)", i+1);	
	saa7146_adap[i].device = dev;
		
	/* remap the memory from virtual to physical adress */
#if LINUX_VERSION_CODE >= 0x020300
	saa7146_adap[i].mem=ioremap((saa7146_adap[i].device->resource[0].start)
				    &PCI_BASE_ADDRESS_MEM_MASK, 0x1000);
#else
	saa7146_adap[i].mem=ioremap((saa7146_adap[i].device->base_address[0])
				    &PCI_BASE_ADDRESS_MEM_MASK, 0x1000);
#endif

	if ( 0 == saa7146_adap[i].mem ) {
	    	printk(KERN_ERR "saa7146_core.o: cannot map pci-address!\n");
		return -1;
	}
	
	/* get clipping memory */	
	saa7146_adap[i].clipping=(u32*)kmalloc(CLIPPING_MEM_SIZE*sizeof(u32),GFP_KERNEL);
	if ( 0 == saa7146_adap[i].clipping ) {
	    	printk(KERN_ERR "saa7146_core.o: not enough kernel-memory for clipping!\n");
		return -1;
	}
	memset(saa7146_adap[i].clipping, 0x0, CLIPPING_MEM_SIZE*sizeof(u32));

	/* get i2c memory */	
	saa7146_adap[i].i2c=(u32*)kmalloc(I2C_MEM_SIZE*sizeof(u32),GFP_KERNEL); /*64*/
	if ( 0 == saa7146_adap[i].i2c ) {
	    	printk(KERN_ERR "saa7146_core.o: not enough kernel-memory for i2c!\n");
		kfree(saa7146_adap[i].clipping);
		return -1;
	}
	memset(saa7146_adap[i].i2c, 0x0, I2C_MEM_SIZE*sizeof(u32));
	
	/* get grabbing memory */
	saa7146_adap[i].grabbing = (u32*)rvmalloc(buffers,&saa7146_adap[i].page_table[0]);
	if ( 0 == saa7146_adap[i].grabbing ) {
	    	printk(KERN_ERR "saa7146_core.o: not enough kernel-memory for grabbing_mem!\n");
		kfree(saa7146_adap[i].i2c);
		kfree(saa7146_adap[i].clipping);
		return -1;
	}

	/* get rps0 memory */
	saa7146_adap[i].rps0 = (u32*)kmalloc(RPS_MEM_SIZE*sizeof(u32),GFP_KERNEL);
	if ( 0 == saa7146_adap[i].rps0 ) {
	    	printk(KERN_ERR "saa7146_core.o: not enough kernel-memory for rps0_mem!\n");
		kfree(saa7146_adap[i].i2c);
		kfree(saa7146_adap[i].clipping);
		rvfree(saa7146_adap[i].grabbing, buffers, &saa7146_adap[i].page_table[0]);
		return -1;
	}
	memset(saa7146_adap[i].rps0, 0x0, RPS_MEM_SIZE*sizeof(u32));

	/* get rps1 memory */
	saa7146_adap[i].rps1 = (u32*)kmalloc(RPS_MEM_SIZE*sizeof(u32),GFP_KERNEL);
	if ( 0 == saa7146_adap[i].rps1 ) {
	    	printk(KERN_ERR "saa7146_core.o: not enough kernel-memory for rps1_mem!\n");
		kfree(saa7146_adap[i].rps0);
		kfree(saa7146_adap[i].i2c);
		kfree(saa7146_adap[i].clipping);
		rvfree(saa7146_adap[i].grabbing, buffers, &saa7146_adap[i].page_table[0]);
		return -1;
	}
	memset(saa7146_adap[i].rps1, 0x0, RPS_MEM_SIZE*sizeof(u32));

	/* get debi memory (32kB) */
	saa7146_adap[i].debi = (u32*)kmalloc(8192*sizeof(u32),GFP_KERNEL);
	if ( 0 == saa7146_adap[i].debi ) {
	    	printk(KERN_ERR "saa7146_core.o: not enough kernel-memory for debi_mem!\n");
		kfree(saa7146_adap[i].rps1);
		kfree(saa7146_adap[i].rps0);
		kfree(saa7146_adap[i].i2c);
		kfree(saa7146_adap[i].clipping);
		rvfree(saa7146_adap[i].grabbing, buffers, &saa7146_adap[i].page_table[0]);
		return -1;
	}
	memset(saa7146_adap[i].debi, 0x0, 8192*sizeof(u32));


	/* clear out memory for grabbing information */
	memset(&saa7146_adap[i].grab_width[0],  0x0, sizeof(int)*SAA7146_MAX_BUF);
	memset(&saa7146_adap[i].grab_height[0], 0x0, sizeof(int)*SAA7146_MAX_BUF);
	memset(&saa7146_adap[i].grab_format[0], 0x0, sizeof(int)*SAA7146_MAX_BUF);
	memset(&saa7146_adap[i].grab_port[0], 0x0, sizeof(int)*SAA7146_MAX_BUF);

	/* init the frame-status array */
	memset(&saa7146_adap[i].frame_stat[0], GBUFFER_UNUSED, sizeof(int)*SAA7146_MAX_BUF);
	
	/* clear out all wait queues */
	init_waitqueue_head(&saa7146_adap[i].rps0_wq);
	init_waitqueue_head(&saa7146_adap[i].rps1_wq);
	

    	/* request an interrupt for the saa7146 */
	result = request_irq(	saa7146_adap[i].device->irq,
				saa7146_irq, SA_SHIRQ | SA_INTERRUPT,
				saa7146_adap[i].name,
				(void *)&saa7146_adap[i]);
			
	switch(result) {
	
		case -EINVAL:
		{
			printk(KERN_ERR "saa7146_core.o: Bad irq number or handler\n");
			return -EINVAL;
		}
		case -EBUSY:
		{
			printk(KERN_ERR "saa7146_core.o: IRQ %d busy, change your PnP config in BIOS\n",saa7146_adap[i].device->irq);
			return -EBUSY;
		}
		case 0:
		{
			break;
		}
		default:
		{
			return result;
		}
	}
	
	/* print status message */
    	printk(KERN_ERR "saa7146_core.o: saa7146(%d): bus:%d, rev:%d, mem:0x%08x.\n", i+1, saa7146_adap[i].device->bus->number, saa7146_adap[i].revision, (unsigned int)saa7146_adap[i].mem);

	/* enable bus-mastering */
 	pci_set_master( saa7146_adap[i].device );

	/* disable everything on the saa7146, perform a software-reset */
	saa7146_write(saa7146_adap[i].mem, MC1, 0xbfff0000);
	mdelay(2);
#if 0
	{
		int j;

		/* clear all registers */
		for( j = 0x0; j < 0xfc; j+=0x4 ) {
			saa7146_write(saa7146_adap[i].mem,j, 0x0000000);
		}
		for( j = 0x104; j < 0x1fc; j+=0x4 ) {
			saa7146_write(saa7146_adap[i].mem,j, 0x0000000); 
		}
	}
#endif 
	/* clear out any rps-signals pending */
	saa7146_write(saa7146_adap[i].mem, MC2, 0xf8000000);

	/* enable video-port-pins*/
	saa7146_write(saa7146_adap[i].mem,MC1, (MASK_10 | MASK_26));

	/* disable all interrupt-conditions, only enable RPS interrupts */
	saa7146_write(saa7146_adap[i].mem, ISR, 0xffffffff);
	saa7146_write(saa7146_adap[i].mem, IER, (MASK_27 | MASK_28));
/*
	printk("main: 0x114: 0x%08x\n",saa7146_read(saa7146_adap[i].mem, 0x114));
	printk("main: 0x0e4: 0x%08x\n",saa7146_read(saa7146_adap[i].mem, 0x0e4));
	printk("PSR:   0x%08x\n",saa7146_read(saa7146_adap[i].mem, PSR));
	printk("SSR:   0x%08x\n",saa7146_read(saa7146_adap[i].mem, SSR));
	printk("IER:   0x%08x\n",saa7146_read(saa7146_adap[i].mem, IER));
	printk("ISR:   0x%08x\n",saa7146_read(saa7146_adap[i].mem, ISR));
*/

	saa7146_write(saa7146_adap[i].mem,PCI_BT_V1, 	0x1c00101f);
	saa7146_write(saa7146_adap[i].mem,BCS_CTRL,	0x80400040);

	/* set dd1 stream a & b */
      	saa7146_write(saa7146_adap[i].mem, DD1_STREAM_B,   0x00000000);
	saa7146_write(saa7146_adap[i].mem, DD1_INIT,	   0x02000000);
	saa7146_write(saa7146_adap[i].mem, MC2, (MASK_09 | MASK_25 | MASK_10 | MASK_26));

	saa7146_write(saa7146_adap[i].mem,MC2, 0x077c077c);

	/* the Siemens DVB needs this if you want to have the I2c chips
           get recognized before the main driver is loaded 
	*/
        saa7146_write(saa7146_adap[i].mem, GPIO_CTRL, 0x500000);

	saa7146_adap[i].command = &saa7146_core_command;
	saa7146_adap[i].client_register   = NULL; /* &client_register; */
	saa7146_adap[i].client_unregister = NULL; /* &client_unregister; */
	saa7146_adap[i].buffers = buffers;
	saa7146_adap[i].mode = mode;
	saa7146_adap[i].interlace = 1;
		
	/* add new bus (adapter gets added automagically) */
	i2c_saa7146_add_bus(&saa7146_adap[i]);

        saa7146_write(saa7146_adap[i].mem, GPIO_CTRL, 0x000000);
	return 0;
}


#ifdef MODULE
MODULE_AUTHOR("Michael Hunold <michael@mihu.de>");
MODULE_DESCRIPTION("core-driver for the saa7146");

struct saa7146* saa7146_get_handle(int i)
{
	hprintk("saa7146_core.o: ==> saa7146_get_handle\n");
	
	if( i >= saa7146_num )
		return NULL;
		
	return &saa7146_adap[i];
}

int	saa7146_add_extension(struct saa7146_extension* ext)
{
	int i = 0, j = 0;

	hprintk("saa7146_core.o: ==> saa7146_attach_extension\n");

	for (i = 0; i < SAA7146_MAX_EXTENSIONS; i++) {
		if (NULL == saa7146_ext[i])
			break;
	}
	if (SAA7146_MAX_EXTENSIONS == i) {
		printk(KERN_WARNING "saa7146.o: attach_extension(%s) - enlarge SAA7146_MAX_EXTENSIONS.\n",ext->name);
		return -ENOMEM;
	}
	saa7146_ext[i] = ext;
	saa7146_extension_count++;

	/* call back extension handler, so extension-specific stuff can be initialized */
	for(j = 0; j < saa7146_num; j++ ) {
		if( NULL != ext->attach ) {
			if( 0 != ext->attach(&saa7146_adap[j],&saa7146_adap[j].data[i])) {
				printk(KERN_WARNING "saa7146.o: attach_extension(%s) - could not increment use-count.\n",ext->name);
			}
		}
	}
	
	return 0;
}

int	saa7146_del_extension(struct saa7146_extension* ext)
{
	int i = 0, j = 0;

	hprintk("saa7146_core.o: ==> saa7146_detach_extension\n");

	for (i = 0; i < SAA7146_MAX_EXTENSIONS; i++)
		if (ext == saa7146_ext[i])
			break;
	if (SAA7146_MAX_EXTENSIONS == i) {
		printk( "saa7146.o: detach_extension extension [%s] not found.\n", ext->name);
		return -ENODEV;
	}		

	/* all done, now unregister */
	saa7146_ext[i] = NULL;
	saa7146_extension_count--;		

	for(j = 0; j < saa7146_num; j++ ) {

		if( NULL != ext->detach ) {
			if( 0 != ext->detach(&saa7146_adap[j],&saa7146_adap[j].data[i])) {
				printk(KERN_WARNING "saa7146.o: detach_extension(%s) - could not decrement use-count.\n",ext->name);
			}
		}
	}
	
	return 0;
}
EXPORT_SYMBOL(saa7146_get_handle);
EXPORT_SYMBOL(saa7146_add_extension);
EXPORT_SYMBOL(saa7146_del_extension);


#endif


void remove_saa7146(struct saa7146 *saa)
{
        /* shut down all dma transfers */
        saa7146_write(saa->mem, MC1, 0xbfff0000);
	
	printk("free irqs\n");
	/* disable alle irqs, release irq-routine */
	saa7146_write(saa->mem, IER, 0x00);
	saa7146_write(saa->mem, ISR, 0xffffffff);
	free_irq(saa->device->irq, (void *)saa);
	dprintk("unmap memory\n");
	/* unmap the memory, if necessary */
	if (saa->mem)
		iounmap((unsigned char *)((unsigned int)saa->mem));
	
	dprintk("release grabbing memory\n");
	/* release grabbing memory */
	if(saa->grabbing)
		rvfree(saa->grabbing, buffers, &saa->page_table[0]);
	
	dprintk("release other memory\n");
	/* release clipping, i2c, rps0 memory */
	kfree(saa->clipping);
	kfree(saa->i2c);
	kfree(saa->rps0);
	kfree(saa->rps1);
	kfree(saa->debi);
	
	dprintk("i2c_del_bus\n");
	/* nothing else to do here */
	i2c_saa7146_del_bus(saa);
}

#if LINUX_VERSION_CODE >= 0x020406
static int saa7146_suspend(struct pci_dev *pdev, u32 state)
#else
static void saa7146_suspend(struct pci_dev *pdev)
#endif
{
        printk("saa7146_suspend()\n");
	saa7146_core_command(((struct saa7146 *) pdev->driver_data)->i2c_adap,
			     SAA7146_SUSPEND, 0);

#if LINUX_VERSION_CODE >= 0x020406
	return 0;
#endif
}

#if LINUX_VERSION_CODE >= 0x020406
static int
#else
static void 
#endif
saa7146_resume(struct pci_dev *pdev)
{
        printk("saa7146_resume()\n");
	saa7146_core_command(((struct saa7146 *) pdev->driver_data)->i2c_adap,
			     SAA7146_RESUME, 0);

#if LINUX_VERSION_CODE >= 0x020406
	return 0;
#endif
}


static int __devinit saa7146_init_one (struct pci_dev *pdev,
		const struct pci_device_id *ent)
{
	uint32_t cardid;
	
	dprintk("saa7146_init_one()\n");

	pci_read_config_dword(pdev, PCI_SUBSYSTEM_VENDOR_ID, &cardid);
	if (cardid == 0x10413)  /* reject the stradis card */
		return 0;
	pci_enable_device(pdev);
	pdev->driver_data=&saa7146_adap[saa7146_num];
	configure_saa7146(pdev, saa7146_num++);
	return 0;
}

static void __devexit saa7146_remove_one (struct pci_dev *pdev)
{
	struct saa7146 *saa=pdev->driver_data;
	uint32_t cardid;
	
	dprintk("saa7146_remove_one()\n");
	pci_read_config_dword(pdev, PCI_SUBSYSTEM_VENDOR_ID, &cardid);
	if (cardid != 0x10413) {  /* reject the stradis card */
		pci_disable_device(pdev);
		remove_saa7146(saa);
	}
}


static struct pci_device_id saa7146_pci_tbl[] __devinitdata = {
	{ PCI_VENDOR_ID_PHILIPS, PCI_DEVICE_ID_PHILIPS_SAA7146,
		PCI_ANY_ID, PCI_ANY_ID, },
	{ 0,},
};

MODULE_DEVICE_TABLE(pci, saa7146_pci_tbl);
	
static struct pci_driver saa7146_driver = {
	name:		"saa7146",
	id_table:	saa7146_pci_tbl,
	probe:		saa7146_init_one,
	remove:		saa7146_remove_one,
	suspend:	saa7146_suspend,
	resume:		saa7146_resume,
};


#ifdef MODULE
static int __init saa7146_init_module(void)
{
	dprintk("saa7146_init_module\n");
	return pci_module_init(&saa7146_driver);
}

static void __exit saa7146_cleanup_module(void)
{
	pci_unregister_driver(&saa7146_driver);
}

module_init(saa7146_init_module);
module_exit(saa7146_cleanup_module);
#endif


