/*********************************************************************************/
/* $Id: AdiUsbAdslDriver.c,v 1.30 2002/05/24 17:56:44 Exp $                      */
/*-------------------------------------------------------------------------------*/
/*										 */
/* Copyright (c) 2002, Analog Devices Inc., All Rights Reserved			 */
/* User space interface rewritten by C.Casteyde (casteyde.christian@free.fr)	 */
/*										 */
/* AdiUsbAdslDriver.c								 */
/*										 */
/* Main code for Linux kernel driver.					         */
/*										 */
/*-------------------------------------------------------------------------------*/
/* This file is part of the "ADI USB ADSL Driver for Linux".   			 */
/*   										 */
/* "ADI USB ADSL Driver for Linux" is free software; you can redistribute it     */
/* and/or modify it under the terms of the GNU General Public License as         */
/* published by the Free Software Foundation; either version 2 of the License,   */
/* or (at your option) any later version.   				         */
/*									         */
/* "ADI USB ADSL Driver for Linux" is distributed in the hope that it will be    */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   		 */
/* GNU General Public License for more details.   				 */
/*   										 */
/* You should have received a copy of the GNU General Public License   		 */
/* along with "ADI USB ADSL Driver for Linux"; if not, write to the Free Software*/
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA     */
/*********************************************************************************/

#include "Adiutil.h"
#include <linux/module.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/if_arp.h>
#include "AdiUsbAdslDriver.h"
#include "Util.h"
#include "Dsp.h"
#include "Msg.h"
#include "Boot.h"
#include "Pipes.h"
#include "Crc.h"
#include "Mpoa.h"
#include "Me.h"
#include "Uni.h"


/*********************************************************************************/
/* Local prototypes								 */
/*********************************************************************************/

/* USB driver */
static void *adi_probe(struct usb_device *usb, unsigned int ifnum, const struct usb_device_id *id);
static void adi_disconnect(struct usb_device *usb, void *ptr);
static void adi_irq(struct urb *urb);
static int adi_user(struct usb_device *dev, unsigned int code, void *buf);
static int adi_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data);

/* Ethernet device */
static int create_etherdev(void);
static int adi_open(struct net_device *ether);
static int adi_close(struct net_device *ether);
static int adi_ioctl(struct net_device *ether, struct ifreq *rq, int cmd);
static struct net_device_stats *adi_stats(struct net_device *ether);
static int adi_start_xmit(struct sk_buff *skb, struct net_device *ether);
static void adi_set_multicast(struct net_device *ether);
static void adi_tx_timeout(struct net_device *ether);


/*********************************************************************************/
/* Globals and local utilities                                                   */
/*********************************************************************************/

/*Tell usb what VID/PID combinations we support*/
static const struct usb_device_id adi_ids[] =
{
    {  USB_DEVICE (kEAGLE_VID, kEAGLE_PID_PREFIRMWARE) }, /* before USB firmware upload */
    {  USB_DEVICE (kEAGLE_VID, kEAGLE_PID_PSTFIRMWARE) }, /* after  USB firmware upload */
    {  USB_DEVICE (kUSR_VID, kMILLER_A_PID_PREFIRMWARE) },
    {  USB_DEVICE (kUSR_VID, kMILLER_A_PID_PSTFIRMWARE) },
    {  USB_DEVICE (kUSR_VID, kMILLER_B_PID_PREFIRMWARE) },
    {  USB_DEVICE (kUSR_VID, kMILLER_B_PID_PSTFIRMWARE) },
    {  USB_DEVICE (kUSR_VID, kHEINEKEN_A_PID_PREFIRMWARE) },
    {  USB_DEVICE (kUSR_VID, kHEINEKEN_A_PID_PSTFIRMWARE) },
    {  USB_DEVICE (kUSR_VID, kHEINEKEN_B_PID_PREFIRMWARE) },
    {  USB_DEVICE (kUSR_VID, kHEINEKEN_B_PID_PSTFIRMWARE) },
    { }
};

/*send this to USB to register the driver*/
static struct usb_driver adi_driver =
{
    .name       = "adiusbadsl",
    .id_table   = adi_ids,
    .probe      = adi_probe,
    .disconnect = adi_disconnect,
    .ioctl      = adi_user
};

/*We only support one physical modem plugged in (who would plug in more than one?)*/
static Hardware modem;

/*Default driver options*/
static const DriverOptions DefaultOptions =
{
    { "OPTN0", DEFAULT_OPTN0 },
    { "OPTN2", DEFAULT_OPTN2 },
    { "OPTN3", DEFAULT_OPTN3 },
    { "OPTN4", DEFAULT_OPTN4 },
    { "OPTN5", DEFAULT_OPTN5 },
    { "OPTN6", DEFAULT_OPTN6 },
    { "OPTN7", DEFAULT_OPTN7 },
    { "OPTN15", DEFAULT_OPTN15 },
    { "VPI", DEFAULT_VPI },
    { "VCI", DEFAULT_VCI },
    { "Encapsulation", DEFAULT_ENCAPS },
    { "Linetype", DEFAULT_LINETYPE },
    { "RatePollFreq", DEFAULT_POLLFREQ },
};

static int CheckOptions(const DriverOptions aOptions)
{
    int result = 1;
    if (aOptions[ConfigEncapsulation].Value != MPOA_MODE_BRIDGED_ETH_LLC &&
	aOptions[ConfigEncapsulation].Value != MPOA_MODE_BRIDGED_ETH_VC &&
	aOptions[ConfigEncapsulation].Value != MPOA_MODE_ROUTED_IP_LLC &&
	aOptions[ConfigEncapsulation].Value != MPOA_MODE_ROUTED_IP_VC &&
	aOptions[ConfigEncapsulation].Value != MPOA_MODE_PPPOA_LLC &&
	aOptions[ConfigEncapsulation].Value != MPOA_MODE_PPPOA_VC)
	result = 0;
    return result;
}

struct EthArpPkt /* copied from if_arp.h */
{
    unsigned char   h_dest[ETH_ALEN];	
    unsigned char   h_source[ETH_ALEN];	
    unsigned short  h_proto;
    unsigned short	ar_hrd;		/* format of hardware address	*/
    unsigned short	ar_pro;		/* format of protocol address	*/
    unsigned char	ar_hln;		/* length of hardware address	*/
    unsigned char	ar_pln;		/* length of protocol address	*/
    unsigned short	ar_op;		/* ARP opcode (command)		*/
    
    /*
     *	 Ethernet looks like this : This bit is variable sized however...
     */
    unsigned char		ar_sha[ETH_ALEN];	/* sender hardware address	*/
    unsigned char		ar_sip[4];		/* sender IP address		*/
    unsigned char		ar_tha[ETH_ALEN];	/* target hardware address	*/
    unsigned char		ar_tip[4];		/* target IP address		*/
};

static int LocalArpReply(Hardware *pHw, UInt8 *pPacket, UInt32 Size)
{
    struct EthArpPkt *pArpPkt = (struct EthArpPkt *)pPacket;
    struct EthArpPkt ReplyArpPkt;   
    struct sk_buff *skb;
   
    if(Size < sizeof(struct EthArpPkt)) 
	return 1;

    if(pArpPkt->ar_op != __constant_htons(ARPOP_REQUEST))
	return 1;
    /* FIXME: Check IP address? ARP should filter that out */
    
    memcpy((unsigned char *)&ReplyArpPkt,pPacket, sizeof(ReplyArpPkt));
    ReplyArpPkt.ar_op = __constant_htons(ARPOP_REPLY);
    memcpy( ReplyArpPkt.ar_sip, pArpPkt->ar_tip, 4);
    memcpy(ReplyArpPkt.ar_tip, pArpPkt->ar_sip, 4);

    /* build HW address using 00:00:IP_ADDRESS[4] */
    memcpy(ReplyArpPkt.ar_sha + 2, pArpPkt->ar_tip, 4);
    ReplyArpPkt.ar_sha[0] = 0;
    ReplyArpPkt.ar_sha[1] = 0;
    memcpy(ReplyArpPkt.ar_tha, pArpPkt->ar_sha, ETH_ALEN);
    memcpy(ReplyArpPkt.h_dest, pArpPkt->h_source, ETH_ALEN);
    memcpy(ReplyArpPkt.h_source, ReplyArpPkt.ar_sha, ETH_ALEN);

    if (!(skb = dev_alloc_skb(sizeof(ReplyArpPkt))))
    {
	ZAP(DEBUG_READ, DEBUG_ERRORS, (kTAB"Error allocating skb!\n"));
	return 1;
    }
    skb->dev = pHw->pLinuxNet;
    eth_copy_and_sum(skb, (UInt8 *)&ReplyArpPkt, sizeof(ReplyArpPkt), 0);
    skb_put(skb, sizeof(ReplyArpPkt));
    skb->protocol = eth_type_trans(skb, pHw->pLinuxNet);
    netif_rx(skb);
    pHw->pLinuxNet->last_rx = jiffies;
    pHw->LinuxStats.rx_packets++;
    pHw->LinuxStats.rx_bytes += sizeof(ReplyArpPkt);
    return 0;
}

/*********************************************************************************/
/* MODULE stuff									 */
/*********************************************************************************/
MODULE_AUTHOR ("Anoosh Naderi <anoosh.naderi@analog.com>");
MODULE_DESCRIPTION ("Analog Devices Inc. USB ADSL Modem driver");
MODULE_DEVICE_TABLE (usb, adi_ids);
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;


/*********************************************************************************/
/* Code that's part of the USB driver interface					 */
/*********************************************************************************/

/*********************************************************************************/
/* adi_init									 */
/*********************************************************************************/
static int __init adi_init (void)
{
    ZAP(DEBUG_INIT, DEBUG_ENTRY, ("+++adi_init\n"));
    
    printk(KERN_INFO "ADI: driver V"ADIDRIVERVERSION" loaded\n");

    CrcGenerateTable();
    usb_register(&adi_driver);

    ZAP(DEBUG_INIT, DEBUG_EXIT, ("---adi_init\n"));
    return 0;
}

module_init (adi_init);


/*********************************************************************************/
/* adi_exit									 */
/*********************************************************************************/
static void __exit adi_exit (void)
{
    ZAP(DEBUG_INIT, DEBUG_ENTRY, ("+++adi_exit\n"));
    
    /* This calls automatically the adi_disconnect method if necessary: */
    usb_deregister (&adi_driver);
    printk(KERN_INFO "ADI: driver unloaded\n");
    
    ZAP(DEBUG_INIT, DEBUG_EXIT, ("---adi_exit\n"));
}

module_exit (adi_exit);


/*********************************************************************************/
/* adi_probe									 */
/*********************************************************************************/
static void *adi_probe(struct usb_device *usb, unsigned int ifnum, const struct usb_device_id *id)
{
    /* Prepare code for multi-device by using an abstract pointer to hardware: */
    UInt32 pid = usb->descriptor.idProduct;
    void *pRet = NULL;

    ZAP(DEBUG_INIT, DEBUG_ENTRY, ("+++adi_probe\n"));

    ZAP(DEBUG_INIT, DEBUG_INFO, (kTAB"vid (%#X) pid (%#X), ifnum (%d)\n",
        usb->descriptor.idVendor, usb->descriptor.idProduct, ifnum));

    /* This driver knows only pre and postfirmware devices. */
    if (!ISPREFIRMWARE(pid) && !ISPOSTFIRMWARE(pid))
    {
        ZAP(DEBUG_INIT, DEBUG_INFO, (kTAB"not an ADI modem\n"));
        goto probe_exit;
    }

    /* Try to select the first USB configuration: */
    if (usb_set_configuration(usb, usb->config[0].bConfigurationValue) < 0)
    {
        ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"set_configuration failed\n"));
        goto probe_exit;
    }

    switch (pid)
    {
    case kCASE_PREFIRMWARE:
	{
	    /* Create a new pre-firmware device: */
	    Hardware *pHw = &modem;
	    printk(KERN_INFO "ADI: New modem detected, waiting for firmware...\n");
	    
	    /* Reinitialize the modem structure: */
	    memset(pHw, 0, sizeof(Hardware));
	    pHw->usbdev = usb;
	    sema_init(&pHw->FirmwareSem, 1);
	    
	    pRet = pHw;
	}
        break;
    case kCASE_POSTFIRMWARE:
	{
	    int i;
	    int ret;
            struct usb_endpoint_descriptor *epindata;
            struct usb_endpoint_descriptor *epoutdata;
            struct usb_endpoint_descriptor *epoutidma;
	    /* Create a new post-firmware device: */
	    Hardware *pHw = &modem;
   	    printk(KERN_INFO "ADI: New USB ADSL device detected, waiting for DSP code...\n");

	    /* Reinitialize the modem structure (and the state machine): */
	    memset(pHw, 0, sizeof(Hardware));
	    pHw->usbdev = usb;
	    init_waitqueue_head(&pHw->OperationalQueue);
	    pHw->DSPLock = RW_LOCK_UNLOCKED;
	    pHw->NetLock = SPIN_LOCK_UNLOCKED;
	    sema_init(&pHw->NetSem, 1);
	    
            /**********************************************************************/
            /* Some one-time init stuff						  */
            /* For now we'll just choose the biggest pipe. This will insure       */
            /* that we'll be ok no matter what the train rate is. Unfortunately,  */
            /* the driver model doesn't fit well with delaying read pipe size     */
            /* figuring, etc., until after we're trained.			  */
            /**********************************************************************/
            pHw->IsoPipeSize     = kFASTEST_ISO_RATE;
            pHw->IsoFramesPerUrb = kFRAMES_PER_ISO_URB;

            /**********************************************************************/
            /*We use constants for the endpoint addresses - but this is the       */
            /*easiest way for now - and most of the other device-specific         */
            /*usb drivers do the same. Eliminates us having to parse through      */
            /*the configuration descriptor, etc., to get these numbers            */
            /**********************************************************************/
            pHw->pipeBulkIdmaOut = usb_sndbulkpipe(usb, kEP_BULK_IDMA_OUT);
            pHw->pipeBulkDataOut = usb_sndbulkpipe(usb, kEP_BULK_DATA_OUT);
            pHw->pipeBulkDataIn  = usb_rcvbulkpipe(usb, kEP_BULK_DATA_IN);
            pHw->pipeIsoDataIn   = usb_rcvisocpipe(usb, kEP_ISOC_DATA_IN);
            pHw->pipeIntIn       = usb_rcvintpipe(usb, kEP_INT_IN);

            /*Make sure we can get all the memory we need*/
            pHw->pInterruptData = GET_KBUFFER(sizeof(CDC_NOTIFY));
            pHw->pReadyData     = GET_KBUFFER(ETHERNET_PACKET_DATA_SIZE);
            pHw->pOutgoingData  = GET_KBUFFER(OUTGOING_DATA_SIZE);
            for (i=0; i<INCOMING_Q_SIZE; i++)
            {
#ifdef USEBULK
                pHw->pIncomingData[i] = GET_KBUFFER(INCOMING_DATA_SIZE);
#else
                pHw->pUrbReadIso[i]   = usb_alloc_urb(pHw->IsoFramesPerUrb);
                pHw->pIncomingData[i] = GET_KBUFFER((pHw->IsoPipeSize * pHw->IsoFramesPerUrb));
#endif
                if (!pHw->pIncomingData[i])
                {
                    ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"NULL pointer returned from GET_KBUFFER\n"));
		    /* FIXME Memory leak */
                    goto probe_exit;
                }
            }

            /*If any of those buffers were not succesfully allocated, we*/
            /*cannot proceed succesfully                                */
            if ( (!pHw->pInterruptData) ||
                 (!pHw->pReadyData)     ||
                 (!pHw->pOutgoingData) )
            {
		/* FIXME Memory leak */
                ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"NULL pointer returned from GET_KBUFFER\n"));
                goto probe_exit;
            }

            /*************************************************************************/
            /* These buffers are used to make the Uni code happy. It was written     */
            /* for MacOS9 - where we had to keep an internal queue of incoming       */
            /* and outgoing packets. Since we don't have to do that here on Linux,   */
            /* we essentially create 1-deep queues and slightly modified the         */
            /* Uni code so it just uses these buffers instead of pulling from queues.*/
            /*************************************************************************/
            pHw->ReassemblyBuffer.GB.AllocedSize   = ETHERNET_PACKET_DATA_SIZE;
            pHw->ReassemblyBuffer.GB.pData         = pHw->pReadyData;
            pHw->SegmentationBuffer.GB.AllocedSize = OUTGOING_DATA_SIZE;
            pHw->SegmentationBuffer.GB.pData       = pHw->pOutgoingData;

            /*Initialize the CTRL URB queue*/
            ret = AllocQueuedBuffers(pHw, CTRL_URB_Q_SIZE, sizeof(QUEUED_URB_BUFFER), 
                                   0, &pHw->CtrlUrbFreeQ, FALSE);
            if (ret != USB_ST_NOERROR)
            {
                ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"Ran out of memory in AllocQueuedBuffers\n"));
		/* FIXME Memory leak */
                goto probe_exit;
            }

            INIT_LIST_HEAD(&pHw->CtrlUrbPendQ);

            /*************************************************************************/
            /* Initialize our kernel timers					     */
            /* WatchdogTimer is the "management entity", per the ADI documentation.  */
            /* It is responsible for initiating a regular status check of the modem, */
            /* and it is the main "driver" for the state machine.		     */
            /*************************************************************************/
            init_timer(&pHw->AdiModemSm.timerSM);
            pHw->AdiModemSm.timerSM.function = WatchdogTimer;
            pHw->AdiModemSm.timerSM.data     = (unsigned long) pHw;

            /*************************************************************************/
            /* The UHCI usb driver does not allow us to queue control urbs.          */
            /* Therefore, to make my own asynchronous control urb queue, I           */
            /* have a timer that goes off after 1 second after I sent a control      */
            /* urb to check to see if another control urb needs submitted. 	     */
            /*************************************************************************/
            init_timer(&pHw->CtrlUrbQueueTimer);
            pHw->CtrlUrbQueueTimer.function  = CtrlUrbQueueWatcher;
            pHw->CtrlUrbQueueTimer.data      = (unsigned long) pHw; 

            /*Other init stuff*/
            pHw->AdiModemSm.HeartbeatCounter = 0; 
            pHw->AdiModemSm.CurrentAdiState  = STATE_JUST_PLUGGED_IN;

#ifdef USEBULK
            /*****************************************************************************/
            /* Other drivers seem to be calling usb_set_interface, but we're not using an*/
            /* alternate interface on any of the interfaces, so we don't need to as far  */
            /* as I can understand. All we really need to do from a USB stack perspective*/
            /* is tell usb that we will handle all the interfaces for this device, so he */
            /* doesn't enumerate through all drivers for each interface.		 */
            /*****************************************************************************/
#endif
            usb_driver_claim_interface(&adi_driver, &(usb->actconfig->interface[0]), pHw);
            usb_driver_claim_interface(&adi_driver, &(usb->actconfig->interface[1]), pHw);
            usb_driver_claim_interface(&adi_driver, &(usb->actconfig->interface[2]), pHw);
            /* Get information about our endpoints*/
            epoutidma = usb->actconfig->interface[1].altsetting[0].endpoint + 0;
            epoutdata = usb->actconfig->interface[1].altsetting[0].endpoint + 1; 
#ifdef USEBULK
            epindata  = usb->actconfig->interface[2].altsetting[0].endpoint + 0;
#else /*ifdef USEBULK*/
            /*******************************************************************************/
            /* Need to eventually choose the alternate interface based on the train ds rate*/
            /* But, like I said above, the driver model doesn't fit well into the	   */
            /* concept of delaying the ISO pipe startup					   */
            /*******************************************************************************/
            if (usb_set_interface(usb, kUSB_INTF_IN, kFASTEST_ISO_INTF) < 0)
            {
                ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"usb_set_interface failed on iso alt 7\n"));
            }
            else
            {
                epindata=usb->actconfig->interface[kUSB_INTF_IN].altsetting[kFASTEST_ISO_INTF].endpoint+0; 
            }
#endif /*ifdef USEBULK else*/

            /*Setup our /proc interface*/
            create_proc_read_entry("driver/adimodem", 0, NULL, adi_read_proc, pHw);
	    
	    /*******************************************************************************/
	    /* Even though the rest of the drivers that claim multiple interfaces	       */
	    /* just go ahead and return NULL from the initial probe - that causes	       */
	    /* the usb stack to think the device isn't claimed, so it goes and tries       */
	    /* to load another driver. In our case, since the device advertises 	       */
	    /* itself as an ethernet control model comm class device, the acm and	       */
	    /* cdcether drivers get loaded. So, we'll return our private pointer	       */
	    /* here too (in addition to the usb_driver_claim_interface above), to	       */
	    /* keep usb from trying to load another driver				       */
	    /*******************************************************************************/
	    pRet = pHw;
        }
        break;
    }

probe_exit:
    ZAP(DEBUG_INIT, DEBUG_EXIT, ("---adi_probe\n"));
    return pRet;
}

/*******************************************************************************/
/* adi_disconnect							       */
/*******************************************************************************/
static void adi_disconnect(struct usb_device *usb, void *ptr)
{
    Hardware *pHw = (Hardware *)ptr;
    UInt32 pid = usb->descriptor.idProduct;

    ZAP(DEBUG_INIT, DEBUG_ENTRY, ("+++adi_disconnect\n"));

    /*We can't do anything if we don't have a valid Hardware pointer*/
    if (pHw == NULL)
    {
        ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"NULL pointer passed to disconnect\n"));
        goto dis_done;
    }
    
    switch (pid)
    {
    case kCASE_PREFIRMWARE:
	/* Nothing has to be done for pre-firmware devices: */
	printk(KERN_INFO "ADI: Modem removed\n");
	break;
    case kCASE_POSTFIRMWARE:
	{
	    int i;
	    /* If timers are currently running, delete them: */
	    if (timer_pending(&pHw->AdiModemSm.timerSM))
		del_timer(&pHw->AdiModemSm.timerSM);
	    if (timer_pending(&pHw->CtrlUrbQueueTimer))
		del_timer(&pHw->CtrlUrbQueueTimer);
	    
	    read_lock(&pHw->DSPLock);
	    /* Unlink pending interrupt URBs: */
	    if (pHw->HasIntURB)
	    {
		pHw->urbInt.transfer_flags &= ~USB_ASYNC_UNLINK;
		usb_unlink_urb(&pHw->urbInt);
		pHw->HasIntURB = FALSE;
	    }
	    /* Free the DSP code: */
	    FreeDspData(&pHw->MainPage, &pHw->pSwapPages, &pHw->SwapPageCount);
	    read_unlock(&pHw->DSPLock);
	    
	    /* Remove our network interface from the kernel: */
	    spin_lock(&pHw->NetLock);
	    if (pHw->pLinuxNet)
	    {
		/* This will call the adi_close method: */
		unregister_netdev(pHw->pLinuxNet);
		kfree(pHw->pLinuxNet);
		pHw->pLinuxNet = NULL;
	    }
	    spin_unlock(&pHw->NetLock);
    
	    /* Free memory we alloced: */
	    FREE_KBUFFER(pHw->pInterruptData);
	    FREE_KBUFFER(pHw->pReadyData);
	    for (i=0; i<INCOMING_Q_SIZE; i++)
	    {
		FREE_KBUFFER(pHw->pIncomingData[i]);
#ifndef USEBULK
		usb_free_urb(pHw->pUrbReadIso[i]);
#endif
	    }
	    DeAllocQueuedBuffers(&pHw->CtrlUrbFreeQ);
	    DeAllocQueuedBuffers(&pHw->CtrlUrbPendQ);
	    
	    /* Tell usb that we no longer claim these interfaces as our property: */
	    usb_driver_release_interface(&adi_driver, &(usb->actconfig->interface[0]));
	    usb_driver_release_interface(&adi_driver, &(usb->actconfig->interface[1]));
	    usb_driver_release_interface(&adi_driver, &(usb->actconfig->interface[2]));
    
	    /* Remove our /proc/adimodem interface: */
	    remove_proc_entry("driver/adimodem", NULL);
	    printk(KERN_INFO "ADI: ADSL device removed\n");
	}
	break;
    }

dis_done:
    ZAP(DEBUG_INIT, DEBUG_EXIT, ("---adi_disconnect\n"));
}

/*******************************************************************************/
/* adi_irq								       */
/*******************************************************************************/
static void adi_irq(struct urb *urb)
{
    Hardware *pHw;

    ZAP(DEBUG_INTS, DEBUG_ENTRY, ("+++adi_irq\n"));
    /*We put the modem pointer as the context*/
    pHw = (Hardware *)urb->context;
    if (NULL == pHw)
    {
	printk(KERN_ERR "ADI: NULL pointer to driver found in interrupt URB!\n");
	goto irq_done;
    }
    if (!pHw->HasIntURB)
    {
	printk(KERN_ERR "ADI: Callback on unregistered interrupt handler!\n");
	goto irq_done;
    }

    /*Lets check the status first, to make sure this was succesfull*/
    if (urb->status<0)
    {
        ZAP(DEBUG_INTS, DEBUG_ERRORS, (kTAB"failed adi_irq urb status = %#x\n", urb->status));
        /* These errors seem to be what we get when the device has been unplugged. */
	/* adi_disconnect will soon be called, meanwhile we must skip this interrupt: */
	urb->status = 0;
        goto irq_done;
    }

    /*All is well, we can process the data!*/
    if (pHw->pInterruptData->bmRequestType == 0x08) /* device-to-host interrupt */
    {
        DeviceInt *pInt;

        pInt = (DeviceInt *)(&(pHw->pInterruptData->Data[0])); 
        switch (pInt->Interrupt)
        {
        case kADI_INT_LOADSWAPPAGE:
	    {
                UInt16 SwapData = __cpu_to_le16(pInt->IntInfo.SwapIntInfo.SwapData);
                ZAP(DEBUG_INTS, DEBUG_INFO, (kTAB"interrupt = kADI_INT_LOADSWAPPAGE\n"));
                if (pHw->AdiModemSm.CurrentAdiState != STATE_HARD_RESET_INITIATE)
		{
		    read_lock(&pHw->DSPLock);
                    IdmaUploadSwapPage(pHw, SwapData);
		    read_unlock(&pHw->DSPLock);
		}
            }
            break;
        case kADI_INT_INCOMINGCMV:
            {
                ZAP(DEBUG_INTS, DEBUG_INFO, (kTAB"interrupt = kADI_INT_INCOMINGCMV\n"));
                ProcessIncomingCmv(pHw, pInt->IntInfo.CmvIntInfo.CmvData);
            }
        break;
        default:
            ZAP(DEBUG_INTS, DEBUG_INFO, (kTAB"interrupt = unsupported(%d)\n", pInt->Interrupt));
            break;
        }
    }

irq_done:
    ZAP(DEBUG_INTS, DEBUG_EXIT, ("---adi_irq\n"));
}

/*******************************************************************************/
/* adi_user								       */
/*									       */
/* This is actually the ioctl handler for ioctls coming from the usermode      */
/* helper application. The other ioctl routine, adi_ioctl, will receive        */
/* ioctls coming from network socket applications.			       */
/*******************************************************************************/
static int adi_user(struct usb_device *dev, unsigned int code, void *buf)
{
    Hardware *pHw = &modem;
    int retval = -ENOTTY;
    UInt32 pid = dev->descriptor.idProduct;
    /* USB automatically transfers user ioctl structure in kernel
    address space (the size is already encoded in the ioctl code): */
    struct adi_ioctl_info *pIOCTLinfo = (struct adi_ioctl_info *) buf;
    MOD_INC_USE_COUNT;
    /* Check this ioctl if for one of our devices: */
    switch (pid)
    {
    case kCASE_PREFIRMWARE:
	switch (code)
	{
	case ADIUSBADSLFIRMWARE:
	    {
		FIRMWARE_RECORD *pFirmware = NULL;
		retval = -ERESTARTSYS;
		if (down_interruptible(&pHw->FirmwareSem))
		    break;
		retval = -EFAULT;
		if (NULL == pIOCTLinfo)
		    goto end_load_firmware;
		pFirmware = (FIRMWARE_RECORD *) GET_VBUFFER(pIOCTLinfo->BufferSize);
		if (NULL == pFirmware)
		{
		    retval = -ENOMEM;
		    goto end_load_firmware;
		}
		if (copy_from_user(pFirmware, pIOCTLinfo->pBuffer, pIOCTLinfo->BufferSize) != 0)
		    goto end_free_firmware;
		printk(KERN_INFO "ADI: ioctl ADIUSBADSLFIRMWARE received\n");
		/* Load the firmware: */
		printk(KERN_INFO "ADI: Loading firmware to modem...\n");
		if (loadUsbFirmware(pHw, pFirmware, pIOCTLinfo->BufferSize) < 0)
		{
		    printk(KERN_ERR "ADI: Unable to load firmware to modem!\n");
		    retval = -EIO;
		}
		else
		{
		    printk(KERN_INFO "ADI: Firmware successfully loaded to modem\n");
		    retval = 0;
		}
	    end_free_firmware:
		FREE_VBUFFER(pFirmware);
	    end_load_firmware:
		up(&pHw->FirmwareSem);
	    }
	}
	break;
    case kCASE_POSTFIRMWARE:
	switch (code)
	{
	case ADIUSBADSLOPTIONS:		/* Set driver options */
	    /* Option cannot be set while network interface is up */
	    retval = -ERESTARTSYS;
	    if (down_interruptible(&pHw->NetSem))
		break;
	    retval = -EBUSY;
	    if (!pHw->IsOpen)
	    {
		DriverOptions aOptions;
		memcpy(&aOptions, &DefaultOptions, sizeof(DriverOptions));
		retval = -EFAULT;
		if (NULL == pIOCTLinfo) 
		    goto end_set_options;
		/* Check the option array size: */
		retval = -EINVAL;
		if (sizeof(aOptions) != pIOCTLinfo->BufferSize)
		    goto end_set_options;
		if (copy_from_user(&aOptions, pIOCTLinfo->pBuffer, sizeof(aOptions)))
		{
		    retval = -EFAULT;
		    goto end_set_options;
		}
		/* Options retrieved, check them: */
		if (CheckOptions(aOptions))
		{
		    /* OK, initialize Mpoa and Msg: */
		    printk(KERN_INFO "ADI: ioctl ADIUSBADSLOPTIONS received\n");
		    MsgInitialize(pHw, aOptions);
		    MpoaInitialize(pHw, aOptions);
		    retval = 0;
		}
	    }
	end_set_options:
	    up(&pHw->NetSem);
	    break;
	case ADIUSBADSLDSP:		/* Load DSP code */
	    /* DSP cannot be loaded while network interface is up */
	    retval = -ERESTARTSYS;
	    if (down_interruptible(&pHw->NetSem))
		break;
	    retval = -EBUSY;
	    if (!pHw->IsOpen)
	    {
		UInt8 *pBuf = NULL;
		IDMAPage MainPage;
		IDMAPage *pSwapPages;
		UInt32   SwapPageCount;
		long flags;
		retval = -EFAULT;
		if (NULL == pIOCTLinfo)
		    goto end_load_adsl_dsp;
		if (pIOCTLinfo->IDMAStart >= pIOCTLinfo->BufferSize)
		    goto end_load_adsl_dsp;
		retval = -ENOMEM;
		pBuf = GET_VBUFFER(pIOCTLinfo->BufferSize);
		if (NULL == pBuf)
		    goto end_load_adsl_dsp;
		retval = -EFAULT;
		if (copy_from_user(pBuf, pIOCTLinfo->pBuffer, pIOCTLinfo->BufferSize))
		    goto free_adsl_dsp;
		/* Get our DSP code ready for IDMA booting */
		printk(KERN_INFO "ADI: ioctl ADIUSBADSLDSP received\n");
		retval = ProcessDSPCode(pHw, pBuf, pIOCTLinfo->IDMAStart, pIOCTLinfo->BufferSize,
		    &MainPage, &pSwapPages, &SwapPageCount);
		if (retval != 0)
		{
		    FreeDspData(&MainPage, &pSwapPages, &SwapPageCount);
		    goto free_adsl_dsp;
		}
		/* Ok, we got our DSP code, replace the old DSP code by the new one: */
		write_lock_irqsave(&pHw->DSPLock, flags);
		FreeDspData(&pHw->MainPage, &pHw->pSwapPages, &pHw->SwapPageCount);
		pHw->MainPage.BlockCount = MainPage.BlockCount;
		pHw->MainPage.Blocks = MainPage.Blocks;
		pHw->pSwapPages = pSwapPages;
		pHw->SwapPageCount = SwapPageCount;
		write_unlock_irqrestore(&pHw->DSPLock, flags);
		/* Load the DSP code if necessary: */
		printk(KERN_INFO "ADI: Loading DSP code to device...\n");
		read_lock(&pHw->DSPLock);
		retval = -EIO;
		if (!pHw->HasIntURB)
		{
		    /* Get the USB endpoint for IDMA uploading: */
		    struct usb_endpoint_descriptor *epint =
			dev->actconfig->interface[0].altsetting[0].endpoint + 0;
		    /* Install the interrupt handler to send IDMA pages to the modem
		    and handle further incoming data: */
		    usb_fill_int_urb(&pHw->urbInt, dev, pHw->pipeIntIn, pHw->pInterruptData,
				 sizeof(CDC_NOTIFY), adi_irq, pHw, epint->bInterval);
		    pHw->HasIntURB = TRUE;
		    if (usb_submit_urb(&pHw->urbInt) < 0)
		    {
			printk(KERN_ERR "ADI: Unable to submit interrupt URB!\n");
			pHw->HasIntURB = FALSE;
			goto unlock_adsl_dsp;
		    }
		}
		/* Send the DSP code to the modem: */
		if (!pHw->HasIntURB || BootTheModem(pHw, FALSE) < 0)
		{
		    printk(KERN_INFO "ADI: Unable to load DSP code to device!\n");
		    goto unlock_adsl_dsp;
		}
		else
		    printk(KERN_INFO "ADI: DSP code successfully loaded to device\n");
		retval = 0;
	    unlock_adsl_dsp:
		read_unlock(&pHw->DSPLock);
	    free_adsl_dsp:
		FREE_VBUFFER(pBuf);
	    }
	end_load_adsl_dsp:
	    up(&pHw->NetSem);
	    break;
	case ADIUSBADSLGETITF:		/* Get the network interface name */
	    {
		uint32_t dwNameLen;
		unsigned long flags;
		retval = -EFAULT;
		if (NULL == pIOCTLinfo)
		    break;
		/* Create the Ethernet interface if necessary: */
		retval = -ERESTARTSYS;
		if (down_interruptible(&pHw->NetSem))
		    break;
		spin_lock_irqsave(&pHw->NetLock, flags);
		if (!pHw->pLinuxNet)
		{
		    retval = create_etherdev();
		    if (retval < 0)
			goto end_get_itf;
		}
		dwNameLen = strlen(pHw->pLinuxNet->name) + sizeof(char);
		if (NULL == pIOCTLinfo->pBuffer)
		{
		    /* Get the requested size */
		    pIOCTLinfo->BufferSize = dwNameLen;
		    retval = 0;
		}
		else
		{
		    /* Get the interface name */
		    if (pIOCTLinfo->BufferSize < dwNameLen)
		    {
			retval = -EINVAL;
			goto end_get_itf;
		    }
		    if (copy_to_user(pIOCTLinfo->pBuffer, pHw->pLinuxNet->name, dwNameLen) != 0)
		    {
			retval = -EFAULT;
			goto end_get_itf;
		    }
		    retval = 0;
		}
	    end_get_itf:
		spin_unlock_irqrestore(&pHw->NetLock, flags);
		up(&pHw->NetSem);
		break;
	    }
	    break;
	case ADIUSBADSLSYNC:		/* Wait for modem "operational" state */
	    if (wait_event_interruptible(pHw->OperationalQueue,
		((pHw->AdiModemSm.CurrentAdiState & STATE_OPERATIONAL) == STATE_OPERATIONAL)))
		retval = -EINTR;
	    else
		retval = 0;
	    break;
	default:
	    /* Bad ioctl */
	    break;
	}
	break;
    default:
	/* Bad device */
	break;
    }
    MOD_DEC_USE_COUNT;
    return retval;
}

/*******************************************************************************/
/* adi_read_proc							       */
/*******************************************************************************/
static int adi_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
    Hardware *pHw = (Hardware *)data;
    int size;
    char *p = page;
    /*If this output needs to be larger than 4K (PAGE_SIZE), we need to do this differently*/
    p += sprintf(p, "\nAnalog Devices USB ADSL Modem Status Display\n");
    p += sprintf(p, "-------------------------------------------------------------\n");
    p += sprintf(p, "Tx Rate  %#8.8x  Rx Rate  %#8.8x  Crc      %#8.8x\n", 
		pHw->AdiModemSm.UpRate / 1024, 
		pHw->AdiModemSm.DownRate / 1024, 
		pHw->Statistics[STAT_CELLS_LOST_CRC]);
    p += sprintf(p, "FEC      %#8.8x  Margin   %#8.8x  Atten    %#8.8x\n",
		pHw->AdiModemSm.stats_Uncorr_Blks,
		pHw->AdiModemSm.stats_Cur_SNR & 0xFF,
		pHw->AdiModemSm.stats_Cur_Atten & 0xFF);
    p += sprintf(p, "VID-CPE  %#8.8x  VID-CO   %#8.8x  HEC      %#8.8x\n",
		pHw->AdiModemSm.INFO14,
		pHw->AdiModemSm.INFO08,
		pHw->AdiModemSm.DIAG03);
    p += sprintf(p, "VPI      %#8.8x  VCI      %#8.8x  Delin    ", 
		pHw->Vc.Vpi, pHw->Vc.Vci);
    /*Delineation is the only one where we print a string instead of a number*/
    if (pHw->AdiModemSm.flags & 0x0C00)
	p += sprintf(p, "ERROR\n");
    else
	if (pHw->AdiModemSm.flags & 0x0030) 
	    p += sprintf(p, "LOSS\n");
	else
	    p += sprintf(p, "GOOD\n");
    p += sprintf(p, "Cells Rx %#8.8x  Cells Tx %#8.8x\n",
		pHw->Statistics[STAT_CELLS_RX],
		pHw->Statistics[STAT_CELLS_TX]);
    p += sprintf(p, "Pkts Rx  %#8.8x  Pkts Tx  %#8.8x\n",
		pHw->Statistics[STAT_PAKTS_RX],
		pHw->Statistics[STAT_PAKTS_TX]);
    switch (pHw->AdiModemSm.CurrentAdiState)
    {
    case STATE_UNDEFINED:
	p += sprintf(p, "Modem is unplugged from USB.\n");
	break;
    case STATE_JUST_PLUGGED_IN:
	p += sprintf(p, "Modem waiting for driver response.\n");
	break;
    case STATE_UNTRAIN:
    case STATE_UNTRAIN_TX:
    case STATE_UNTRAIN_RX:
    case STATE_INITIALIZING:
    case STATE_INITIALIZING_TX:
    case STATE_INITIALIZING_RX:
	p += sprintf(p, "Modem is initializing\n");
	break;
    case STATE_HARD_RESET_INITIATE:
    case STATE_HARD_RESET_END:
    case STATE_BOOT_WAIT:
	p += sprintf(p, "Modem is booting\n");
	break;
    case STATE_OPERATIONAL:
    case STATE_OPERATIONAL_TX:
    case STATE_OPERATIONAL_RX:
	p += sprintf(p, "Modem is operational\n");
	break;
    case STATE_STALLED_FOREVER:
	p += sprintf(p, "Modem cannot boot. Unplug other devices and retry.\n");
	break;
#ifdef ZAPS
    default:
	p += sprintf(p, "Unhandled state\n");
	Zap(("Timer: unhandled state 0x%X\n", pHw->AdiModemSm.CurrentAdiState));
	break;
#endif
    }
    p += sprintf(p, "\n");
    /*Figure out how much of the page we used*/
    size  = p - page;
    size -= off;
    
    if (size < count)
    {
	*eof = 1; 
	if (size <= 0)
	    return 0;
    }
    else
	size = count;
    /*Fool caller into thinking we started where he told us to in the page*/
    *start = page + off;
    return size;
}


/*******************************************************************************/
/* Code that's part of the ethernet driver interface			       */
/*******************************************************************************/

/*******************************************************************************/
/* create_etherdev							       */
/*******************************************************************************/
static int create_etherdev(void)
{
    Hardware *pHw = &modem;
    struct net_device *ether = init_etherdev(NULL, 0);
    if (!ether)
    {
	return -ENOMEM;
    }    
    /* Initialize the Ethernet device: */
    ether->priv               = pHw;
    ether->open               = adi_open;
    ether->stop               = adi_close;
    ether->do_ioctl           = adi_ioctl;
    ether->get_stats          = adi_stats;
    ether->hard_start_xmit    = adi_start_xmit;
    ether->set_multicast_list = adi_set_multicast;
    ether->tx_timeout         = adi_tx_timeout;
    ether->watchdog_timeo     = kTRANSMIT_TIMEOUT;
    GetMacAddress(pHw);
    memcpy(ether->dev_addr, pHw->MAC, ETH_LENGTH_OF_ADDRESS);
    SET_MODULE_OWNER(ether);
    /*Link to be able to get back to the net_device from the Hardware*/
    pHw->pLinuxNet = ether;
    printk(KERN_INFO "ADI: Modem ethernet interface is '%s' (mtu %d)\n",
	ether->name, ether->mtu);
    return 0;
}


/*******************************************************************************/
/* adi_open								       */
/*******************************************************************************/
static int adi_open(struct net_device *ether)
{
    Hardware *pHw = (Hardware *)ether->priv;

    MOD_INC_USE_COUNT;
    ZAP(DEBUG_ENET, DEBUG_ENTRY, ("+++adi_open\n"));

    /*Tell ourselves that we're open for business!*/
    if (down_interruptible(&pHw->NetSem))
	return -ERESTARTSYS;
    pHw->IsOpen = TRUE;
    up(&pHw->NetSem);
    
    /* Create incoming data URBs: */
    StartReadPipe(pHw);

    /*Tell the kernel we're ready for network traffic flow*/
    pHw->IsWriting = FALSE;
    netif_start_queue(ether);

    ZAP(DEBUG_ENET, DEBUG_EXIT, ("---adi_open\n"));
    return 0;
}

/*******************************************************************************/
/* adi_close								       */
/*******************************************************************************/
static int adi_close(struct net_device *ether)
{
    Hardware *pHw = (Hardware *)ether->priv;

    ZAP(DEBUG_ENET, DEBUG_ENTRY, ("+++adi_close\n"));
    if (down_interruptible(&pHw->NetSem))
	return -ERESTARTSYS;
    /*Tell the kernel that we do not want any more network traffic*/
    netif_stop_queue(ether);

    /* Stop our outgoing data urbs */
    if (pHw->IsWriting)
    {
	pHw->urbWrite.transfer_flags &= ~USB_ASYNC_UNLINK;
	usb_unlink_urb(&pHw->urbWrite);
    }
    
    /*Stop our incoming data urbs*/
    StopReadPipe(pHw);

    /*Tell ourselves that we're closed for business!*/
    pHw->IsOpen = FALSE;
    up(&pHw->NetSem);

    ZAP(DEBUG_ENET, DEBUG_EXIT, ("---adi_close\n"));
    MOD_DEC_USE_COUNT;
    return 0;
}

/*******************************************************************************/
/* adi_ioctl 								       */
/*******************************************************************************/
static int adi_ioctl(struct net_device *ether, struct ifreq *rq, int cmd)
{
    ZAP(DEBUG_ENET, DEBUG_ENTRY, ("+++adi_ioctl\n"));

    ZAP(DEBUG_ENET, DEBUG_EXIT, ("---adi_ioctl\n"));
    /*This particular ioctl is accessed through the socket API*/
    /*For now, I don't know of any ioctls that we need to support*/
    return -ENOTTY;
}

/*******************************************************************************/
/* adi_stats 								       */
/*******************************************************************************/
static struct net_device_stats *adi_stats(struct net_device *ether)
{
    Hardware *pHw = (Hardware *)ether->priv;

    ZAP(DEBUG_ENET, DEBUG_ENTRY, ("+++adi_stats\n"));

    ZAP(DEBUG_ENET, DEBUG_EXIT, ("---adi_stats\n"));
    return &pHw->LinuxStats;
}

/*******************************************************************************/
/* adi_start_xmit 							       */
/*******************************************************************************/
static int adi_start_xmit(struct sk_buff *skb, struct net_device *ether)
{
    Hardware *pHw = (Hardware *)ether->priv;
    int result=0, packet_sent=0;

    ZAP(DEBUG_ENET, DEBUG_ENTRY, ("+++adi_start_xmit\n"));

    /*Tell the kernel to hold off until we're done sending this packet*/
    netif_stop_queue(ether);

    /*Send the packet through the SAR to get ATM cells*/
    switch (pHw->MpoaMode)
    {
    case MPOA_MODE_BRIDGED_ETH_LLC :
    case MPOA_MODE_BRIDGED_ETH_VC  :
        result = UniProcessOutboundPdu(pHw, skb->data, skb->len);
        break;
    case MPOA_MODE_ROUTED_IP_LLC   :
    case MPOA_MODE_ROUTED_IP_VC    :
	/* In routed mode, ARP request will be replied locally: */
	if(*((short *)(skb->data+12))  == __constant_htons(ETH_P_ARP))
	{ 
	    result = LocalArpReply(pHw, skb->data,skb->len);
	    if (!result)
	    {
		++pHw->LinuxStats.tx_packets;
		pHw->LinuxStats.tx_bytes += skb->len;
		netif_wake_queue(ether);
		packet_sent = 1;
	    }
	}
	else
	    result = UniProcessOutboundPdu(pHw, skb->data+14, skb->len-14);
	break;
    case MPOA_MODE_PPPOA_LLC       :
    case MPOA_MODE_PPPOA_VC        :
        result = UniProcessOutboundPdu(pHw, skb->data+16, skb->len-16);
        break;
    }
    /* If all is OK, send the data: */
    if (!result && !packet_sent)
    {
	/* Fix the timestamp of transaction for timeout detection: */
	ether->trans_start = jiffies;
	/* Store packet information for transmit statistics: */
	pHw->OutgoingPacketSize = skb->len;
	/* Uni says all is ok to proceed with sending the data: */
	result = WriteAtmData(pHw);
    }
    /* We've copied the data into our own internal buffers, so we can */ 
    /* go ahead and free the kernel's buffer for this packet. */
    dev_kfree_skb(skb);
    if (0 != result)
    {
	/*******************************************************************************/
	/* hard_start_xmit must never return 1 in normal condition. Returning 1 means  */
	/* there is an unrecoverable hardware error and that the adapter cannot be used*/
	/* anymore. In this case, we must not destroy the socket buffer neither resume */
	/* packet transmission. I really don't know who will wake up the softnet layer */
	/* after, and indeed it seems the network interface is definitively dead. As we*/
	/* do not and cannot handle hardware errors at this level, all we can do is to */
	/* consider the socket buffer has been lost on the wire and resume	       */
	/* transmissions.							       */
	/*******************************************************************************/
	
	/* Update statistics: */
	printk(KERN_WARNING "ADI: Packet transmission error: result=0x%x\n", result);
        ++pHw->LinuxStats.tx_errors;
	++pHw->LinuxStats.tx_carrier_errors;
	
	/* There will be no URB completion, so we must wake up the network layer ourself: */
	pHw->IsWriting = FALSE;
	netif_wake_queue(ether);
    }
    else
    {
	/*******************************************************************************/
	/* You'll notice that, in this case, we do not wake up the netif queue	       */
	/* .... that's because we're only using one write URB, so we have	       */
	/* to wait until the completion handler for the write is called		       */
	/* or a transmit timeout occur before we can tell kernel that we can handle    */
	/* more network data.							       */
	/* A potential performance improvement would be to have an internal	       */
	/* URB queue for outgoing data, so we could (theoretically) start	       */
	/* the software-processing of a packet while another one is being sent.	       */
	/*******************************************************************************/
    }
    ZAP(DEBUG_ENET, DEBUG_EXIT, ("---adi_start_xmit\n"));
    return 0;
}

/*******************************************************************************/
/* adi_set_multicast							       */
/*******************************************************************************/
static void adi_set_multicast(struct net_device *ether)
{
    ZAP(DEBUG_ENET, DEBUG_ENTRY, ("+++adi_set_multicast\n"));

    /*******************************************************************************/
    /* Our ethernet side is really all software, so anything that		   */
    /* a normal ethernet card would do in hardware, we do here in the driver.	   */
    /* On a normal ethernet card, you would upload the filter address list	   */
    /* to the card - but we just loop through a fast software compare. 		   */
    /* Since we will just always check whats in the net_device *, we don't 	   */
    /* have to do anything here!						   */
    /*******************************************************************************/

    ZAP(DEBUG_ENET, DEBUG_EXIT, ("---adi_set_multicast\n"));
}


/*******************************************************************************/
/* adi_tx_timeout							       */
/*******************************************************************************/
static void adi_tx_timeout(struct net_device *ether)
{
    Hardware *pHw = (Hardware *)ether->priv;
    printk(KERN_WARNING "ADI: Transmit timed out!\n");
    
    /*
    Other linux drivers don't seem to recover and resend
    the socket buffer Linux gives them. However, O'Reilly's
    Linux Device Driver says we must not lose the buffer.
    But, as packets may anyway be lost by media in general,
    it's quite certain the protocol will resend it sooner or
    later, or consider packet lost as normal. Therefore, this
    driver do not attempt anything more than removing the broken
    URB. This will call WriteCompletion, which will call
    netif_wakequeue:
    */
    pHw->urbWrite.transfer_flags |= USB_ASYNC_UNLINK;
    usb_unlink_urb(&pHw->urbWrite);
    
    /* We must reset the transaction time to keep the watchdog quiet: */
    ether->trans_start = jiffies;

    ZAP(DEBUG_ENET, DEBUG_EXIT, ("---adi_tx_timeout\n"));
}


/**************************************************************************************
$Log: AdiUsbAdslDriver.c,v $
Revision 1.30  2002/05/24 17:56:44  Anoosh Naderi
Clean up the code.
Added support for PPPoA (LLC & VC).

Added ISO support for incoming data.

Revision 1.27  2002/02/12 21:50:08  steve.urquhart
Added statement to /proc output for modem status.

Revision 1.26  2002/02/11 21:12:35  chris.edgington
Added support for possible multiple read urbs.

Revision 1.25  2002/01/23 17:23:02  chris.edgington
Moved code that claims usb interfaces to be after kernel memory
allocations, so we don't claim the interfaces until we're sure we
will use them.
Added verification of pointer validity for all memory allocs.
Added comments to adi_user.

Revision 1.24  2002/01/18 18:00:09  chris.edgington
Added include of Options.h.
Removed temporary MODULE_PARMs section, since we're using .conf file now.
Moved adi_user to bottom of file, cleaned it up some.
Write ioctl handler for options data transfer.
Moved calls to MsgInitialize and MpoaInitialize to options ioctl.

Revision 1.23  2002/01/18 00:01:20  steve.urquhart
Upgraded adi_user to handle 2 ioctls

Revision 1.22  2002/01/17 17:49:18  steve.urquhart
Fixed GPL Header, keyword subsititution

Revision 1.21  2002/01/16 22:59:33  steve.urquhart
Moved portion of DspFoodProcessor and invocation of BootTheModem to
adi_user ioctl handler.  Removed LoadAndPrepareDsp.  Cleaned up the ioctl
interface.  Decreased initial DSP vmalloc to less than .5MB.

Revision 1.20  2002/01/15 22:15:59  steve.urquhart
Second revision of user mode code, includes example of means necessary to
ioctl the driver.

Revision 1.19  2002/01/14 21:59:30  chris.edgington
Added GPL header.

Revision 1.18  2002/01/14 17:38:41  chris.edgington
Changed include order.

Revision 1.17  2002/01/10 17:14:15  chris.edgington
Added /proc/adimodem interface.

Revision 1.16  2002/01/09 22:15:56  chris.edgington
Added support for proper kernel symbol name importing through MODVERSIONS.

Revision 1.15  2002/01/09 20:16:57  chris.edgington
Added transmit timeout support via adi_tx_timeout.

Revision 1.14  2002/01/08 20:03:37  chris.edgington
Removed a few unused local vars.
Removed hardcoded MAC address setting code.

Revision 1.13  2002/01/08 16:22:25  chris.edgington
Fixed parameters to memcpy of MAC address in adi_probe.

Revision 1.12  2002/01/08 16:06:28  chris.edgington
Added some necessary comments.
Removed invalid DeAllocQueuedBuffers calls in adi_disconnect.
Finished details of adi_set_multicast.

Revision 1.11  2002/01/04 20:47:47  chris.edgington
Added initialization of data send/receive buffers in Hardware.
Removed initialization of send/receive queues - not needed for Linux.
Moved cancellation of timers in adi_disconnect to be right at the beginning.
Added calls to StartReadPipe and StopReadPipe in adi_open and adi_close.
Added transmit logic to adi_start_xmit (call WriteAtmData).

Revision 1.10  2002/01/02 21:59:00  chris.edgington
Added initialization of IdleOutQ, LiveOutQ, IdleWaitQ, LiveWaitQ.
Added calls to CrcGenerateTable and MpoaInitialize.

Revision 1.9  2001/12/28 22:08:26  chris.edgington
Added creation and initialization of net_device interface.
Added skeletons for adi_open, adi_close, adi_ioctl, adi_start_xmit,
      adi_stats, and adi_set_multicast.

Revision 1.8  2001/12/28 19:28:03  chris.edgington
Added initialization of CtrlUrbFreeQ and CtrlUrbPendQ.
Added initialization of CtrlUrbQueueTimer.
Make adi_disconnect only attemp free when pid = post firmware pid.
Only upload swap page when state != HARD_RESET_INITIATE.

Revision 1.7  2001/12/26 22:43:33  chris.edgington
Added initialization code for kernel timer.
Changed parameters to BootTheModem.

Revision 1.6  2001/12/22 21:43:23  chris.edgington
Added call to MsgInitialize.
Added call to ProcessIncomingCmv on CMV interrupt.

Revision 1.5  2001/12/22 19:53:00  chris.edgington
Moved utility routines to Util.c - loadUsbFirmware, EZUSBWrite.
Added loading of swap pages in response to swap page request interrupt.

Revision 1.4  2001/12/21 23:56:51  chris.edgington
Added interrupt handler and startup of interrupt pipe.
Added call to BootTheModem.

Revision 1.3  2001/12/20 22:39:25  chris.edgington
Removed call to usb_set_interface - I don't think we need it.
Added calls to usb_driver_claim_interface, to keep usb from calling adi_probe
for each interface exposed by the device ('cause we use them all and we don't
want another driver trying to claim an interface.)
Added call to LoadAndPrepareDspCode
Added calls to usb_driver_release_interface in adi_disconnect.

Revision 1.2  2001/12/17 22:21:37  chris.edgington
Wrote core driver init code, including USB firmware upload.
Use "insmod -f adiusbadsl.o" to get this version to load.

Revision 1.1.1.1  2001/12/14 21:27:07  chris.edgington
Initial import
**************************************************************************************/
