/**********************************************************************************/
/* $Id: Util.c,v 2.00 2002/05/24 21:59:30 Anoosh Naderi Exp $           	  */
/*										  */
/* Copyright (c) 2002, Analog Devices Inc., All Rights Reserved			  */
/*										  */
/* Util.c									  */
/*										  */
/* Miscellaneous utility routines used in the driver				  */
/*										  */
/* This file is part of the "ADI USB ADSL Driver for Linux".			  */
/*										  */
/* "ADI USB ADSL Driver for Linux" is free software; you can redistribute it      */
/* and/or modify it under the terms of the GNU General Public License as          */
/* published by the Free Software Foundation; either version 2 of the License,    */
/* or (at your option) any later version.					  */
/*										  */
/* "ADI USB ADSL Driver for Linux" is distributed in the hope that it will be     */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty of	  */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the		  */
/* GNU General Public License for more details.					  */
/*										  */
/* You should have received a copy of the GNU General Public License		  */
/* along with "ADI USB ADSL Driver for Linux"; if not, write to the Free Software */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA      */
/**********************************************************************************/

#include "Adiutil.h"
#include "Util.h"
#include "Konst.h"

/**********************************************************************************/
/* Local utilities								  */
/**********************************************************************************/

/**********************************************************************************/
/* htoi										  */
/*										  */
/* Converts a hex char to an integer (my own wacky implementation).		  */
/* I don't know why this isn't part of a standard library or something,		  */
/* it seems like this is a quite-frequently-rewritten function.			  */
/**********************************************************************************/
static UInt8 htoi(UInt8 h)
{
    switch (toupper(h))
    {
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
	return h-'0';
	break;
    case 'A':
    case 'B':
    case 'C':
    case 'D':
    case 'E':
    case 'F':
	return h-'A'+10;
	break;
    case 'a':
    case 'b':
    case 'c':
    case 'd':
    case 'e':
    case 'f':
	return h-'a'+10;
	break;
    }
    
    /*If we get here ... the character was invalid*/
    return -1;
}


/**********************************************************************************/
/* EZUSBWrite									  */
/**********************************************************************************/
static int EZUSBWrite(Hardware *pHw, UInt16 EZUSBAddress, UInt16 count, UInt8 writeBuffer[])
{
    int ret = -ENOMEM;
    UInt8 *xfer_buf = GET_KBUFFER(count);
    if (NULL != xfer_buf)
    {
	memcpy(xfer_buf, writeBuffer, count);
	ret = usb_control_msg(pHw->usbdev,
			      usb_sndctrlpipe( pHw->usbdev, 0 ),
			      kEZUSB_LOAD_INTERNAL,
			      USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			      EZUSBAddress, 0, xfer_buf, count, 300);
	FREE_KBUFFER(xfer_buf);
    }
    return ret;
}


/**********************************************************************************/
/* Local prototypes								  */
/**********************************************************************************/

static void QueueControlUrb(Hardware *pHw, struct urb *urb);
static void CtrlUrbCompletion(struct urb *urb);


/**********************************************************************************/
/* Helper implementations							  */
/**********************************************************************************/

/**********************************************************************************/
/* GetMacAddress								  */
/*										  */
/* Queries the ethernet (MAC) address from the USB firmware, storing it in	  */
/* pHw->MAC in numeric form (as opposed to ascii).				  */
/**********************************************************************************/
void GetMacAddress(Hardware *pHw)
{
    int maclen, i;
    UInt8 macbuf[ETH_LENGTH_OF_ADDRESS*2+1];
    
    /****************************************************************************/
    /* Even though this device advertises itself as being an ATM class          */
    /* device, it doesn't have ATM functional descriptors - it has Ethernet ones*/
    /* We already know what the string index (iMACAddress) is from the		*/
    /* Ethernet descriptor, so all we have to do is query the string		*/
    /****************************************************************************/
    maclen = usb_string(pHw->usbdev, kMAC_ADDRESS_INDEX, macbuf, ETH_LENGTH_OF_ADDRESS*2+1);
    
    /*Make sure we got a string of the right length*/
    if (maclen != ETH_LENGTH_OF_ADDRESS*2)
    {
	ZAP(DEBUG_UTILS, DEBUG_ERRORS, ("GetMacAddress: invalid length %x.\n", maclen));
	return;
    }
    
    /*Convert the string into a numeric format    */
    /*(each string byte is a nibble in the result)*/
    for (i=0; i<ETH_LENGTH_OF_ADDRESS; i++)
    {
	pHw->MAC[i] = (htoi(macbuf[i*2])<<4) + (htoi(macbuf[i*2+1]));
    }
}


/**********************************************************************************/
/* loadUsbFirmware								  */
/*										  */
/* Called by ioctl from user space to upload the ADI-specific firmware            */
/* to the EZUSB chip.                                                             */
/**********************************************************************************/
int loadUsbFirmware(Hardware *pHw, FIRMWARE_RECORD *aRecords, UInt32 BufSize)
{
    /*Assert reset*/
    UInt8 writeVal = 1;
    int ret = EZUSBWrite(pHw, k8051_USBCS, 1, &writeVal);
    if (ret >= 0)
    {
	int ret2;
	/*Download code*/
	UInt32 i=0;
	while (1)
	{
	    ret = EZUSBWrite(pHw, ntohs(aRecords[i].Address),
			     (aRecords[i].Length < MAX_FIRMWARE_RECORD_LENGTH ?
			     aRecords[i].Length : MAX_FIRMWARE_RECORD_LENGTH),
			     aRecords[i].Data);
	    /*Check for write error*/
	    if (ret < 0)
	    {
		printk(KERN_ERR"ADI: Error while writing firmware to device!\n");
		break;
	    }
	    /*Check for end of firmware*/
	    if (aRecords[i].Type != 0)
		break;
	    /*Check for end of buffer*/
	    if (i*sizeof(FIRMWARE_RECORD) >= BufSize)
	    {
		printk(KERN_ERR"ADI: Incomplete firmware received, some parts already downloaded to device!\n");
		ret = -1;
		break;
	    }
	    /*Go to next record*/
	    ++i;
	}
	/*De-assert reset*/
	writeVal = 0;
	ret2 = EZUSBWrite(pHw, k8051_USBCS, 1, &writeVal);
	if (ret2 < 0)
	{
	    printk(KERN_ERR"ADI: EZUSBWrite de-reset failure\n");
	    ret = ret2;
	}
    }
    else
    {
	printk(KERN_ERR"ADI: EZUSBWrite reset failure with error %x\n", ret);
    }
    return ret;
}


/**********************************************************************************/
/* AllocQueuedBuffers								  */
/*										  */
/* One-time initialization of buffers that will go into queues. All queued	  */
/* buffers have the same structure at their beginning, so this allows us to	  */
/* use common routines for buffer allocation, freeing, and manipulation.	  */
/* If SetQueueEntry == TRUE, the routine also associates this buffer with 	  */
/* that Pb by setting the Pb's pQueueEntry to the buffer pointer. 		  */
/* Assumes that the Q is "fresh" - so it initializes it first.			  */
/**********************************************************************************/
int AllocQueuedBuffers(Hardware *pHw, 
                       UInt32 BufCount, 
                       UInt32 BufSize, 
                       UInt32 DataSize, 
                       QHdr  *pQ, 
                       Boolean SetQueueEntry)
{
    int stat = USB_ST_NOERROR;
    GENERIC_BUFFER *pBuf;
    UInt8 *pData = NULL;
    UInt32 i;

    ZAP(DEBUG_UTILS, DEBUG_ENTRY,
        ("+++AllocQueuedBuffers: BufCount=0x%X, BufSize=0x%X, "
	"DataSize=0x%x, pQ=%p, SetQueueEntry=0x%X\n",
        BufCount, BufSize, DataSize, pQ, SetQueueEntry));

    /*Initialize the queue to a known state before adding members to it*/
    INIT_LIST_HEAD(pQ);

    for (i=0; i<BufCount; i++)
    {
        /*First allocate space for the buffer structure itself.*/ 
        pBuf = (GENERIC_BUFFER *)GET_VBUFFER(BufSize);
        if (NULL == pBuf)
        {
	    ZAP(DEBUG_UTILS, DEBUG_ERRORS, (kTAB"Not enough memory for buffers!\n"));
	    stat = -ENOMEM;
	    goto Alloc_exit;
        }

        /*Clear out the buffer*/
        memset(pBuf, 0, BufSize);

        /*Add this buffer to the queue*/
        list_add(&pBuf->List, pQ);

        if (DataSize)
        {
	    /*Now allocate the actual data storage part of the buffer, if the buffer needs it*/
	    pData = GET_KBUFFER(DataSize);
	    
	    if (NULL == pData)
	    {
		ZAP(DEBUG_UTILS, DEBUG_ERRORS, (kTAB"Not enough memory for data in buffers!\n"));
		stat = -ENOMEM;
		goto Alloc_exit;
	    }
        }

        pBuf->AllocedSize = DataSize;
        pBuf->pData       = pData;
    }

Alloc_exit:
    ZAP(DEBUG_UTILS, DEBUG_ENTRY, ("---AllocQueuedBuffers\n"));
    return stat;
}

/**********************************************************************************/
/* DeAllocQueuedBuffers								  */
/*										  */
/* Frees all buffers in the given queue						  */
/**********************************************************************************/
void DeAllocQueuedBuffers(QHdr *pQ)
{
    GENERIC_BUFFER *pBuf;

    pBuf = (GENERIC_BUFFER *)GetBufferFromQ(pQ);

    while (pBuf != NULL)
    {
	/*If we allocated data storage for this buffer, free that first*/
	if (pBuf->AllocedSize != 0)
	    FREE_KBUFFER(pBuf->pData);
	
	/*Now free the memory we allocated for this buffer*/
	FREE_VBUFFER(pBuf);
	/*And get the next one ...*/
	pBuf = (GENERIC_BUFFER *)GetBufferFromQ(pQ);
    }
}

/**********************************************************************************/
/* GetBufferFromQ								  */
/**********************************************************************************/
QHdr *GetBufferFromQ(QHdr *pQ)
{
    QHdr *pBuf;

    /*Check first for an empty list*/
    if (list_empty(pQ))
	return NULL;
    
    /*List isn't empty, so extract an element and return it*/
    pBuf = pQ->next;
    list_del(pQ->next);
    
    return pBuf;
}

/**********************************************************************************/
/* PutBufferInQ									  */
/**********************************************************************************/
void PutBufferInQ(GENERIC_BUFFER *pBuf, QHdr *pQ)
{
    /*Add this buffer to the queue*/
    list_add(&pBuf->List, pQ);
}


/**********************************************************************************/
/* WriteDspMsg									  */
/**********************************************************************************/
void WriteDspMsg(Hardware *pHw, UInt16 Addr, UInt16 *pData)
{
    /*get space for data and request struct*/
    struct urb *urb;
    devrequest *pdr;
    UInt8 *xfer_buf = GET_KBUFFER((CMV_DATA_WORDS*2)+sizeof(devrequest));

    if (!xfer_buf)
    {
	ZAP(DEBUG_BOOT, DEBUG_ERRORS,
	    (kTAB"kmalloc in WriteDspMsg for %#X bytes failed\n",
	    CMV_DATA_WORDS*2+sizeof(devrequest)));
	goto WriteDsp_exit;
    }

    memcpy(xfer_buf, pData, CMV_DATA_WORDS*2);
    pdr = (devrequest *)(xfer_buf+CMV_DATA_WORDS*2);

    /*Get an URB and prepare it for submission*/
    urb = usb_alloc_urb(0);

    FILL_USB_CTRL_REQUEST(pdr,
	    USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
	    kCDC_REQ_SEND_ENCAPSULATED_COMMAND,
	    kADI_CMD_SET_BLOCK, Addr | 0x4000, CMV_DATA_WORDS*2);

    usb_fill_control_urb(urb, pHw->usbdev, usb_sndctrlpipe(pHw->usbdev,0), (UInt8 *)pdr,
                     xfer_buf, CMV_DATA_WORDS*2, CtrlUrbCompletion, NULL);
    
    QueueControlUrb(pHw, urb);

WriteDsp_exit:
    return ;
}

/**********************************************************************************/
/* AdiCustomWrite								  */
/**********************************************************************************/
int AdiCustomWrite(Hardware *pHw, UInt32 Cmd, UInt32 Index, UInt32 Count, UInt8 *pData)
{
    /*get space for data and request struct*/
    struct urb *urb;
	devrequest *pdr;
    UInt8 *xfer_buf = GET_KBUFFER(Count+sizeof(devrequest));
    int ret = USB_ST_NOERROR;

    if (!xfer_buf)
    {
	ZAP(DEBUG_BOOT, DEBUG_ERRORS,
	    (kTAB"kmalloc in AdiCustomWrite for %#X bytes failed\n", Count+sizeof(devrequest)));
	ret = -ENOMEM;
	goto custom_exit;
    }

    memcpy(xfer_buf, pData, Count);
    pdr = (devrequest *)(xfer_buf+CMV_DATA_WORDS*2);

    /*Get an URB and prepare it for submission*/
    urb = usb_alloc_urb(0);

    FILL_USB_CTRL_REQUEST(pdr,
	    USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
	    kCDC_REQ_SEND_ENCAPSULATED_COMMAND,
	    Cmd, Index, Count);

    usb_fill_control_urb(urb, pHw->usbdev, usb_sndctrlpipe(pHw->usbdev,0), (UInt8 *)pdr,
                     xfer_buf, Count, CtrlUrbCompletion, NULL);
    
    QueueControlUrb(pHw, urb);
custom_exit:
    return ret;
} 

/**********************************************************************************/
/* QueueControlUrb								  */
/**********************************************************************************/
static void QueueControlUrb(Hardware *pHw, struct urb *urb)
{
    int ret;
    QUEUED_URB_BUFFER *qItem;

    ZAP(DEBUG_UTILS, DEBUG_ENTRY, ("+++QueueControlUrb (%p)\n", urb));

    /***********************************************************************/
    /* If there's already something in the queue, we don't want to         */
    /* submit this urb directly - we want to put it in the queue, so the   */
    /* urbs get sent in the order they came to us. So, if the queue is not */
    /* empty, fake an error so the urb will be queued locally		   */
    /***********************************************************************/
    if (list_empty(&pHw->CtrlUrbPendQ))
    {
	ZAP(DEBUG_UTILS, DEBUG_INFO, (kTAB"Queue empty - attempting direct submission of urb\n"));
	ret = usb_submit_urb(urb);
    }
    else
    {
	ZAP(DEBUG_UTILS, DEBUG_INFO, (kTAB"Queue not empty - queuing urb locally\n"));
	ret = -ENXIO;
    }

    /*************************************************************************/
    /* The linux usb port driver for UHCI usb hardware does not queue control*/
    /* urbs internally. If there is already an urb on the control pipe it    */
    /* returns the -ENXIO error to us - which means we need to put this urb  */
    /* in our "send this urb later" list. 				     */
    /*************************************************************************/
    if (-ENXIO == ret)
    {
	qItem = (QUEUED_URB_BUFFER *)GetBufferFromQ(&pHw->CtrlUrbFreeQ);
	if (NULL == qItem)
	{
	    ZAP(DEBUG_UTILS, DEBUG_ERRORS, (kTAB"No QUEUED_URB_BUFFERs available!\n"));
	    goto QueueControl_exit;
	}
	
	qItem->urb = urb;
	qItem->pHw = pHw;
	list_add_tail((QHdr *)qItem, &pHw->CtrlUrbPendQ);
	
	/*We're adding something to the queue ... make our timer check RSN*/
	mod_timer(&pHw->CtrlUrbQueueTimer, jiffies+MSEC_TO_JIFFIES(100));
    }

QueueControl_exit:
    ZAP(DEBUG_UTILS, DEBUG_EXIT, ("---QueueControlUrb\n"));
    return;
}

/**********************************************************************************/
/* CtrlUrbCompletion								  */
/**********************************************************************************/
static void CtrlUrbCompletion(struct urb *urb)
{
    void *q;

    ZAP(DEBUG_UTILS, DEBUG_ENTRY, ("+++CtrlUrbCompletion\n"));

    /*If this was queued, put it back in the free pool*/
    if ((q = urb->context) != NULL)
    {
	ZAP(DEBUG_UTILS, DEBUG_INFO, (kTAB"Returned CTRL URB BUFFER %p to free queue\n", q));
	list_add_tail((QHdr *)q, &((Hardware *)(((QUEUED_URB_BUFFER *)q)->pHw))->CtrlUrbFreeQ);
	((QUEUED_URB_BUFFER *)q)->urb = NULL;
    }

    /*All we really need to do is free the memory we alloced and free the urb*/
    FREE_KBUFFER(urb->transfer_buffer);
    usb_free_urb(urb);

    ZAP(DEBUG_UTILS, DEBUG_INFO, (kTAB"Freed CTRL URB %p\n", urb));

    ZAP(DEBUG_UTILS, DEBUG_EXIT, ("---CtrlUrbCompletion\n"));
}

/**********************************************************************************/
/* CtrlUrbQueueWatcher								  */
/**********************************************************************************/
void CtrlUrbQueueWatcher(unsigned long ptr)
{
    Hardware *pHw = (Hardware *)ptr;
    QUEUED_URB_BUFFER *qItem;

    /*Are there any URBs queued for delivery? If so, get the next one and send it*/
    qItem = (QUEUED_URB_BUFFER *)GetBufferFromQ(&pHw->CtrlUrbPendQ);
    if ((qItem != NULL) && (qItem->urb != NULL))
    {
	int ret;
	ret = usb_submit_urb(qItem->urb);
	if (ret < 0)
	    printk(KERN_ERR "ADI: Failed to submut control URB %p\n", qItem->urb);
    }
    else
    if ((qItem != NULL) && (qItem->urb == NULL))
    {
	ZAP(DEBUG_UTILS, DEBUG_ERRORS,
	    (kTAB"Warning - queued URB has NULL urb pointer - this shouldn't happen!\n"));
    }

    /*Reschedule the watcher based on the queue status*/
    if (qItem == NULL)
    {
	/*No items were in the queue ... so lets make the timer go off in 1 sec*/
	pHw->CtrlUrbQueueTimer.expires = jiffies + MSEC_TO_JIFFIES(1000);
    }
    else
    {
	/*There was at least one item in the queue, check again RSN*/
	pHw->CtrlUrbQueueTimer.expires = jiffies + MSEC_TO_JIFFIES(100);
    }

    add_timer(&pHw->CtrlUrbQueueTimer);
    return;
}



/*********************************************************************************
$Log: Util.c,v $
Revision 2.00  2002/05/24 21:59:30  Anoosh Naderi
Clean upthe code.

Revision 1.9  2002/01/14 21:59:30  chris.edgington
Added GPL header.

Revision 1.8  2002/01/08 20:06:12  chris.edgington
Removed UnicodeHexStringToBytes - usb does Unicode conversion for us.
Wrote htoi to assist in converting MAC string to numeric format.
Finished GetMacAddress.

Revision 1.7  2002/01/08 16:23:40  chris.edgington
Added #ifdef around GetMacAddress code until it is working.

Revision 1.6  2002/01/08 16:10:10  chris.edgington
Added some necessary comments.
Removed AdiCustomWriteIm - no longer needed.
Added GetMacAddress and UnicodeHexStringToBytes.

Revision 1.5  2002/01/02 22:02:49  chris.edgington
Added PutBufferInQ function.

Revision 1.4  2001/12/28 19:31:33  chris.edgington
Fixed AllocQueuedBuffers and DeAllocQueuedBuffers to work on Linux.
Changed all routines that send Control urbs to use QueueControlUrb.
Added QueueControlUrb, CtrlUrbCompletion, and CtrlUrbQueueWatcher.

Revision 1.3  2001/12/26 22:46:51  chris.edgington
Changed WriteDspMsg to use URB - fixes kernel crash when in interrupt.
Moved AdiCustomWrite to here from Boot.c and changed it to use a dynamic URB.

Revision 1.2  2001/12/22 21:40:29  chris.edgington
Added WriteDspMsg.

Revision 1.1  2001/12/22 19:54:33  chris.edgington
First version - moved routines from AdiUsbAdslDriver.c to here.
***********************************************************************************/
