/* $Id: audiofunc.c,v 1.28 2003/01/30 02:30:57 cochran Exp $
 *
 * Driver for the Extigy (tm) USB sound card - generic audio funtions.
 *
 * Copyright (C) 1999, 2000 Alan Cox (alan@lxorguk.ukuu.org.uk) and
 * Thomas Sailer (sailer@ife.ee.ethz.ch)
 *
 * Adapted from the generic USB audio driver.
 * Copyright (c) 2002 Richard Cochran (cochran@cs.umass.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 */

/* Notes from the audio.c authors:
 *
 * Data IO implementation issues
 *
 * A mmap'able ring buffer per direction is implemented, because
 * almost every OSS app expects it. It is however impractical to
 * transmit/receive USB data directly into and out of the ring buffer,
 * due to alignment and synchronisation issues. Instead, the ring buffer
 * feeds a constant time delay line that handles the USB issues.
 *
 * Now we first try to find an alternate setting that exactly matches
 * the sample format requested by the user. If we find one, we do not
 * need to perform any sample rate conversions. If there is no matching
 * altsetting, we choose the closest one and perform sample format
 * conversions. We never do sample rate conversion; these are too
 * expensive to be performed in the kernel.
 *
 * The AudioStreaming parser simply enumerates all altsettings belonging
 * to the specified interface. It looks for AS_GENERAL and FORMAT_TYPE
 * class specific descriptors to extract the sample format/sample rate
 * data. Only sample format types PCM and PCM8 are supported right now, and
 * only FORMAT_TYPE_I is handled. The isochronous data endpoint needs to
 * be the first endpoint of the interface, and the optional synchronisation
 * isochronous endpoint the second one.
 */

#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/module.h>
#include <linux/sound.h>
#include <linux/soundcard.h>
#include <linux/list.h>
#include <linux/vmalloc.h>
#include <linux/wrapper.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/bitops.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/usb.h>
#include <linux/proc_fs.h>

#include "exaudio.h"
#include "remote.h"

#define DRIVER_AUTHOR  "Richard Cochran, Alan Cox, Thomas Sailer"
#define DRIVER_DESC    "exaudio driver"
#define DRIVER_VERSION "$Revision: 1.28 $"

#define USB_DT_CS_DEVICE                0x21
#define USB_DT_CS_CONFIG                0x22
#define USB_DT_CS_STRING                0x23
#define USB_DT_CS_INTERFACE             0x24
#define USB_DT_CS_ENDPOINT              0x25

#define CS_AUDIO_UNDEFINED		0x20
#define CS_AUDIO_DEVICE			0x21
#define CS_AUDIO_CONFIGURATION		0x22
#define CS_AUDIO_STRING			0x23
#define CS_AUDIO_INTERFACE		0x24
#define CS_AUDIO_ENDPOINT		0x25

#define HEADER				0x01
#define INPUT_TERMINAL			0x02
#define OUTPUT_TERMINAL			0x03
#define MIXER_UNIT			0x04
#define SELECTOR_UNIT			0x05
#define FEATURE_UNIT			0x06
#define PROCESSING_UNIT			0x07
#define EXTENSION_UNIT			0x08

#define AS_GENERAL			0x01
#define FORMAT_TYPE			0x02
#define FORMAT_SPECIFIC			0x03
#define EP_GENERAL			0x01
#define MAX_CHAN   9
#define MAX_FREQ   16
#define MAX_IFACE  8
#define MAX_FORMAT 8
#define MAX_ALT    32 /* Sorry, we need quite a few for the Philips webcams */

#define COPY_PROTECT_CONTROL       0x01   /* Terminal Control Selectors */
#define MUTE_CONTROL               0x01   /* Feature Unit Control Selectors */
#define VOLUME_CONTROL             0x02
#define BASS_CONTROL               0x03
#define MID_CONTROL                0x04
#define TREBLE_CONTROL             0x05
#define GRAPHIC_EQUALIZER_CONTROL  0x06
#define AUTOMATIC_GAIN_CONTROL     0x07
#define DELAY_CONTROL              0x08
#define BASS_BOOST_CONTROL         0x09
#define LOUDNESS_CONTROL           0x0a
#define SAMPLING_FREQ_CONTROL      0x01   /* Endpoint Control Selectors */
#define PITCH_CONTROL              0x02

#define FLG_URB0RUNNING   1
#define FLG_URB1RUNNING   2
#define FLG_SYNC0RUNNING  4
#define FLG_SYNC1RUNNING  8
#define FLG_RUNNING      16
#define FLG_CONNECTED    32

#define SND_DEV_DSP16 5

#define MAXFORMATS        MAX_ALT
#define DMABUFSHIFT       17	/* 128k worth of DMA buffer */
#define NRSGBUF           (1U<<(DMABUFSHIFT-PAGE_SHIFT))

/*
 * This influences:
 * - Latency
 * - Interrupt rate
 * - Synchronisation behaviour
 * Don't touch this if you don't understand all of the above.
 */
#define DESCFRAMES  5
#define SYNCFRAMES  DESCFRAMES

/* Mixer defines, especially for exaudio */
#define TRUE  1
#define FALSE 0
#define RECORD   1
#define PLAYBACK 2

#define N_RECMIX_CH 4
#define REC_LINE_UNIT   8
#define REC_MIC_UNIT    10
#define REC_SRC_UNIT    11
#define REC_MASTER_UNIT 12

#define N_PLAYMIX_CH 10
#define PLAY_PCM_UNIT     2
#define PLAY_DIGIN_UNIT   6
#define PLAY_MASTER_UNIT  18
#define PLAY_CPAIR1_UNIT  18
#define PLAY_CPAIR2_UNIT  18
#define PLAY_CPAIR3_UNIT  18
#define PLAY_SPEAKER_UNIT 28
#define PLAY_DIGOUT1_UNIT 25
#define PLAY_DIGOUT2_UNIT 23
#define PLAY_DIGOUTS_UNIT 22

#define EX1_UNIT  27
#define EX2_UNIT  21
#define FX3D_UNIT 15

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,19)
typedef struct iso_packet_descriptor iso_packet_descriptor_t;
#endif

/* Linked list of all audio devices... */
struct list_head audiodevs = LIST_HEAD_INIT(audiodevs);

DECLARE_MUTEX(open_sem);

/* Wait queue for processes wanting to open an USB audio device */
static DECLARE_WAIT_QUEUE_HEAD(open_wait);

struct audioformat {
  unsigned int format;
  unsigned int sratelo;
  unsigned int sratehi;
  unsigned char altsetting;
  unsigned char attributes;
};

struct dmabuf {
  unsigned int format; /* buffer data format */
  unsigned int srate;
  unsigned char *sgbuf[NRSGBUF]; /* physical buffer */
  unsigned bufsize;
  unsigned numfrag;
  unsigned fragshift;
  unsigned wrptr, rdptr;
  unsigned total_bytes;
  int count;
  unsigned error; /* over/underrun */
  wait_queue_head_t wait;
  unsigned fragsize; /* redundant, but makes calculations easier */
  unsigned dmasize;
  unsigned mapped:1; /* OSS stuff */
  unsigned ready:1;
  unsigned ossfragshift;
  int ossmaxfrags;
  unsigned subdivision;
};

struct my_data_urb {
  struct urb urb;
  iso_packet_descriptor_t isoframe[DESCFRAMES];
};

struct my_sync_urb {
  struct urb urb;
  iso_packet_descriptor_t isoframe[SYNCFRAMES];
};

/*  A "struct audiodev" is created for each Audio Streaming interface
 *  on the extigy deivce.
 */
struct audiodev {
  struct list_head list; /* points to next ASIF in chain */
  struct extigy *state;  /* points back to the device struct */
  int ossdsp;       /* soundcore stuff */
  mode_t open_mode; /* wave stuff */
  spinlock_t lock;  /* DMA buffer access spinlock */
  struct usbin {
    int interface;         /* Interface number, -1 means not used */
    unsigned int format;   /* USB data format */
    unsigned int datapipe; /* the data input pipe */
    unsigned int syncpipe; /* the sync pipe for adaptive IN mode */
    unsigned int syncinterval; /* P for adaptive IN mode, 0 otherwise */
    unsigned int freqn; /* nominal sampling rate USB format fs/1000 in Q10.14 */
    unsigned int freqmax; /* maximum sampling rate, for buffer management */
    unsigned int phase;		/* phase accumulator */
    unsigned int flags;		/* see FLG_ defines */
    struct my_data_urb durb[2];	/* ISO descriptors for the data endpoint */
    struct my_sync_urb surb[2];	/* ISO sync pipe descriptor if needed */
    struct dmabuf dma;
  } usbin;
  struct usbout {
    int interface;         /* Interface number, -1 means not used */
    unsigned int format;   /* USB data format */
    unsigned int datapipe; /* the data input pipe */
    unsigned int syncpipe; /* the sync pipe for asynchronous OUT mode */
    unsigned int syncinterval; /* P for asynchronous OUT mode, 0 otherwise */
    unsigned int freqn; /* nominal sampling rate USB format fs/1000 in Q10.14 */
    unsigned int freqm; /* momentary sampling rate in USB format */
    unsigned int freqmax; /* maximum sampling rate, for buffer management */
    unsigned int phase;         /* phase accumulator */
    unsigned int flags;         /* see FLG_ defines */
    struct my_data_urb durb[2];	/* ISO descriptors for the data endpoint */
    struct my_sync_urb surb[2];	/* ISO sync pipe descriptor if needed */
    struct dmabuf dma;
  } usbout;
  unsigned int numfmtin, numfmtout;
  struct audioformat fmtin[MAXFORMATS];
  struct audioformat fmtout[MAXFORMATS];
};

struct mixerchannel {
  __u16 osschannel;   /* Type of OSS channel */
  __u16 value;        /* Current OSS value */
  __u8  volctl;       /* Feature unit id for VOLUME */
  __u8  mutectl;      /* Feature unit id for MUTE */
  __u8  leftch;       /* Feature unit channel number, where zero means */
  __u8  rightch;      /*  a mono control using the master FU channel   */
  __s16 min;          /* USB min value */
  __s16 max;          /* USB max value */
  __s16 step;         /* USB resolution */
};

struct mixerdev {
  struct list_head list;
  struct extigy *state;
  unsigned char type;         /* RECORD or PLAYBACK */
  unsigned short selector;    /* For setting the recording source */
  int ossmixer;               /* soundcore stuff */
  unsigned char iface;        /* interface number of AudioControl interface */
  unsigned int numch, modcnt; /* USB format descriptions */
  struct mixerchannel ch[0];  /* mixch is last and gets allocated dynamically */
};

/* private audio format extensions */
#define AFMT_STEREO        0x80000000
#define AFMT_ISSTEREO(x)   ((x) & AFMT_STEREO)
#define AFMT_IS16BIT(x)    ((x) & (AFMT_S16_LE|AFMT_S16_BE|AFMT_U16_LE|AFMT_U16_BE))
#define AFMT_ISUNSIGNED(x) ((x) & (AFMT_U8|AFMT_U16_LE|AFMT_U16_BE))
#define AFMT_BYTESSHIFT(x) ((AFMT_ISSTEREO(x) ? 1 : 0) + (AFMT_IS16BIT(x) ? 1 : 0))
#define AFMT_BYTES(x)      (1<<AFMT_BYTESSHFIT(x))

/* private prototypes */

static void construct_mixer(struct extigy *s,unsigned char *desc,
			    unsigned int len,unsigned int ctrlif);

static void* exaudio_probe(struct usb_device *dev, unsigned int ifnum,
			   const struct usb_device_id *id);

static void exaudio_disconnect(struct usb_device *dev, void *ptr);

/*----------------------------------------------------------------------------*/
extern inline unsigned ld2(unsigned int x)
{
  unsigned r = 0;

  if (x >= 0x10000) {
    x >>= 16;
    r += 16;
  }
  if (x >= 0x100) {
    x >>= 8;
    r += 8;
  }
  if (x >= 0x10) {
    x >>= 4;
    r += 4;
  }
  if (x >= 4) {
    x >>= 2;
    r += 2;
  }
  if (x >= 2)
    r++;
  return r;
}

/*----------------------------------------------------------------------------*/
/*
 *  OSS compatible ring buffer management. The ring buffer may be
 *  mmap'ed into an application address space.
 *
 *  I first used the rvmalloc stuff copied from bttv. Alan Cox did not
 *  like it, so we now use an array of pointers to a single page
 *  each. This saves us the kernel page table manipulations, but we
 *  have to do a page table alike mechanism (though only one
 *  indirection) in software.
 */
static void dmabuf_release(struct dmabuf *db)
{
  unsigned int nr;
  void *p;

  for (nr = 0; nr < NRSGBUF; nr++) {
    if (!(p = db->sgbuf[nr]))
      continue;
    mem_map_unreserve(virt_to_page(p));
    free_page((unsigned long) p);
    db->sgbuf[nr] = NULL;
  }
  db->mapped = db->ready = 0;
}

static int dmabuf_init(struct dmabuf *db)
{
  unsigned int nr, bytepersec, bufs;
  void *p;

  /* initialize some fields */
  db->rdptr = db->wrptr = db->total_bytes = db->count = db->error = 0;
  /* calculate required buffer size */
  bytepersec = db->srate << AFMT_BYTESSHIFT(db->format);
  bufs = 1U << DMABUFSHIFT;
  if (db->ossfragshift) {
    if ((1000 << db->ossfragshift) < bytepersec)
      db->fragshift = ld2(bytepersec / 1000);
    else
      db->fragshift = db->ossfragshift;
  } else {
    db->fragshift =
      ld2(bytepersec / 100 / (db->subdivision ? db->subdivision : 1));
    if (db->fragshift < 3)
      db->fragshift = 3;
  }
  db->numfrag = bufs >> db->fragshift;
  while (db->numfrag < 4 && db->fragshift > 3) {
    db->fragshift--;
    db->numfrag = bufs >> db->fragshift;
  }
  db->fragsize = 1 << db->fragshift;
  if (db->ossmaxfrags >= 4 && db->ossmaxfrags < db->numfrag)
    db->numfrag = db->ossmaxfrags;
  db->dmasize = db->numfrag << db->fragshift;
  for (nr = 0; nr < NRSGBUF; nr++) {
    if (!db->sgbuf[nr]) {
      p = (void *) get_free_page(GFP_KERNEL);
      if (!p)
	return -ENOMEM;
      db->sgbuf[nr] = p;
      mem_map_reserve(virt_to_page(p));
    }
    memset(db->sgbuf[nr], AFMT_ISUNSIGNED(db->format) ? 0x80 : 0, PAGE_SIZE);
    if ((nr << PAGE_SHIFT) >= db->dmasize)
      break;
  }
  db->bufsize = nr << PAGE_SHIFT;
  db->ready = 1;
  dbg("dmabuf_init: bytepersec %d bufs %d ossfragshift %d ossmaxfrags %d "
      "fragshift %d fragsize %d numfrag %d dmasize %d bufsize %d fmt 0x%x",
      bytepersec, bufs, db->ossfragshift, db->ossmaxfrags,
      db->fragshift, db->fragsize, db->numfrag, db->dmasize,
      db->bufsize, db->format);
  return 0;
}

static int
dmabuf_mmap(struct dmabuf *db, unsigned long start, unsigned long size,
	    pgprot_t prot)
{
  unsigned int nr;

  if (!db->ready || db->mapped || (start | size) & (PAGE_SIZE - 1)
      || size > db->bufsize)
    return -EINVAL;
  size >>= PAGE_SHIFT;
  for (nr = 0; nr < size; nr++)
    if (!db->sgbuf[nr])
      return -EINVAL;
  db->mapped = 1;
  for (nr = 0; nr < size; nr++) {
    if (remap_page_range(start, virt_to_phys(db->sgbuf[nr]), PAGE_SIZE, prot))
      return -EAGAIN;
    start += PAGE_SIZE;
  }
  return 0;
}

static void
dmabuf_copyin(struct dmabuf *db, const void *buffer, unsigned int size)
{
  unsigned int pgrem, rem;

  db->total_bytes += size;
  for (;;) {
    if (size <= 0)
      return;
    pgrem = ((~db->wrptr) & (PAGE_SIZE - 1)) + 1;
    if (pgrem > size)
      pgrem = size;
    rem = db->dmasize - db->wrptr;
    if (pgrem > rem)
      pgrem = rem;
    memcpy((db->sgbuf[db->wrptr >> PAGE_SHIFT]) +
	   (db->wrptr & (PAGE_SIZE - 1)), buffer, pgrem);
    size -= pgrem;
    (char *) buffer += pgrem;
    db->wrptr += pgrem;
    if (db->wrptr >= db->dmasize)
      db->wrptr = 0;
  }
}

static void dmabuf_copyout(struct dmabuf *db, void *buffer, unsigned int size)
{
  unsigned int pgrem, rem;

  db->total_bytes += size;
  for (;;) {
    if (size <= 0)
      return;
    pgrem = ((~db->rdptr) & (PAGE_SIZE - 1)) + 1;
    if (pgrem > size)
      pgrem = size;
    rem = db->dmasize - db->rdptr;
    if (pgrem > rem)
      pgrem = rem;
    memcpy(buffer,
	   (db->sgbuf[db->rdptr >> PAGE_SHIFT]) +
	   (db->rdptr & (PAGE_SIZE - 1)), pgrem);
    size -= pgrem;
    (char *) buffer += pgrem;
    db->rdptr += pgrem;
    if (db->rdptr >= db->dmasize)
      db->rdptr = 0;
  }
}

static int
dmabuf_copyin_user(struct dmabuf *db, unsigned int ptr, const void *buffer,
		   unsigned int size)
{
  unsigned int pgrem, rem;

  if (!db->ready || db->mapped)
    return -EINVAL;
  for (;;) {
    if (size <= 0)
      return 0;
    pgrem = ((~ptr) & (PAGE_SIZE - 1)) + 1;
    if (pgrem > size)
      pgrem = size;
    rem = db->dmasize - ptr;
    if (pgrem > rem)
      pgrem = rem;
    if (copy_from_user
	((db->sgbuf[ptr >> PAGE_SHIFT]) + (ptr & (PAGE_SIZE - 1)), buffer,
	 pgrem))
      return -EFAULT;
    size -= pgrem;
    (char *) buffer += pgrem;
    ptr += pgrem;
    if (ptr >= db->dmasize)
      ptr = 0;
  }
}

static int
dmabuf_copyout_user(struct dmabuf *db, unsigned int ptr, void *buffer,
		    unsigned int size)
{
  unsigned int pgrem, rem;

  if (!db->ready || db->mapped)
    return -EINVAL;
  for (;;) {
    if (size <= 0)
      return 0;
    pgrem = ((~ptr) & (PAGE_SIZE - 1)) + 1;
    if (pgrem > size)
      pgrem = size;
    rem = db->dmasize - ptr;
    if (pgrem > rem)
      pgrem = rem;
    if (copy_to_user
	(buffer, (db->sgbuf[ptr >> PAGE_SHIFT]) + (ptr & (PAGE_SIZE - 1)),
	 pgrem))
      return -EFAULT;
    size -= pgrem;
    (char *) buffer += pgrem;
    ptr += pgrem;
    if (ptr >= db->dmasize)
      ptr = 0;
  }
}

/*----------------------------------------------------------------------------*/
/* USB I/O code. We do sample format conversion if necessary. */

static void usbin_stop(struct audiodev *as)
{
  struct usbin *u = &as->usbin;
  unsigned long flags;
  unsigned int i, notkilled = 1;

  spin_lock_irqsave(&as->lock, flags);
  u->flags &= ~FLG_RUNNING;
  i = u->flags;
  spin_unlock_irqrestore(&as->lock, flags);
  while (i &
	 (FLG_URB0RUNNING | FLG_URB1RUNNING | FLG_SYNC0RUNNING |
	  FLG_SYNC1RUNNING)) {
    set_current_state(notkilled ? TASK_INTERRUPTIBLE : TASK_UNINTERRUPTIBLE);
    schedule_timeout(1);
    spin_lock_irqsave(&as->lock, flags);
    i = u->flags;
    spin_unlock_irqrestore(&as->lock, flags);
    if (notkilled && signal_pending(current)) {
      if (i & FLG_URB0RUNNING)
	usb_unlink_urb(&u->durb[0].urb);
      if (i & FLG_URB1RUNNING)
	usb_unlink_urb(&u->durb[1].urb);
      if (i & FLG_SYNC0RUNNING)
	usb_unlink_urb(&u->surb[0].urb);
      if (i & FLG_SYNC1RUNNING)
	usb_unlink_urb(&u->surb[1].urb);
      notkilled = 0;
    }
  }
  set_current_state(TASK_RUNNING);
  if (u->durb[0].urb.transfer_buffer)
    kfree(u->durb[0].urb.transfer_buffer);
  if (u->durb[1].urb.transfer_buffer)
    kfree(u->durb[1].urb.transfer_buffer);
  if (u->surb[0].urb.transfer_buffer)
    kfree(u->surb[0].urb.transfer_buffer);
  if (u->surb[1].urb.transfer_buffer)
    kfree(u->surb[1].urb.transfer_buffer);
  u->durb[0].urb.transfer_buffer = u->durb[1].urb.transfer_buffer =
    u->surb[0].urb.transfer_buffer = u->surb[1].urb.transfer_buffer = NULL;
}

static inline void usbin_release(struct audiodev *as)
{
  usbin_stop(as);
}

static void usbin_disc(struct audiodev *as)
{
  struct usbin *u = &as->usbin;

  unsigned long flags;

  spin_lock_irqsave(&as->lock, flags);
  u->flags &= ~(FLG_RUNNING | FLG_CONNECTED);
  spin_unlock_irqrestore(&as->lock, flags);
  usbin_stop(as);
}

static void
conversion(const void *ibuf, unsigned int ifmt, void *obuf,
	   unsigned int ofmt, void *tmp, unsigned int scnt)
{
  unsigned int cnt, i;
  __s16 *sp, *sp2, s;
  unsigned char *bp;

  cnt = scnt;
  if (AFMT_ISSTEREO(ifmt))
    cnt <<= 1;
  sp = ((__s16 *) tmp) + cnt;
  switch (ifmt & ~AFMT_STEREO) {
  case AFMT_U8:
    for (bp = ((unsigned char *) ibuf) + cnt, i = 0; i < cnt; i++) {
      bp--;
      sp--;
      *sp = (*bp ^ 0x80) << 8;
    }
    break;

  case AFMT_S8:
    for (bp = ((unsigned char *) ibuf) + cnt, i = 0; i < cnt; i++) {
      bp--;
      sp--;
      *sp = *bp << 8;
    }
    break;

  case AFMT_U16_LE:
    for (bp = ((unsigned char *) ibuf) + 2 * cnt, i = 0; i < cnt; i++) {
      bp -= 2;
      sp--;
      *sp = (bp[0] | (bp[1] << 8)) ^ 0x8000;
    }
    break;

  case AFMT_U16_BE:
    for (bp = ((unsigned char *) ibuf) + 2 * cnt, i = 0; i < cnt; i++) {
      bp -= 2;
      sp--;
      *sp = (bp[1] | (bp[0] << 8)) ^ 0x8000;
    }
    break;

  case AFMT_S16_LE:
    for (bp = ((unsigned char *) ibuf) + 2 * cnt, i = 0; i < cnt; i++) {
      bp -= 2;
      sp--;
      *sp = bp[0] | (bp[1] << 8);
    }
    break;

  case AFMT_S16_BE:
    for (bp = ((unsigned char *) ibuf) + 2 * cnt, i = 0; i < cnt; i++) {
      bp -= 2;
      sp--;
      *sp = bp[1] | (bp[0] << 8);
    }
    break;
  }
  if (!AFMT_ISSTEREO(ifmt) && AFMT_ISSTEREO(ofmt)) {
    /* expand from mono to stereo */
    for (sp = ((__s16 *) tmp) + scnt, sp2 = ((__s16 *) tmp) + 2 * scnt, i =
	 0; i < scnt; i++) {
      sp--;
      sp2 -= 2;
      sp2[0] = sp2[1] = sp[0];
    }
  }
  if (AFMT_ISSTEREO(ifmt) && !AFMT_ISSTEREO(ofmt)) {
    /* contract from stereo to mono */
    for (sp = sp2 = ((__s16 *) tmp), i = 0; i < scnt; i++, sp++, sp2 += 2)
      sp[0] = (sp2[0] + sp2[1]) >> 1;
  }
  cnt = scnt;
  if (AFMT_ISSTEREO(ofmt))
    cnt <<= 1;
  sp = ((__s16 *) tmp);
  bp = ((unsigned char *) obuf);
  switch (ofmt & ~AFMT_STEREO) {
  case AFMT_U8:
    for (i = 0; i < cnt; i++, sp++, bp++)
      *bp = (*sp >> 8) ^ 0x80;
    break;

  case AFMT_S8:
    for (i = 0; i < cnt; i++, sp++, bp++)
      *bp = *sp >> 8;
    break;

  case AFMT_U16_LE:
    for (i = 0; i < cnt; i++, sp++, bp += 2) {
      s = *sp;
      bp[0] = s;
      bp[1] = (s >> 8) ^ 0x80;
    }
    break;

  case AFMT_U16_BE:
    for (i = 0; i < cnt; i++, sp++, bp += 2) {
      s = *sp;
      bp[1] = s;
      bp[0] = (s >> 8) ^ 0x80;
    }
    break;

  case AFMT_S16_LE:
    for (i = 0; i < cnt; i++, sp++, bp += 2) {
      s = *sp;
      bp[0] = s;
      bp[1] = s >> 8;
    }
    break;

  case AFMT_S16_BE:
    for (i = 0; i < cnt; i++, sp++, bp += 2) {
      s = *sp;
      bp[1] = s;
      bp[0] = s >> 8;
    }
    break;
  }

}

static void
usbin_convert(struct usbin *u, unsigned char *buffer, unsigned int samples)
{
  union {
    __s16 s[64];
    unsigned char b[0];
  } tmp;
  unsigned int scnt, maxs, ufmtsh, dfmtsh;

  ufmtsh = AFMT_BYTESSHIFT(u->format);
  dfmtsh = AFMT_BYTESSHIFT(u->dma.format);
  maxs = (AFMT_ISSTEREO(u->dma.format | u->format)) ? 32 : 64;
  while (samples > 0) {
    scnt = samples;
    if (scnt > maxs)
      scnt = maxs;
    conversion(buffer, u->format, tmp.b, u->dma.format, tmp.b, scnt);
    dmabuf_copyin(&u->dma, tmp.b, scnt << dfmtsh);
    buffer += scnt << ufmtsh;
    samples -= scnt;
  }
}

static int usbin_prepare_desc(struct usbin *u, struct urb* urb)
{
  unsigned int i, maxsize, offs;

  maxsize = (u->freqmax + 0x3fff) >> (14 - AFMT_BYTESSHIFT(u->format));
  for (i = offs = 0; i < DESCFRAMES; i++, offs += maxsize) {
    urb->iso_frame_desc[i].length = maxsize;
    urb->iso_frame_desc[i].offset = offs;
  }
  return 0;
}

/*
 * return value: 0 if descriptor should be restarted, -1 otherwise
 * convert sample format on the fly if necessary
 */
static int usbin_retire_desc(struct usbin *u, struct urb* urb)
{
  unsigned int i, ufmtsh, dfmtsh, err = 0, cnt, scnt, dmafree;
  unsigned char *cp;

  ufmtsh = AFMT_BYTESSHIFT(u->format);
  dfmtsh = AFMT_BYTESSHIFT(u->dma.format);
  for (i = 0; i < DESCFRAMES; i++) {
    cp =
      ((unsigned char *) urb->transfer_buffer) +
      urb->iso_frame_desc[i].offset;
    if (urb->iso_frame_desc[i].status) {
      dbg("usbin_retire_desc: frame %u status %d",
	  i,urb->iso_frame_desc[i].status);
      continue;
    }
    scnt = urb->iso_frame_desc[i].actual_length >> ufmtsh;
    if (!scnt)
      continue;
    cnt = scnt << dfmtsh;
    if (!u->dma.mapped) {
      dmafree = u->dma.dmasize - u->dma.count;
      if (cnt > dmafree) {
	scnt = dmafree >> dfmtsh;
	cnt = scnt << dfmtsh;
	err++;
      }
    }
    u->dma.count += cnt;
    if (u->format == u->dma.format) {
      /* we do not need format conversion */
      dmabuf_copyin(&u->dma, cp, cnt);
    } else {
      /* we need sampling format conversion */
      usbin_convert(u, cp, scnt);
    }
  }
  if (err)
    u->dma.error++;
  if (u->dma.count >= (signed) u->dma.fragsize)
    wake_up(&u->dma.wait);
  return err ? -1 : 0;
}

static void usbin_completed(struct urb *urb)
{
  struct audiodev *as = (struct audiodev *) urb->context;
  struct usbin *u = &as->usbin;
  unsigned long flags;
  unsigned int mask;
  int suret = USB_ST_NOERROR;

  dbg("usbin_completed: status %d errcnt %d flags 0x%x",
      urb->status,urb->error_count,u->flags);

  if (urb == &u->durb[0].urb)
    mask = FLG_URB0RUNNING;
  else if (urb == &u->durb[1].urb)
    mask = FLG_URB1RUNNING;
  else {
    mask = 0;
    printk(KERN_ERR "usbin_completed: panic: unknown URB\n");
  }
  urb->dev = as->state->usbdev;
  spin_lock_irqsave(&as->lock, flags);
  if (!usbin_retire_desc(u, urb) &&
      u->flags & FLG_RUNNING &&
      !usbin_prepare_desc(u, urb) &&
      (suret = usb_submit_urb(urb)) == USB_ST_NOERROR) {
    u->flags |= mask;
  } else {
    u->flags &= ~(mask | FLG_RUNNING);
    wake_up(&u->dma.wait);
    dbg("usbin_completed: descriptor not restarted (usb_submit_urb: %d)",suret);
  }
  spin_unlock_irqrestore(&as->lock, flags);
}

/*
 * we output sync data
 */
static int usbin_sync_prepare_desc(struct usbin *u, struct urb* urb)
{
  unsigned char *cp = urb->transfer_buffer;
  unsigned int i, offs;

  for (i = offs = 0; i < SYNCFRAMES; i++, offs += 3, cp += 3) {
    urb->iso_frame_desc[i].length = 3;
    urb->iso_frame_desc[i].offset = offs;
    cp[0] = u->freqn;
    cp[1] = u->freqn >> 8;
    cp[2] = u->freqn >> 16;
  }
  return 0;
}

/*
 * return value: 0 if descriptor should be restarted, -1 otherwise
 */
static int usbin_sync_retire_desc(struct usbin *u, struct urb* urb)
{
  unsigned int i;

  for (i = 0; i < SYNCFRAMES; i++)
    if (urb->iso_frame_desc[0].status)
      dbg("usbin_sync_retire_desc: frame %u status %d",
	  i,urb->iso_frame_desc[i].status);
  return 0;
}

static void usbin_sync_completed(struct urb *urb)
{
  struct audiodev *as = (struct audiodev *) urb->context;
  struct usbin *u = &as->usbin;
  unsigned long flags;
  unsigned int mask;
  int suret = USB_ST_NOERROR;

#if 0
  printk(KERN_DEBUG "usbin_sync_completed: status %d errcnt %d flags 0x%x\n",
	 urb->status, urb->error_count, u->flags);
#endif
  if (urb == &u->surb[0].urb)
    mask = FLG_SYNC0RUNNING;
  else if (urb == &u->surb[1].urb)
    mask = FLG_SYNC1RUNNING;
  else {
    mask = 0;
    printk(KERN_ERR "usbin_sync_completed: panic: unknown URB\n");
  }
  urb->dev = as->state->usbdev;
  spin_lock_irqsave(&as->lock, flags);
  if (!usbin_sync_retire_desc(u, urb) &&
      u->flags & FLG_RUNNING &&
      !usbin_sync_prepare_desc(u, urb) &&
      (suret = usb_submit_urb(urb)) == USB_ST_NOERROR) {
    u->flags |= mask;
  } else {
    u->flags &= ~(mask | FLG_RUNNING);
    wake_up(&u->dma.wait);
    dbg("usbin_sync_completed: descriptor not restarted (usb_submit_urb: %d)",
	suret);
  }
  spin_unlock_irqrestore(&as->lock, flags);
}

static int usbin_start(struct audiodev *as)
{
  struct usb_device *dev = as->state->usbdev;
  struct usbin *u = &as->usbin;
  struct urb* urb;
  unsigned long flags;
  unsigned int maxsze, bufsz;

#if 0
  printk(KERN_DEBUG
	 "usbin_start: device %d ufmt 0x%08x dfmt 0x%08x srate %d\n",
	 dev->devnum, u->format, u->dma.format, u->dma.srate);
#endif
  /* allocate USB storage if not already done */
  spin_lock_irqsave(&as->lock, flags);
  if (!(u->flags & FLG_CONNECTED)) {
    spin_unlock_irqrestore(&as->lock, flags);
    return -EIO;
  }
  if (!(u->flags & FLG_RUNNING)) {
    spin_unlock_irqrestore(&as->lock, flags);
    u->freqn = ((u->dma.srate << 11) + 62) / 125; /* this will overflow at approx 2MSPS */
    u->freqmax = u->freqn + (u->freqn >> 2);
    u->phase = 0;
    maxsze = (u->freqmax + 0x3fff) >> (14 - AFMT_BYTESSHIFT(u->format));
    bufsz = DESCFRAMES * maxsze;
    if (u->durb[0].urb.transfer_buffer)
      kfree(u->durb[0].urb.transfer_buffer);
    u->durb[0].urb.transfer_buffer = kmalloc(bufsz, GFP_KERNEL);
    u->durb[0].urb.transfer_buffer_length = bufsz;
    if (u->durb[1].urb.transfer_buffer)
      kfree(u->durb[1].urb.transfer_buffer);
    u->durb[1].urb.transfer_buffer = kmalloc(bufsz, GFP_KERNEL);
    u->durb[1].urb.transfer_buffer_length = bufsz;
    if (u->syncpipe) {
      if (u->surb[0].urb.transfer_buffer)
	kfree(u->surb[0].urb.transfer_buffer);
      u->surb[0].urb.transfer_buffer = kmalloc(3 * SYNCFRAMES, GFP_KERNEL);
      u->surb[0].urb.transfer_buffer_length = 3 * SYNCFRAMES;
      if (u->surb[1].urb.transfer_buffer)
	kfree(u->surb[1].urb.transfer_buffer);
      u->surb[1].urb.transfer_buffer = kmalloc(3 * SYNCFRAMES, GFP_KERNEL);
      u->surb[1].urb.transfer_buffer_length = 3 * SYNCFRAMES;
    }
    if (!u->durb[0].urb.transfer_buffer || !u->durb[1].urb.transfer_buffer
	|| (u->syncpipe
	    && (!u->surb[0].urb.transfer_buffer
		|| !u->surb[1].urb.transfer_buffer))) {
      printk(KERN_ERR "exaudio: cannot start playback device %d\n",
	     dev->devnum);
      return 0;
    }
    spin_lock_irqsave(&as->lock, flags);
  }
  if (u->dma.count >= u->dma.dmasize && !u->dma.mapped) {
    spin_unlock_irqrestore(&as->lock, flags);
    return 0;
  }
  u->flags |= FLG_RUNNING;
  if (!(u->flags & FLG_URB0RUNNING)) {
    urb = &u->durb[0].urb;
    urb->dev = dev;
    urb->pipe = u->datapipe;
    urb->transfer_flags = USB_ISO_ASAP;
    urb->number_of_packets = DESCFRAMES;
    urb->context = as;
    urb->complete = usbin_completed;
    if (!usbin_prepare_desc(u, urb) && !usb_submit_urb(urb))
      u->flags |= FLG_URB0RUNNING;
    else
      u->flags &= ~FLG_RUNNING;
  }
  if (u->flags & FLG_RUNNING && !(u->flags & FLG_URB1RUNNING)) {
    urb = &u->durb[1].urb;
    urb->dev = dev;
    urb->pipe = u->datapipe;
    urb->transfer_flags = USB_ISO_ASAP;
    urb->number_of_packets = DESCFRAMES;
    urb->context = as;
    urb->complete = usbin_completed;
    if (!usbin_prepare_desc(u, urb) && !usb_submit_urb(urb))
      u->flags |= FLG_URB1RUNNING;
    else
      u->flags &= ~FLG_RUNNING;
  }
  if (u->syncpipe) {
    if (u->flags & FLG_RUNNING && !(u->flags & FLG_SYNC0RUNNING)) {
      urb = &u->surb[0].urb;
      urb->dev = dev;
      urb->pipe = u->syncpipe;
      urb->transfer_flags = USB_ISO_ASAP;
      urb->number_of_packets = SYNCFRAMES;
      urb->context = as;
      urb->complete = usbin_sync_completed;
      /* stride: u->syncinterval */
      if (!usbin_sync_prepare_desc(u, urb) && !usb_submit_urb(urb))
	u->flags |= FLG_SYNC0RUNNING;
      else
	u->flags &= ~FLG_RUNNING;
    }
    if (u->flags & FLG_RUNNING && !(u->flags & FLG_SYNC1RUNNING)) {
      urb = &u->surb[1].urb;
      urb->dev = dev;
      urb->pipe = u->syncpipe;
      urb->transfer_flags = USB_ISO_ASAP;
      urb->number_of_packets = SYNCFRAMES;
      urb->context = as;
      urb->complete = usbin_sync_completed;
      /* stride: u->syncinterval */
      if (!usbin_sync_prepare_desc(u, urb) && !usb_submit_urb(urb))
	u->flags |= FLG_SYNC1RUNNING;
      else
	u->flags &= ~FLG_RUNNING;
    }
  }
  spin_unlock_irqrestore(&as->lock, flags);
  return 0;
}

static void usbout_stop(struct audiodev *as)
{
  struct usbout *u = &as->usbout;
  unsigned long flags;
  unsigned int i, notkilled = 1;

  spin_lock_irqsave(&as->lock, flags);
  u->flags &= ~FLG_RUNNING;
  i = u->flags;
  spin_unlock_irqrestore(&as->lock, flags);
  while (i &
	 (FLG_URB0RUNNING | FLG_URB1RUNNING | FLG_SYNC0RUNNING |
	  FLG_SYNC1RUNNING)) {
    set_current_state(notkilled ? TASK_INTERRUPTIBLE : TASK_UNINTERRUPTIBLE);
    schedule_timeout(1);
    spin_lock_irqsave(&as->lock, flags);
    i = u->flags;
    spin_unlock_irqrestore(&as->lock, flags);
    if (notkilled && signal_pending(current)) {
      if (i & FLG_URB0RUNNING)
	usb_unlink_urb(&u->durb[0].urb);
      if (i & FLG_URB1RUNNING)
	usb_unlink_urb(&u->durb[1].urb);
      if (i & FLG_SYNC0RUNNING)
	usb_unlink_urb(&u->surb[0].urb);
      if (i & FLG_SYNC1RUNNING)
	usb_unlink_urb(&u->surb[1].urb);
      notkilled = 0;
    }
  }
  set_current_state(TASK_RUNNING);
  if (u->durb[0].urb.transfer_buffer)
    kfree(u->durb[0].urb.transfer_buffer);
  if (u->durb[1].urb.transfer_buffer)
    kfree(u->durb[1].urb.transfer_buffer);
  if (u->surb[0].urb.transfer_buffer)
    kfree(u->surb[0].urb.transfer_buffer);
  if (u->surb[1].urb.transfer_buffer)
    kfree(u->surb[1].urb.transfer_buffer);
  u->durb[0].urb.transfer_buffer = u->durb[1].urb.transfer_buffer =
    u->surb[0].urb.transfer_buffer = u->surb[1].urb.transfer_buffer = NULL;
}

static inline void usbout_release(struct audiodev *as)
{
  usbout_stop(as);
}

static void usbout_disc(struct audiodev *as)
{
  struct usbout *u = &as->usbout;
  unsigned long flags;

  spin_lock_irqsave(&as->lock, flags);
  u->flags &= ~(FLG_RUNNING | FLG_CONNECTED);
  spin_unlock_irqrestore(&as->lock, flags);
  usbout_stop(as);
}

static void
usbout_convert(struct usbout *u, unsigned char *buffer, unsigned int samples)
{
  union
  {
    __s16 s[64];
    unsigned char b[0];
  }
  tmp;
  unsigned int scnt, maxs, ufmtsh, dfmtsh;

  ufmtsh = AFMT_BYTESSHIFT(u->format);
  dfmtsh = AFMT_BYTESSHIFT(u->dma.format);
  maxs = (AFMT_ISSTEREO(u->dma.format | u->format)) ? 32 : 64;
  while (samples > 0) {
    scnt = samples;
    if (scnt > maxs)
      scnt = maxs;
    dmabuf_copyout(&u->dma, tmp.b, scnt << dfmtsh);
    conversion(tmp.b, u->dma.format, buffer, u->format, tmp.b, scnt);
    buffer += scnt << ufmtsh;
    samples -= scnt;
  }
}

static int usbout_prepare_desc(struct usbout *u, struct urb* urb)
{
  unsigned int i, ufmtsh, dfmtsh, err = 0, cnt, scnt, offs;
  unsigned char *cp = urb->transfer_buffer;

  ufmtsh = AFMT_BYTESSHIFT(u->format);
  dfmtsh = AFMT_BYTESSHIFT(u->dma.format);
  for (i = offs = 0; i < DESCFRAMES; i++) {
    urb->iso_frame_desc[i].offset = offs;
    u->phase = (u->phase & 0x3fff) + u->freqm;
    scnt = u->phase >> 14;
    if (!scnt) {
      urb->iso_frame_desc[i].length = 0;
      continue;
    }
    cnt = scnt << dfmtsh;
    if (!u->dma.mapped) {
      if (cnt > u->dma.count) {
	scnt = u->dma.count >> dfmtsh;
	cnt = scnt << dfmtsh;
	err++;
      }
      u->dma.count -= cnt;
    } else
      u->dma.count += cnt;
    if (u->format == u->dma.format) {
      /* we do not need format conversion */
      dmabuf_copyout(&u->dma, cp, cnt);
    } else {
      /* we need sampling format conversion */
      usbout_convert(u, cp, scnt);
    }
    cnt = scnt << ufmtsh;
    urb->iso_frame_desc[i].length = cnt;
    offs += cnt;
    cp += cnt;
  }
  if (err)
    u->dma.error++;
  if (u->dma.mapped) {
    if (u->dma.count >= (signed) u->dma.fragsize)
      wake_up(&u->dma.wait);
  } else {
    if ((signed) u->dma.dmasize >= u->dma.count + (signed) u->dma.fragsize)
      wake_up(&u->dma.wait);
  }
  return err ? -1 : 0;
}

/*
 * return value: 0 if descriptor should be restarted, -1 otherwise
 */
static int usbout_retire_desc(struct usbout *u, struct urb* urb)
{
  unsigned int i;

  for (i = 0; i < DESCFRAMES; i++) {
    if (urb->iso_frame_desc[i].status) {
      dbg("usbout_retire_desc: frame %u status %d",
	  i,urb->iso_frame_desc[i].status);
      continue;
    }
  }
  return 0;
}

static void usbout_completed(struct urb *urb)
{
  struct audiodev *as = (struct audiodev *) urb->context;
  struct usbout *u = &as->usbout;
  unsigned long flags;
  unsigned int mask;
  int suret = USB_ST_NOERROR;

#if 0
  printk(KERN_DEBUG "usbout_completed: status %d errcnt %d flags 0x%x\n",
	 urb->status, urb->error_count, u->flags);
#endif
  if (urb == &u->durb[0].urb)
    mask = FLG_URB0RUNNING;
  else if (urb == &u->durb[1].urb)
    mask = FLG_URB1RUNNING;
  else {
    mask = 0;
    printk(KERN_ERR "usbout_completed: panic: unknown URB\n");
  }
  urb->dev = as->state->usbdev;
  spin_lock_irqsave(&as->lock, flags);
  if (!usbout_retire_desc(u, urb) &&
      u->flags & FLG_RUNNING &&
      !usbout_prepare_desc(u, urb) &&
      (suret = usb_submit_urb(urb)) == USB_ST_NOERROR) {
    u->flags |= mask;
  } else {
    u->flags &= ~(mask | FLG_RUNNING);
    wake_up(&u->dma.wait);
    dbg("usbout_completed: descriptor not restarted (usb_submit_urb %d)",suret);
  }
  spin_unlock_irqrestore(&as->lock, flags);
}

static int usbout_sync_prepare_desc(struct usbout *u, struct urb* urb)
{
  unsigned int i, offs;

  for (i = offs = 0; i < SYNCFRAMES; i++, offs += 3) {
    urb->iso_frame_desc[i].length = 3;
    urb->iso_frame_desc[i].offset = offs;
  }
  return 0;
}

/*
 * return value: 0 if descriptor should be restarted, -1 otherwise
 */
static int usbout_sync_retire_desc(struct usbout *u, struct urb* urb)
{
  unsigned char *cp = urb->transfer_buffer;
  unsigned int f, i;

  for (i = 0; i < SYNCFRAMES; i++, cp += 3) {
    if (urb->iso_frame_desc[i].status) {
      dbg("usbout_sync_retire_desc: frame %u status %d",
	  i,urb->iso_frame_desc[i].status);
      continue;
    }
    if (urb->iso_frame_desc[i].actual_length < 3) {
      dbg("usbout_sync_retire_desc: frame %u length %d",
	  i,urb->iso_frame_desc[i].actual_length);
      continue;
    }
    f = cp[0] | (cp[1] << 8) | (cp[2] << 16);
    if (abs(f - u->freqn) > (u->freqn >> 3) || f > u->freqmax) {
      printk(KERN_WARNING
	     "usbout_sync_retire_desc: requested frequency %u (nominal %u) out of range!\n",
	     f, u->freqn);
      continue;
    }
    u->freqm = f;
  }
  return 0;
}

static void usbout_sync_completed(struct urb *urb)
{
  struct audiodev *as = (struct audiodev *) urb->context;
  struct usbout *u = &as->usbout;
  unsigned long flags;
  unsigned int mask;
  int suret = USB_ST_NOERROR;

#if 0
  printk(KERN_DEBUG
	 "usbout_sync_completed: status %d errcnt %d flags 0x%x\n",
	 urb->status, urb->error_count, u->flags);
#endif
  if (urb == &u->surb[0].urb)
    mask = FLG_SYNC0RUNNING;
  else if (urb == &u->surb[1].urb)
    mask = FLG_SYNC1RUNNING;
  else {
    mask = 0;
    printk(KERN_ERR "usbout_sync_completed: panic: unknown URB\n");
  }
  urb->dev = as->state->usbdev;
  spin_lock_irqsave(&as->lock, flags);
  if (!usbout_sync_retire_desc(u, urb) &&
      u->flags & FLG_RUNNING &&
      !usbout_sync_prepare_desc(u, urb) &&
      (suret = usb_submit_urb(urb)) == USB_ST_NOERROR) {
    u->flags |= mask;
  } else {
    u->flags &= ~(mask | FLG_RUNNING);
    wake_up(&u->dma.wait);
    dbg("usbout_sync_completed: descriptor not restarted (usb_submit_urb %d)",
	suret);
  }
  spin_unlock_irqrestore(&as->lock, flags);
}

static int usbout_start(struct audiodev *as)
{
  struct usb_device *dev = as->state->usbdev;
  struct usbout *u = &as->usbout;
  struct urb* urb;
  unsigned long flags;
  unsigned int maxsze, bufsz;

#if 0
  printk(KERN_DEBUG
	 "usbout_start: device %d ufmt 0x%08x dfmt 0x%08x srate %d\n",
	 dev->devnum, u->format, u->dma.format, u->dma.srate);
#endif
  /* allocate USB storage if not already done */
  spin_lock_irqsave(&as->lock, flags);
  if (!(u->flags & FLG_CONNECTED)) {
    spin_unlock_irqrestore(&as->lock, flags);
    return -EIO;
  }
  if (!(u->flags & FLG_RUNNING)) {
    spin_unlock_irqrestore(&as->lock, flags);
    u->freqn = u->freqm = ((u->dma.srate << 11) + 62) / 125; /* this will overflow at approx 2MSPS */
    u->freqmax = u->freqn + (u->freqn >> 2);
    u->phase = 0;
    maxsze = (u->freqmax + 0x3fff) >> (14 - AFMT_BYTESSHIFT(u->format));
    bufsz = DESCFRAMES * maxsze;
    if (u->durb[0].urb.transfer_buffer)
      kfree(u->durb[0].urb.transfer_buffer);
    u->durb[0].urb.transfer_buffer = kmalloc(bufsz, GFP_KERNEL);
    u->durb[0].urb.transfer_buffer_length = bufsz;
    if (u->durb[1].urb.transfer_buffer)
      kfree(u->durb[1].urb.transfer_buffer);
    u->durb[1].urb.transfer_buffer = kmalloc(bufsz, GFP_KERNEL);
    u->durb[1].urb.transfer_buffer_length = bufsz;
    if (u->syncpipe) {
      if (u->surb[0].urb.transfer_buffer)
	kfree(u->surb[0].urb.transfer_buffer);
      u->surb[0].urb.transfer_buffer = kmalloc(3 * SYNCFRAMES, GFP_KERNEL);
      u->surb[0].urb.transfer_buffer_length = 3 * SYNCFRAMES;
      if (u->surb[1].urb.transfer_buffer)
	kfree(u->surb[1].urb.transfer_buffer);
      u->surb[1].urb.transfer_buffer = kmalloc(3 * SYNCFRAMES, GFP_KERNEL);
      u->surb[1].urb.transfer_buffer_length = 3 * SYNCFRAMES;
    }
    if (!u->durb[0].urb.transfer_buffer || !u->durb[1].urb.transfer_buffer
	|| (u->syncpipe
	    && (!u->surb[0].urb.transfer_buffer
		|| !u->surb[1].urb.transfer_buffer))) {
      printk(KERN_ERR "exaudio: cannot start playback device %d\n",
	     dev->devnum);
      return 0;
    }
    spin_lock_irqsave(&as->lock, flags);
  }
  if (u->dma.count <= 0 && !u->dma.mapped) {
    spin_unlock_irqrestore(&as->lock, flags);
    return 0;
  }
  u->flags |= FLG_RUNNING;
  if (!(u->flags & FLG_URB0RUNNING)) {
    urb = &u->durb[0].urb;
    urb->dev = dev;
    urb->pipe = u->datapipe;
    urb->transfer_flags = USB_ISO_ASAP;
    urb->number_of_packets = DESCFRAMES;
    urb->context = as;
    urb->complete = usbout_completed;
    if (!usbout_prepare_desc(u, urb) && !usb_submit_urb(urb))
      u->flags |= FLG_URB0RUNNING;
    else
      u->flags &= ~FLG_RUNNING;
  }
  if (u->flags & FLG_RUNNING && !(u->flags & FLG_URB1RUNNING)) {
    urb = &u->durb[1].urb;
    urb->dev = dev;
    urb->pipe = u->datapipe;
    urb->transfer_flags = USB_ISO_ASAP;
    urb->number_of_packets = DESCFRAMES;
    urb->context = as;
    urb->complete = usbout_completed;
    if (!usbout_prepare_desc(u, urb) && !usb_submit_urb(urb))
      u->flags |= FLG_URB1RUNNING;
    else
      u->flags &= ~FLG_RUNNING;
  }
  if (u->syncpipe) {
    if (u->flags & FLG_RUNNING && !(u->flags & FLG_SYNC0RUNNING)) {
      urb = &u->surb[0].urb;
      urb->dev = dev;
      urb->pipe = u->syncpipe;
      urb->transfer_flags = USB_ISO_ASAP;
      urb->number_of_packets = SYNCFRAMES;
      urb->context = as;
      urb->complete = usbout_sync_completed;
      /* stride: u->syncinterval */
      if (!usbout_sync_prepare_desc(u, urb) && !usb_submit_urb(urb))
	u->flags |= FLG_SYNC0RUNNING;
      else
	u->flags &= ~FLG_RUNNING;
    }
    if (u->flags & FLG_RUNNING && !(u->flags & FLG_SYNC1RUNNING)) {
      urb = &u->surb[1].urb;
      urb->dev = dev;
      urb->pipe = u->syncpipe;
      urb->transfer_flags = USB_ISO_ASAP;
      urb->number_of_packets = SYNCFRAMES;
      urb->context = as;
      urb->complete = usbout_sync_completed;
      /* stride: u->syncinterval */
      if (!usbout_sync_prepare_desc(u, urb) && !usb_submit_urb(urb))
	u->flags |= FLG_SYNC1RUNNING;
      else
	u->flags &= ~FLG_RUNNING;
    }
  }
  spin_unlock_irqrestore(&as->lock, flags);
  return 0;
}

/*----------------------------------------------------------------------------*/
static unsigned int
format_goodness(struct audioformat *afp, unsigned int fmt, unsigned int srate)
{
  unsigned int g = 0;

  if (srate < afp->sratelo)
    g += afp->sratelo - srate;
  if (srate > afp->sratehi)
    g += srate - afp->sratehi;
  if (AFMT_ISSTEREO(afp->format) && !AFMT_ISSTEREO(fmt))
    g += 0x100000;
  if (!AFMT_ISSTEREO(afp->format) && AFMT_ISSTEREO(fmt))
    g += 0x400000;
  if (AFMT_IS16BIT(afp->format) && !AFMT_IS16BIT(fmt))
    g += 0x100000;
  if (!AFMT_IS16BIT(afp->format) && AFMT_IS16BIT(fmt))
    g += 0x400000;
  return g;
}

static int
find_format(struct audioformat *afp, unsigned int nr, unsigned int fmt,
	    unsigned int srate)
{
  unsigned int i, g, gb = ~0;
  int j = -1;			/* default to failure */

  /* find "best" format (according to format_goodness) */
  for (i = 0; i < nr; i++) {
    g = format_goodness(&afp[i], fmt, srate);
    if (g >= gb)
      continue;
    j = i;
    gb = g;
  }
  return j;
}

static int set_format_in(struct audiodev *as)
{
  struct usb_device *dev = as->state->usbdev;
  struct usb_config_descriptor *config = dev->actconfig;
  struct usb_interface_descriptor *alts;
  struct usb_interface *iface;
  struct usbin *u = &as->usbin;
  struct dmabuf *d = &u->dma;
  struct audioformat *fmt;
  unsigned int ep;
  unsigned char data[3];
  int fmtnr, ret;

  if (u->interface < 0 || u->interface >= config->bNumInterfaces)
    return 0;
  iface = &config->interface[u->interface];

  fmtnr = find_format(as->fmtin, as->numfmtin, d->format, d->srate);
  if (fmtnr < 0) {
    printk(KERN_ERR
	   "exaudio: set_format_in(): failed to find desired format/speed combination.\n");
    return -1;
  }

  fmt = as->fmtin + fmtnr;
  alts = &iface->altsetting[fmt->altsetting];
  u->format = fmt->format;
  u->datapipe =
    usb_rcvisocpipe(dev, alts->endpoint[0].bEndpointAddress & 0xf);
  u->syncpipe = u->syncinterval = 0;
  if ((alts->endpoint[0].bmAttributes & 0x0c) == 0x08) {
    if (alts->bNumEndpoints < 2 ||
	alts->endpoint[1].bmAttributes != 0x01 ||
	alts->endpoint[1].bSynchAddress != 0 ||
	alts->endpoint[1].bEndpointAddress !=
	(alts->endpoint[0].bSynchAddress & 0x7f)) {
      printk(KERN_ERR
	     "exaudio: device %d interface %d altsetting %d invalid synch pipe\n",
	     dev->devnum, u->interface, fmt->altsetting);
      return -1;
    }
    u->syncpipe =
      usb_sndisocpipe(dev, alts->endpoint[1].bEndpointAddress & 0xf);
    u->syncinterval = alts->endpoint[1].bRefresh;
  }
  if (d->srate < fmt->sratelo)
    d->srate = fmt->sratelo;
  if (d->srate > fmt->sratehi)
    d->srate = fmt->sratehi;
  dbg("set_format_in: usb_set_interface %u %u",
      alts->bInterfaceNumber,fmt->altsetting);
  if (usb_set_interface(dev, alts->bInterfaceNumber, fmt->altsetting) < 0) {
    printk(KERN_WARNING
	   "exaudio: usb_set_interface failed, device %d interface %d altsetting %d\n",
	   dev->devnum, u->interface, fmt->altsetting);
    return -1;
  }
  if (fmt->sratelo == fmt->sratehi)
    return 0;
  ep = usb_pipeendpoint(u->datapipe) | (u->datapipe & USB_DIR_IN);
  /* if endpoint has pitch control, enable it */
  if (fmt->attributes & 0x02) {
    data[0] = 1;
    if ((ret =
	 usb_control_msg(dev, usb_sndctrlpipe(dev, 0), SET_CUR,
			 USB_TYPE_CLASS | USB_RECIP_ENDPOINT | USB_DIR_OUT,
			 PITCH_CONTROL << 8, ep, data, 1, HZ)) < 0) {
      printk(KERN_ERR
	     "exaudio: failure (error %d) to set output pitch control device %d IF%u endpoint 0x%x to %u\n",
	     ret, dev->devnum, u->interface, ep, d->srate);
      return -1;
    }
  }
  /* if endpoint has sampling rate control, set it */
  if (fmt->attributes & 0x01) {
    data[0] = d->srate;
    data[1] = d->srate >> 8;
    data[2] = d->srate >> 16;
    if ((ret =
	 usb_control_msg(dev, usb_sndctrlpipe(dev, 0), SET_CUR,
			 USB_TYPE_CLASS | USB_RECIP_ENDPOINT | USB_DIR_OUT,
			 SAMPLING_FREQ_CONTROL << 8, ep, data, 3, HZ)) < 0) {
      printk(KERN_ERR
	     "exaudio: failure (error %d) to set input sampling frequency device %d IF%u endpoint 0x%x to %u\n",
	     ret, dev->devnum, u->interface, ep, d->srate);
      return -1;
    }
    if ((ret =
	 usb_control_msg(dev, usb_rcvctrlpipe(dev, 0), GET_CUR,
			 USB_TYPE_CLASS | USB_RECIP_ENDPOINT | USB_DIR_IN,
			 SAMPLING_FREQ_CONTROL << 8, ep, data, 3, HZ)) < 0) {
      printk(KERN_ERR
	     "exaudio: failure (error %d) to get input sampling frequency device %d IF%u endpoint 0x%x\n",
	     ret, dev->devnum, u->interface, ep);
      return -1;
    }
    printk(KERN_DEBUG
	   "exaudio: set_format_in: device %d interface %d altsetting %d srate req: %u real %u\n",
	   dev->devnum, u->interface, fmt->altsetting, d->srate,
	   data[0] | (data[1] << 8) | (data[2] << 16));
    d->srate = data[0] | (data[1] << 8) | (data[2] << 16);
  }
  printk(KERN_DEBUG
	 "exaudio: set_format_in: USB format 0x%x, DMA format 0x%x srate %u\n",
	 u->format, d->format, d->srate);
  return 0;
}

static int set_format_out(struct audiodev *as)
{
  struct usb_device *dev = as->state->usbdev;
  struct usb_config_descriptor *config = dev->actconfig;
  struct usb_interface_descriptor *alts;
  struct usb_interface *iface;
  struct usbout *u = &as->usbout;
  struct dmabuf *d = &u->dma;
  struct audioformat *fmt;
  unsigned int ep;
  unsigned char data[3];
  int fmtnr, ret;

  if (u->interface < 0 || u->interface >= config->bNumInterfaces)
    return 0;
  iface = &config->interface[u->interface];

  fmtnr = find_format(as->fmtout, as->numfmtout, d->format, d->srate);
  if (fmtnr < 0) {
    printk(KERN_ERR
	   "exaudio: set_format_out(): failed to find desired format/speed combination.\n");
    return -1;
  }

  fmt = as->fmtout + fmtnr;
  u->format = fmt->format;
  alts = &iface->altsetting[fmt->altsetting];
  u->datapipe =
    usb_sndisocpipe(dev, alts->endpoint[0].bEndpointAddress & 0xf);
  u->syncpipe = u->syncinterval = 0;
  if ((alts->endpoint[0].bmAttributes & 0x0c) == 0x04) {
#if 0
    printk(KERN_DEBUG
	   "bNumEndpoints 0x%02x endpoint[1].bmAttributes 0x%02x\n"
	   KERN_DEBUG
	   "endpoint[1].bSynchAddress 0x%02x endpoint[1].bEndpointAddress 0x%02x\n"
	   KERN_DEBUG "endpoint[0].bSynchAddress 0x%02x\n",
	   alts->bNumEndpoints, alts->endpoint[1].bmAttributes,
	   alts->endpoint[1].bSynchAddress,
	   alts->endpoint[1].bEndpointAddress,
	   alts->endpoint[0].bSynchAddress);
#endif
    if (alts->bNumEndpoints < 2 ||
	alts->endpoint[1].bmAttributes != 0x01 ||
	alts->endpoint[1].bSynchAddress != 0 ||
	alts->endpoint[1].bEndpointAddress !=
	(alts->endpoint[0].bSynchAddress | 0x80)) {
      printk(KERN_ERR
	     "exaudio: device %d interface %d altsetting %d invalid synch pipe\n",
	     dev->devnum, u->interface, fmt->altsetting);
      return -1;
    }
    u->syncpipe =
      usb_rcvisocpipe(dev, alts->endpoint[1].bEndpointAddress & 0xf);
    u->syncinterval = alts->endpoint[1].bRefresh;
  }
  if (d->srate < fmt->sratelo)
    d->srate = fmt->sratelo;
  if (d->srate > fmt->sratehi)
    d->srate = fmt->sratehi;
  dbg("set_format_out: usb_set_interface %u %u",alts->bInterfaceNumber,
      fmt->altsetting);
  if (usb_set_interface(dev, u->interface, fmt->altsetting) < 0) {
    printk(KERN_WARNING
	   "exaudio: usb_set_interface failed, device %d interface %d altsetting %d\n",
	   dev->devnum, u->interface, fmt->altsetting);
    return -1;
  }
  if (fmt->sratelo == fmt->sratehi)
    return 0;
  ep = usb_pipeendpoint(u->datapipe) | (u->datapipe & USB_DIR_IN);
  /* if endpoint has pitch control, enable it */
  if (fmt->attributes & 0x02) {
    data[0] = 1;
    if ((ret =
	 usb_control_msg(dev, usb_sndctrlpipe(dev, 0), SET_CUR,
			 USB_TYPE_CLASS | USB_RECIP_ENDPOINT | USB_DIR_OUT,
			 PITCH_CONTROL << 8, ep, data, 1, HZ)) < 0) {
      printk(KERN_ERR
	     "exaudio: failure (error %d) to set output pitch control device %d IF%u endpoint 0x%x to %u\n",
	     ret, dev->devnum, u->interface, ep, d->srate);
      return -1;
    }
  }
  /* if endpoint has sampling rate control, set it */
  if (fmt->attributes & 0x01) {
    data[0] = d->srate;
    data[1] = d->srate >> 8;
    data[2] = d->srate >> 16;
    if ((ret =
	 usb_control_msg(dev, usb_sndctrlpipe(dev, 0), SET_CUR,
			 USB_TYPE_CLASS | USB_RECIP_ENDPOINT | USB_DIR_OUT,
			 SAMPLING_FREQ_CONTROL << 8, ep, data, 3, HZ)) < 0) {
      printk(KERN_ERR
	     "exaudio: failure (error %d) to set output sampling frequency device %d IF%u endpoint 0x%x to %u\n",
	     ret, dev->devnum, u->interface, ep, d->srate);
      return -1;
    }
    if ((ret =
	 usb_control_msg(dev, usb_rcvctrlpipe(dev, 0), GET_CUR,
			 USB_TYPE_CLASS | USB_RECIP_ENDPOINT | USB_DIR_IN,
			 SAMPLING_FREQ_CONTROL << 8, ep, data, 3, HZ)) < 0) {
      printk(KERN_ERR
	     "exaudio: failure (error %d) to get output sampling frequency device %d IF%u endpoint 0x%x\n",
	     ret, dev->devnum, u->interface, ep);
      return -1;
    }
    printk(KERN_DEBUG
	   "exaudio: set_format_out: device %d interface %d altsetting %d srate req: %u real %u\n",
	   dev->devnum, u->interface, fmt->altsetting, d->srate,
	   data[0] | (data[1] << 8) | (data[2] << 16));
    d->srate = data[0] | (data[1] << 8) | (data[2] << 16);
  }
  printk(KERN_DEBUG
	 "exaudio: set_format_out: USB format 0x%x, DMA format 0x%x srate %u\n",
	 u->format, d->format, d->srate);
  return 0;
}

static int
set_format(struct audiodev *s, unsigned int fmode, unsigned int fmt,
	   unsigned int srate)
{
  int ret1 = 0, ret2 = 0;

  if (!(fmode & (FMODE_READ | FMODE_WRITE)))
    return -EINVAL;
  if (fmode & FMODE_READ) {
    usbin_stop(s);
    s->usbin.dma.ready = 0;
    if (fmt == AFMT_QUERY)
      fmt = s->usbin.dma.format;
    else
      s->usbin.dma.format = fmt;
    if (!srate)
      srate = s->usbin.dma.srate;
    else
      s->usbin.dma.srate = srate;
  }
  if (fmode & FMODE_WRITE) {
    usbout_stop(s);
    s->usbout.dma.ready = 0;
    if (fmt == AFMT_QUERY)
      fmt = s->usbout.dma.format;
    else
      s->usbout.dma.format = fmt;
    if (!srate)
      srate = s->usbout.dma.srate;
    else
      s->usbout.dma.srate = srate;
  }
  if (fmode & FMODE_READ)
    ret1 = set_format_in(s);
  if (fmode & FMODE_WRITE)
    ret2 = set_format_out(s);
  return ret1 ? ret1 : ret2;
}

/*----------------------------------------------------------------------------*/
static int wrmixer(struct mixerdev* ms,struct mixerchannel* ch,
		   unsigned value)
{ struct usb_device* dev = ms->state->usbdev;
  int err, range, v1, v2;
  short usbval;
  unsigned char buf[2];
  double step;

  v1 = value & 0xFF;
  v2 = (value >> 8) & 0xFF;

  if (ch->mutectl) {
    int setmute = FALSE;
    if (v1 > 0 || v2 > 0) {
      if (0 == ch->value) { /* Unmute the control. */
	buf[0] = 0;
	setmute = TRUE;
      }
    } else {
      if (ch->value > 0) { /* Mute the control */
	buf[0] = 1;
	setmute = TRUE;
      }
    }
    if (setmute) {
      err = usb_control_msg(dev,usb_sndctrlpipe(dev,0),SET_CUR,
			    USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_OUT,
			    (MUTE_CONTROL<<8)|ch->leftch,
			    (ch->mutectl<<8)|ms->iface,
			    buf,1,HZ);
      if (err < 0)
	err("cannot set mute on unit%u",ch->mutectl);
      if (ch->rightch) {
	err = usb_control_msg(dev,usb_sndctrlpipe(dev,0),SET_CUR,
			      USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_OUT,
			      (MUTE_CONTROL<<8)|ch->rightch,
			      (ch->mutectl<<8)|ms->iface,
			      buf,1,HZ);
	if (err < 0)
	  err("cannot set mute on unit%u",ch->mutectl);
      }
    }
  }

  if (ch->volctl) {
    range = ch->max - ch->min + 1;
    step = range/100.0;
    usbval = ch->min + (int)(v1*step);
    buf[0] = usbval & 0xFF;
    buf[1] = (usbval>>8) & 0xFF;
    err = usb_control_msg(dev,usb_sndctrlpipe(dev,0),SET_CUR,
			  USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_OUT,
			  (VOLUME_CONTROL<<8)|ch->leftch,
			  (ch->volctl<<8)|ms->iface,
			  buf,2,HZ);
    if (err < 0)
      err("cannot set VOL on unit%u",ch->volctl);
    if (ch->rightch) {
      usbval = ch->min + (int)(v2*step);
      buf[0] = usbval & 0xFF;
      buf[1] = (usbval>>8) & 0xFF;
      err = usb_control_msg(dev,usb_sndctrlpipe(dev,0),SET_CUR,
			    USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_OUT,
			    (VOLUME_CONTROL<<8)|ch->rightch,
			    (ch->volctl<<8)|ms->iface,
			    buf,2,HZ);
      if (err < 0) err("exaudio: cannot set VOL on unit%u",ch->volctl);
    }
    ch->value = value;  /* Remember OSS value. */
  }

  return(0);
}

/*----------------------------------------------------------------------------*/
static int get_rec_src(struct mixerdev *ms)
{ struct usb_device *dev = ms->state->usbdev;
  unsigned int mask = 0;
  unsigned char val;
  int err;

  switch (ms->type) {
  case RECORD:
    err = usb_control_msg(dev,usb_rcvctrlpipe(dev,0),GET_CUR,
			  USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_IN,
			  0,ms->iface|(ms->selector<<8),&val,1,HZ);
    if (err < 0) {
      err("selector read request DEV%u if %u unit %u failed",
	  dev->devnum,ms->iface,ms->selector&0xff);
      return(-EIO);
    } else {
      switch (val) {
      case 1: mask = 1<<SOUND_MIXER_LINE; break;
      case 2: mask = 1<<SOUND_MIXER_MIC;  break;
      case 3: mask = 1<<SOUND_MIXER_IMIX; break;
      default:
      }
      return(mask);
    }
  case PLAYBACK: return(0);
  default:       return(-EINVAL);
  }
}

/*----------------------------------------------------------------------------*/
static int set_rec_src(struct mixerdev *ms, int srcmask)
{ struct usb_device *dev = ms->state->usbdev;
  unsigned char val = 1;
  int err;

  switch (ms->type) {
  case RECORD:
    if      (srcmask & (1<<SOUND_MIXER_LINE)) val = 1;
    else if (srcmask & (1<<SOUND_MIXER_MIC))  val = 2;
    else if (srcmask & (1<<SOUND_MIXER_IMIX)) val = 3;
    err = usb_control_msg(dev,usb_sndctrlpipe(dev,0),SET_CUR,
			  USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_OUT,
			  0,ms->iface|(ms->selector<<8),&val,1,HZ);
    if (err < 0) {
      err("selector write request DEV%u IF%u unit%u failed",
	  dev->devnum,ms->iface,ms->selector);
      return(-EIO);
    }
    info("set recording selector to pin %u",val);
    return(0);
  case PLAYBACK:
  default:
    return(-EINVAL);
  }
}

/*----------------------------------------------------------------------------*/
static int mixer_ioctl(struct inode *inode,struct file *file,
		       unsigned int cmd,unsigned long arg)
{ struct mixerdev *ms = (struct mixerdev*)file->private_data;
  int i, j, val;

  if (!ms->state->usbdev) return(-ENODEV);
  if (cmd == SOUND_MIXER_INFO) {
    mixer_info info;
    strncpy(info.id,"EXAUDIO",sizeof(info.id));
    if (RECORD == ms->type)
      strncpy(info.name,"EXAUDIO Recording Mixer",sizeof(info.name));
    else
      strncpy(info.name,"EXAUDIO Playback Mixer",sizeof(info.name));
    info.modify_counter = ms->modcnt;
    if (copy_to_user((void*)arg,&info,sizeof(info))) return(-EFAULT);
    return(0);
  }
  if (cmd == SOUND_OLD_MIXER_INFO) {
    _old_mixer_info info;
    strncpy(info.id,"EXAUDIO",sizeof(info.id));
    if (RECORD == ms->type)
      strncpy(info.name,"EXAUDIO Recording Mixer",sizeof(info.name));
    else
      strncpy(info.name,"EXAUDIO Playback Mixer",sizeof(info.name));
    if (copy_to_user((void*)arg,&info,sizeof(info))) return(-EFAULT);
    return(0);
  }
  if (cmd == OSS_GETVERSION) return(put_user(SOUND_VERSION,(int*)arg));
  if (_IOC_TYPE(cmd) != 'M' || _IOC_SIZE(cmd) != sizeof(int)) return(-EINVAL);

  if (_IOC_DIR(cmd) == _IOC_READ) {
    switch (_IOC_NR(cmd)) {

    case SOUND_MIXER_RECSRC: /* Set a bit for the current recording source. */
      val = get_rec_src(ms);
      dbg("mixer RECSRC request, reply %04X",val);
      if (val < 0) return(val);
      return(put_user(val,(int*)arg));

    case SOUND_MIXER_DEVMASK: /* Set a bit for each supported device. */
      for (val = i = 0; i < ms->numch; i++) val |= 1 << ms->ch[i].osschannel;
      dbg("mixer DEVMASK request, reply %04X",val);
      return(put_user(val,(int*)arg));

    case SOUND_MIXER_RECMASK: /* a bit for each supported recording source */
      if (RECORD == ms->type) {
	val =
	  1<<ms->ch[0].osschannel |
	  1<<ms->ch[1].osschannel |
	  1<<ms->ch[2].osschannel;
      } else {
	val = 0;
      }
      dbg("mixer RECMASK request, reply %04X",val);
      return(put_user(val,(int*)arg));

    case SOUND_MIXER_STEREODEVS: /* Mixer channels supporting stereo */
      for (val = i = 0; i < ms->numch; i++)
	if (ms->ch[i].leftch) val |= 1 << ms->ch[i].osschannel;
      dbg("mixer STEREODEVS request, reply %04X",val);
      return(put_user(val,(int*)arg));

    case SOUND_MIXER_CAPS: /* We only allow one recording source at a time. */
      return(put_user(SOUND_CAP_EXCL_INPUT,(int*)arg));

    default: /* Read channel's current OSS value. */
      i = _IOC_NR(cmd);
      if (i >= SOUND_MIXER_NRDEVICES) return(-EINVAL);
      for (j = 0; j < ms->numch; j++)
	if (ms->ch[j].osschannel == i) {
	  return(put_user(ms->ch[j].value,(int*)arg));
	}
      return(-EINVAL);
    }
  }

  if (_IOC_DIR(cmd) != (_IOC_READ|_IOC_WRITE)) return(-EINVAL);
  ms->modcnt++;
  switch (_IOC_NR(cmd)) {
  case SOUND_MIXER_RECSRC: /* Arg contains a bit for each recording source */
    if (get_user(val,(int*)arg)) return(-EFAULT);
    return(set_rec_src(ms,val));

  default: /* Set the channel value. */
    i = _IOC_NR(cmd);
    if (i >= SOUND_MIXER_NRDEVICES) return(-EINVAL);
    for (j = 0; j < ms->numch && ms->ch[j].osschannel != i; j++)
      ;
    if (j >= ms->numch)             return(-EINVAL);
    if (get_user(val,(int*)arg))    return(-EFAULT);
    if (wrmixer(ms,&ms->ch[j],val)) return(-EIO);
    return(put_user(ms->ch[j].value,(int*)arg));
  }
}

/*----------------------------------------------------------------------------*/
/*
 * should be called with open_sem hold, so that no new processes
 * look at the audio device to be destroyed
 */
void release(struct extigy *s)
{
  struct audiodev *as;
  struct mixerdev *ms;

  s->count--;
  if (s->count) {
    up(&open_sem);
    return;
  }
  up(&open_sem);
  wake_up(&open_wait);
  while (!list_empty(&s->audiolist)) {
    as = list_entry(s->audiolist.next, struct audiodev, list);
    list_del(&as->list);
    usbin_release(as);
    usbout_release(as);
    dmabuf_release(&as->usbin.dma);
    dmabuf_release(&as->usbout.dma);
    kfree(as);
  }
  while (!list_empty(&s->mixerlist)) {
    ms = list_entry(s->mixerlist.next, struct mixerdev, list);
    list_del(&ms->list);
    kfree(ms);
  }
  kfree(s->desc);
  kfree(s);
}

/*----------------------------------------------------------------------------*/
extern inline int prog_dmabuf_in(struct audiodev *as)
{
  usbin_stop(as);
  return dmabuf_init(&as->usbin.dma);
}

extern inline int prog_dmabuf_out(struct audiodev *as)
{
  usbout_stop(as);
  return dmabuf_init(&as->usbout.dma);
}

/*----------------------------------------------------------------------------*/
static loff_t exaudio_llseek(struct file *file, loff_t offset, int origin)
{
  return -ESPIPE;
}

/*----------------------------------------------------------------------------*/
static int open_mixer(struct inode *inode,struct file *file)
{ int minor = MINOR(inode->i_rdev);
  struct list_head *devs, *mdevs;
  struct mixerdev *ms;
  struct extigy *s;

  down(&open_sem);
  for (devs = audiodevs.next; devs != &audiodevs; devs = devs->next) {
    s = list_entry(devs, struct extigy, audiodev);
    for (mdevs = s->mixerlist.next; mdevs != &s->mixerlist;
	 mdevs = mdevs->next) {
      ms = list_entry(mdevs, struct mixerdev, list);
      if (ms->ossmixer == minor)
	goto mixer_found;
    }
  }
  up(&open_sem);
  return -ENODEV;

mixer_found:
  if (!s->usbdev) {
    up(&open_sem);
    return -EIO;
  }
  file->private_data = ms;
  s->count++;
  up(&open_sem);
  return 0;
}

/*----------------------------------------------------------------------------*/
static int release_mixer(struct inode *inode, struct file *file)
{ struct mixerdev *ms = (struct mixerdev*)file->private_data;
  struct extigy *s;
  lock_kernel();
  s = ms->state;
  down(&open_sem);
  release(s);
  unlock_kernel();
  return 0;
}

static struct file_operations exaudio_mixer_fops = {
  owner:   THIS_MODULE,
  llseek:  exaudio_llseek,
  ioctl:   mixer_ioctl,
  open:    open_mixer,
  release: release_mixer,
};

/*----------------------------------------------------------------------------*/
static int drain_out(struct audiodev *as, int nonblock)
{
  DECLARE_WAITQUEUE(wait, current);
  unsigned long flags;
  int count, tmo;

  if (as->usbout.dma.mapped || !as->usbout.dma.ready)
    return 0;
  usbout_start(as);
  add_wait_queue(&as->usbout.dma.wait, &wait);
  for (;;) {
    __set_current_state(TASK_INTERRUPTIBLE);
    spin_lock_irqsave(&as->lock, flags);
    count = as->usbout.dma.count;
    spin_unlock_irqrestore(&as->lock, flags);
    if (count <= 0)
      break;
    if (signal_pending(current))
      break;
    if (nonblock) {
      remove_wait_queue(&as->usbout.dma.wait, &wait);
      set_current_state(TASK_RUNNING);
      return -EBUSY;
    }
    tmo = 3 * HZ * count / as->usbout.dma.srate;
    tmo >>= AFMT_BYTESSHIFT(as->usbout.dma.format);
    if (!schedule_timeout(tmo + 1)) {
      printk(KERN_DEBUG "exaudio: dma timed out??\n");
      break;
    }
  }
  remove_wait_queue(&as->usbout.dma.wait, &wait);
  set_current_state(TASK_RUNNING);
  if (signal_pending(current))
    return -ERESTARTSYS;
  return 0;
}

/*----------------------------------------------------------------------------*/
static ssize_t exaudio_read(struct file *file,char *buffer,
			    size_t count,loff_t *ppos)
{
  struct audiodev *as = (struct audiodev *) file->private_data;
  DECLARE_WAITQUEUE(wait, current);
  ssize_t ret = 0;
  unsigned long flags;
  unsigned int ptr;
  int cnt, err;

  if (ppos != &file->f_pos)
    return -ESPIPE;
  if (as->usbin.dma.mapped)
    return -ENXIO;
  if (!as->usbin.dma.ready && (ret = prog_dmabuf_in(as)))
    return ret;
  if (!access_ok(VERIFY_WRITE, buffer, count))
    return -EFAULT;
  add_wait_queue(&as->usbin.dma.wait, &wait);
  while (count > 0) {
    spin_lock_irqsave(&as->lock, flags);
    ptr = as->usbin.dma.rdptr;
    cnt = as->usbin.dma.count;
    /* set task state early to avoid wakeup races */
    if (cnt <= 0)
      __set_current_state(TASK_INTERRUPTIBLE);
    spin_unlock_irqrestore(&as->lock, flags);
    if (cnt > count)
      cnt = count;
    if (cnt <= 0) {
      if (usbin_start(as)) {
	if (!ret)
	  ret = -ENODEV;
	break;
      }
      if (file->f_flags & O_NONBLOCK) {
	if (!ret)
	  ret = -EAGAIN;
	break;
      }
      schedule();
      if (signal_pending(current)) {
	if (!ret)
	  ret = -ERESTARTSYS;
	break;
      }
      continue;
    }
    if ((err = dmabuf_copyout_user(&as->usbin.dma, ptr, buffer, cnt))) {
      if (!ret)
	ret = err;
      break;
    }
    ptr += cnt;
    if (ptr >= as->usbin.dma.dmasize)
      ptr -= as->usbin.dma.dmasize;
    spin_lock_irqsave(&as->lock, flags);
    as->usbin.dma.rdptr = ptr;
    as->usbin.dma.count -= cnt;
    spin_unlock_irqrestore(&as->lock, flags);
    count -= cnt;
    buffer += cnt;
    ret += cnt;
  }
  __set_current_state(TASK_RUNNING);
  remove_wait_queue(&as->usbin.dma.wait, &wait);
  return ret;
}

static ssize_t exaudio_write(struct file *file,const char *buffer,
			     size_t count,loff_t *ppos)
{
  struct audiodev *as = (struct audiodev *) file->private_data;
  DECLARE_WAITQUEUE(wait, current);
  ssize_t ret = 0;
  unsigned long flags;
  unsigned int ptr;
  unsigned int start_thr;
  int cnt, err;

  if (ppos != &file->f_pos)
    return -ESPIPE;
  if (as->usbout.dma.mapped)
    return -ENXIO;
  if (!as->usbout.dma.ready && (ret = prog_dmabuf_out(as)))
    return ret;
  if (!access_ok(VERIFY_READ, buffer, count))
    return -EFAULT;
  start_thr =
    (as->usbout.dma.srate << AFMT_BYTESSHIFT(as->usbout.dma.format)) /
    (1000 / (3 * DESCFRAMES));
  add_wait_queue(&as->usbout.dma.wait, &wait);
  while (count > 0) {
#if 0
    printk(KERN_DEBUG
	   "exaudio_write: count %u dma: count %u rdptr %u wrptr %u dmasize %u fragsize %u flags 0x%02x taskst 0x%lx\n",
	   count, as->usbout.dma.count, as->usbout.dma.rdptr,
	   as->usbout.dma.wrptr, as->usbout.dma.dmasize,
	   as->usbout.dma.fragsize, as->usbout.flags, current->state);
#endif
    spin_lock_irqsave(&as->lock, flags);
    if (as->usbout.dma.count < 0) {
      as->usbout.dma.count = 0;
      as->usbout.dma.rdptr = as->usbout.dma.wrptr;
    }
    ptr = as->usbout.dma.wrptr;
    cnt = as->usbout.dma.dmasize - as->usbout.dma.count;
    /* set task state early to avoid wakeup races */
    if (cnt <= 0)
      __set_current_state(TASK_INTERRUPTIBLE);
    spin_unlock_irqrestore(&as->lock, flags);
    if (cnt > count)
      cnt = count;
    if (cnt <= 0) {
      if (usbout_start(as)) {
	if (!ret)
	  ret = -ENODEV;
	break;
      }
      if (file->f_flags & O_NONBLOCK) {
	if (!ret)
	  ret = -EAGAIN;
	break;
      }
      schedule();
      if (signal_pending(current)) {
	if (!ret)
	  ret = -ERESTARTSYS;
	break;
      }
      continue;
    }
    if ((err = dmabuf_copyin_user(&as->usbout.dma, ptr, buffer, cnt))) {
      if (!ret)
	ret = err;
      break;
    }
    ptr += cnt;
    if (ptr >= as->usbout.dma.dmasize)
      ptr -= as->usbout.dma.dmasize;
    spin_lock_irqsave(&as->lock, flags);
    as->usbout.dma.wrptr = ptr;
    as->usbout.dma.count += cnt;
    spin_unlock_irqrestore(&as->lock, flags);
    count -= cnt;
    buffer += cnt;
    ret += cnt;
    if (as->usbout.dma.count >= start_thr && usbout_start(as)) {
      if (!ret)
	ret = -ENODEV;
      break;
    }
  }
  __set_current_state(TASK_RUNNING);
  remove_wait_queue(&as->usbout.dma.wait, &wait);
  return ret;
}

/* Called without the kernel lock - fine */
static unsigned int exaudio_poll(struct file *file,
				 struct poll_table_struct *wait)
{
  struct audiodev *as = (struct audiodev *) file->private_data;
  unsigned long flags;
  unsigned int mask = 0;

  if (file->f_mode & FMODE_WRITE) {
    if (!as->usbout.dma.ready)
      prog_dmabuf_out(as);
    poll_wait(file, &as->usbout.dma.wait, wait);
  }
  if (file->f_mode & FMODE_READ) {
    if (!as->usbin.dma.ready)
      prog_dmabuf_in(as);
    poll_wait(file, &as->usbin.dma.wait, wait);
  }
  spin_lock_irqsave(&as->lock, flags);
  if (file->f_mode & FMODE_READ) {
    if (as->usbin.dma.count >= (signed) as->usbin.dma.fragsize)
      mask |= POLLIN | POLLRDNORM;
  }
  if (file->f_mode & FMODE_WRITE) {
    if (as->usbout.dma.mapped) {
      if (as->usbout.dma.count >= (signed) as->usbout.dma.fragsize)
	mask |= POLLOUT | POLLWRNORM;
    } else {
      if ((signed) as->usbout.dma.dmasize >=
	  as->usbout.dma.count + (signed) as->usbout.dma.fragsize)
	mask |= POLLOUT | POLLWRNORM;
    }
  }
  spin_unlock_irqrestore(&as->lock, flags);
  return mask;
}

static int exaudio_mmap(struct file *file, struct vm_area_struct *vma)
{
  struct audiodev *as = (struct audiodev *) file->private_data;
  struct dmabuf *db;
  int ret = -EINVAL;

  lock_kernel();
  if (vma->vm_flags & VM_WRITE) {
    if ((ret = prog_dmabuf_out(as)) != 0)
      goto out;
    db = &as->usbout.dma;
  } else if (vma->vm_flags & VM_READ) {
    if ((ret = prog_dmabuf_in(as)) != 0)
      goto out;
    db = &as->usbin.dma;
  } else
    goto out;

  ret = -EINVAL;
  if (vma->vm_pgoff != 0)
    goto out;

  ret =
    dmabuf_mmap(db, vma->vm_start, vma->vm_end - vma->vm_start,
		vma->vm_page_prot);
out:
  unlock_kernel();
  return ret;
}

static int exaudio_ioctl(struct inode *inode,struct file *file,
			 unsigned int cmd,unsigned long arg)
{
  struct audiodev *as = (struct audiodev *) file->private_data;
  struct extigy *s = as->state;
  unsigned long flags;
  audio_buf_info abinfo;
  count_info cinfo;
  int val, val2, mapped, ret;

  if (!s->usbdev)
    return -EIO;
  mapped = ((file->f_mode & FMODE_WRITE) && as->usbout.dma.mapped) ||
    ((file->f_mode & FMODE_READ) && as->usbin.dma.mapped);
  switch (cmd) {
  case OSS_GETVERSION:
    return put_user(SOUND_VERSION, (int *) arg);

  case SNDCTL_DSP_SYNC:
    if (file->f_mode & FMODE_WRITE)
      return drain_out(as, 0 /*file->f_flags & O_NONBLOCK */ );
    return 0;

  case SNDCTL_DSP_SETDUPLEX:
    return 0;

  case SNDCTL_DSP_GETCAPS:
    return put_user(DSP_CAP_DUPLEX | DSP_CAP_REALTIME | DSP_CAP_TRIGGER |
		    DSP_CAP_MMAP | DSP_CAP_BATCH, (int *) arg);

  case SNDCTL_DSP_RESET:
    if (file->f_mode & FMODE_WRITE) {
      usbout_stop(as);
      as->usbout.dma.rdptr = as->usbout.dma.wrptr = as->usbout.dma.count =
	as->usbout.dma.total_bytes = 0;
    }
    if (file->f_mode & FMODE_READ) {
      usbin_stop(as);
      as->usbin.dma.rdptr = as->usbin.dma.wrptr = as->usbin.dma.count =
	as->usbin.dma.total_bytes = 0;
    }
    return 0;

  case SNDCTL_DSP_SPEED:
    if (get_user(val, (int *) arg))
      return -EFAULT;
    if (val >= 0) {
      if (val < 4000)
	val = 4000;
      if (val > 100000)
	val = 100000;
      if (set_format(as, file->f_mode, AFMT_QUERY, val))
	return -EIO;
    }
    return put_user((file->f_mode & FMODE_READ) ? as->usbin.dma.
		    srate : as->usbout.dma.srate, (int *) arg);

  case SNDCTL_DSP_STEREO:
    val2 =
      (file->f_mode & FMODE_READ) ? as->usbin.dma.format : as->usbout.dma.
      format;
    if (set_format(as, file->f_mode, val2 | AFMT_STEREO, 0))
      return -EIO;
    return 0;

  case SNDCTL_DSP_CHANNELS:
    if (get_user(val, (int *) arg))
      return -EFAULT;
    if (val != 0) {
      val2 =
	(file->f_mode & FMODE_READ) ? as->usbin.dma.format : as->usbout.
	dma.format;
      if (val == 1)
	val2 &= ~AFMT_STEREO;
      else
	val2 |= AFMT_STEREO;
      if (set_format(as, file->f_mode, val2, 0))
	return -EIO;
    }
    val2 =
      (file->f_mode & FMODE_READ) ? as->usbin.dma.format : as->usbout.dma.
      format;
    return put_user(AFMT_ISSTEREO(val2) ? 2 : 1, (int *) arg);

  case SNDCTL_DSP_GETFMTS:	/* Returns a mask */
    return put_user(AFMT_U8 | AFMT_U16_LE | AFMT_U16_BE |
		    AFMT_S8 | AFMT_S16_LE | AFMT_S16_BE, (int *) arg);

  case SNDCTL_DSP_SETFMT:	/* Selects ONE fmt */
    if (get_user(val, (int *) arg))
      return -EFAULT;
    if (val != AFMT_QUERY) {
      if (hweight32(val) != 1)
	return -EINVAL;
      if (!(val & (AFMT_U8 | AFMT_U16_LE | AFMT_U16_BE |
		   AFMT_S8 | AFMT_S16_LE | AFMT_S16_BE)))
	return -EINVAL;
      val2 =
	(file->f_mode & FMODE_READ) ? as->usbin.dma.format : as->usbout.
	dma.format;
      val |= val2 & AFMT_STEREO;
      if (set_format(as, file->f_mode, val, 0))
	return -EIO;
    }
    val2 =
      (file->f_mode & FMODE_READ) ? as->usbin.dma.format : as->usbout.dma.
      format;
    return put_user(val2 & ~AFMT_STEREO, (int *) arg);

  case SNDCTL_DSP_POST:
    return 0;

  case SNDCTL_DSP_GETTRIGGER:
    val = 0;
    if (file->f_mode & FMODE_READ && as->usbin.flags & FLG_RUNNING)
      val |= PCM_ENABLE_INPUT;
    if (file->f_mode & FMODE_WRITE && as->usbout.flags & FLG_RUNNING)
      val |= PCM_ENABLE_OUTPUT;
    return put_user(val, (int *) arg);

  case SNDCTL_DSP_SETTRIGGER:
    if (get_user(val, (int *) arg))
      return -EFAULT;
    if (file->f_mode & FMODE_READ) {
      if (val & PCM_ENABLE_INPUT) {
	if (!as->usbin.dma.ready && (ret = prog_dmabuf_in(as)))
	  return ret;
	if (usbin_start(as))
	  return -ENODEV;
      } else
	usbin_stop(as);
    }
    if (file->f_mode & FMODE_WRITE) {
      if (val & PCM_ENABLE_OUTPUT) {
	if (!as->usbout.dma.ready && (ret = prog_dmabuf_out(as)))
	  return ret;
	if (usbout_start(as))
	  return -ENODEV;
      } else
	usbout_stop(as);
    }
    return 0;

  case SNDCTL_DSP_GETOSPACE:
    if (!(file->f_mode & FMODE_WRITE))
      return -EINVAL;
    if (!(as->usbout.flags & FLG_RUNNING)
	&& (val = prog_dmabuf_out(as)) != 0)
      return val;
    spin_lock_irqsave(&as->lock, flags);
    abinfo.fragsize = as->usbout.dma.fragsize;
    abinfo.bytes = as->usbout.dma.dmasize - as->usbout.dma.count;
    abinfo.fragstotal = as->usbout.dma.numfrag;
    abinfo.fragments = abinfo.bytes >> as->usbout.dma.fragshift;
    spin_unlock_irqrestore(&as->lock, flags);
    return copy_to_user((void *) arg, &abinfo, sizeof(abinfo)) ? -EFAULT : 0;

  case SNDCTL_DSP_GETISPACE:
    if (!(file->f_mode & FMODE_READ))
      return -EINVAL;
    if (!(as->usbin.flags & FLG_RUNNING)
	&& (val = prog_dmabuf_in(as)) != 0)
      return val;
    spin_lock_irqsave(&as->lock, flags);
    abinfo.fragsize = as->usbin.dma.fragsize;
    abinfo.bytes = as->usbin.dma.count;
    abinfo.fragstotal = as->usbin.dma.numfrag;
    abinfo.fragments = abinfo.bytes >> as->usbin.dma.fragshift;
    spin_unlock_irqrestore(&as->lock, flags);
    return copy_to_user((void *) arg, &abinfo, sizeof(abinfo)) ? -EFAULT : 0;

  case SNDCTL_DSP_NONBLOCK:
    file->f_flags |= O_NONBLOCK;
    return 0;

  case SNDCTL_DSP_GETODELAY:
    if (!(file->f_mode & FMODE_WRITE))
      return -EINVAL;
    spin_lock_irqsave(&as->lock, flags);
    val = as->usbout.dma.count;
    spin_unlock_irqrestore(&as->lock, flags);
    return put_user(val, (int *) arg);

  case SNDCTL_DSP_GETIPTR:
    if (!(file->f_mode & FMODE_READ))
      return -EINVAL;
    spin_lock_irqsave(&as->lock, flags);
    cinfo.bytes = as->usbin.dma.total_bytes;
    cinfo.blocks = as->usbin.dma.count >> as->usbin.dma.fragshift;
    cinfo.ptr = as->usbin.dma.wrptr;
    if (as->usbin.dma.mapped)
      as->usbin.dma.count &= as->usbin.dma.fragsize - 1;
    spin_unlock_irqrestore(&as->lock, flags);
    return copy_to_user((void *) arg, &cinfo, sizeof(cinfo));

  case SNDCTL_DSP_GETOPTR:
    if (!(file->f_mode & FMODE_WRITE))
      return -EINVAL;
    spin_lock_irqsave(&as->lock, flags);
    cinfo.bytes = as->usbout.dma.total_bytes;
    cinfo.blocks = as->usbout.dma.count >> as->usbout.dma.fragshift;
    cinfo.ptr = as->usbout.dma.rdptr;
    if (as->usbout.dma.mapped)
      as->usbout.dma.count &= as->usbout.dma.fragsize - 1;
    spin_unlock_irqrestore(&as->lock, flags);
    return copy_to_user((void *) arg, &cinfo, sizeof(cinfo));

  case SNDCTL_DSP_GETBLKSIZE:
    if (file->f_mode & FMODE_WRITE) {
      if ((val = prog_dmabuf_out(as)))
	return val;
      return put_user(as->usbout.dma.fragsize, (int *) arg);
    }
    if ((val = prog_dmabuf_in(as)))
      return val;
    return put_user(as->usbin.dma.fragsize, (int *) arg);

  case SNDCTL_DSP_SETFRAGMENT:
    if (get_user(val, (int *) arg))
      return -EFAULT;
    if (file->f_mode & FMODE_READ) {
      as->usbin.dma.ossfragshift = val & 0xffff;
      as->usbin.dma.ossmaxfrags = (val >> 16) & 0xffff;
      if (as->usbin.dma.ossfragshift < 4)
	as->usbin.dma.ossfragshift = 4;
      if (as->usbin.dma.ossfragshift > 15)
	as->usbin.dma.ossfragshift = 15;
      if (as->usbin.dma.ossmaxfrags < 4)
	as->usbin.dma.ossmaxfrags = 4;
    }
    if (file->f_mode & FMODE_WRITE) {
      as->usbout.dma.ossfragshift = val & 0xffff;
      as->usbout.dma.ossmaxfrags = (val >> 16) & 0xffff;
      if (as->usbout.dma.ossfragshift < 4)
	as->usbout.dma.ossfragshift = 4;
      if (as->usbout.dma.ossfragshift > 15)
	as->usbout.dma.ossfragshift = 15;
      if (as->usbout.dma.ossmaxfrags < 4)
	as->usbout.dma.ossmaxfrags = 4;
    }
    return 0;

  case SNDCTL_DSP_SUBDIVIDE:
    if ((file->f_mode & FMODE_READ && as->usbin.dma.subdivision) ||
	(file->f_mode & FMODE_WRITE && as->usbout.dma.subdivision))
      return -EINVAL;
    if (get_user(val, (int *) arg))
      return -EFAULT;
    if (val != 1 && val != 2 && val != 4)
      return -EINVAL;
    if (file->f_mode & FMODE_READ)
      as->usbin.dma.subdivision = val;
    if (file->f_mode & FMODE_WRITE)
      as->usbout.dma.subdivision = val;
    return 0;

  case SOUND_PCM_READ_RATE:
    return put_user((file->f_mode & FMODE_READ) ? as->usbin.dma.
		    srate : as->usbout.dma.srate, (int *) arg);

  case SOUND_PCM_READ_CHANNELS:
    val2 =
      (file->f_mode & FMODE_READ) ? as->usbin.dma.format : as->usbout.dma.
      format;
    return put_user(AFMT_ISSTEREO(val2) ? 2 : 1, (int *) arg);

  case SOUND_PCM_READ_BITS:
    val2 =
      (file->f_mode & FMODE_READ) ? as->usbin.dma.format : as->usbout.dma.
      format;
    return put_user(AFMT_IS16BIT(val2) ? 16 : 8, (int *) arg);

  case SOUND_PCM_WRITE_FILTER:
  case SNDCTL_DSP_SETSYNCRO:
  case SOUND_PCM_READ_FILTER:
    return -EINVAL;
  }
  return -ENOIOCTLCMD;
}

static int exaudio_open(struct inode *inode, struct file *file)
{
  int minor = MINOR(inode->i_rdev);
  DECLARE_WAITQUEUE(wait, current);
  struct list_head *devs, *adevs;
  struct audiodev *as;
  struct extigy *s;

  for (;;) {
    down(&open_sem);
    for (devs = audiodevs.next; devs != &audiodevs; devs = devs->next) {
      s = list_entry(devs, struct extigy, audiodev);
      for (adevs = s->audiolist.next; adevs != &s->audiolist;
	   adevs = adevs->next) {
	as = list_entry(adevs, struct audiodev, list);
	if (!((as->ossdsp ^ minor) & ~0xf))
	  goto device_found;
      }
    }
    up(&open_sem);
    return -ENODEV;

  device_found:
    if (!s->usbdev) {
      up(&open_sem);
      return -EIO;
    }
    /* wait for device to become free */
    if (!(as->open_mode & file->f_mode))
      break;
    if (file->f_flags & O_NONBLOCK) {
      up(&open_sem);
      return -EBUSY;
    }
    __set_current_state(TASK_INTERRUPTIBLE);
    add_wait_queue(&open_wait, &wait);
    up(&open_sem);
    schedule();
    __set_current_state(TASK_RUNNING);
    remove_wait_queue(&open_wait, &wait);
    if (signal_pending(current))
      return -ERESTARTSYS;
  }
  if (file->f_mode & FMODE_READ)
    as->usbin.dma.ossfragshift = as->usbin.dma.ossmaxfrags =
      as->usbin.dma.subdivision = 0;
  if (file->f_mode & FMODE_WRITE)
    as->usbout.dma.ossfragshift = as->usbout.dma.ossmaxfrags =
      as->usbout.dma.subdivision = 0;
  if (set_format(as,file->f_mode,((minor&0xf) == SND_DEV_DSP16) ? 
		 AFMT_S16_LE:AFMT_U8/*AFMT_ULAW*/,8000)) {
    up(&open_sem);
    return -EIO;
  }
  file->private_data = as;
  as->open_mode |= file->f_mode & (FMODE_READ|FMODE_WRITE);
  s->count++;
  up(&open_sem);
  return 0;
}

static int exaudio_release(struct inode *inode, struct file *file)
{
  struct audiodev *as = (struct audiodev *) file->private_data;
  struct extigy *s;
  struct usb_device *dev;
  struct usb_interface *iface;

  lock_kernel();
  s = as->state;
  dev = s->usbdev;
  if (file->f_mode & FMODE_WRITE)
    drain_out(as, file->f_flags & O_NONBLOCK);
  down(&open_sem);
  if (file->f_mode & FMODE_WRITE) {
    usbout_stop(as);
    if (dev && as->usbout.interface >= 0) {
      iface = &dev->actconfig->interface[as->usbout.interface];
      usb_set_interface(dev, iface->altsetting->bInterfaceNumber, 0);
    }
    dmabuf_release(&as->usbout.dma);
    usbout_release(as);
  }
  if (file->f_mode & FMODE_READ) {
    usbin_stop(as);
    if (dev && as->usbin.interface >= 0) {
      iface = &dev->actconfig->interface[as->usbin.interface];
      usb_set_interface(dev, iface->altsetting->bInterfaceNumber, 0);
    }
    dmabuf_release(&as->usbin.dma);
    usbin_release(as);
  }
  as->open_mode &= (~file->f_mode) & (FMODE_READ | FMODE_WRITE);
  release(s);
  wake_up(&open_wait);
  unlock_kernel();
  return 0;
}

static struct file_operations exaudio_fops = {
  owner:   THIS_MODULE,
  llseek:  exaudio_llseek,
  read:    exaudio_read,
  write:   exaudio_write,
  poll:    exaudio_poll,
  ioctl:   exaudio_ioctl,
  mmap:    exaudio_mmap,
  open:    exaudio_open,
  release: exaudio_release,
};

/*----------------------------------------------------------------------------*/
static struct usb_device_id exaudio_ids[] = {
  { USB_DEVICE(0x041E,0x3000),
    bInterfaceClass:    USB_CLASS_AUDIO,
    bInterfaceSubClass: 1
  },
  {} /* Terminating entry */
};

MODULE_DEVICE_TABLE(usb,exaudio_ids);

static struct usb_driver exaudio_driver = {
  name:        "exaudio",
  probe:       exaudio_probe,
  disconnect:  exaudio_disconnect,
  driver_list: LIST_HEAD_INIT(exaudio_driver.driver_list),
  id_table:    exaudio_ids,
};

/*----------------------------------------------------------------------------*/
static void* find_desc(void *descstart,unsigned int desclen,void *after,
		       u8 dtype,int iface,int altsetting)
{
  u8 *p, *end, *next;
  int ifc = -1, as = -1;

  p = descstart;
  end = p + desclen;
  for (; p < end;) {
    if (p[0] < 2)
      return NULL;
    next = p + p[0];
    if (next > end)
      return NULL;
    if (p[1] == USB_DT_INTERFACE) {
      /* minimum length of interface descriptor */
      if (p[0] < 9)
	return NULL;
      ifc = p[2];
      as = p[3];
    }
    if ( p[1] == dtype &&
	 (!after || (void*) p > after) &&
	 (iface == -1 || iface == ifc) &&
	 (altsetting == -1 || altsetting == as) ) {
      return p;
    }
    p = next;
  }
  return NULL;
}

static void* find_csif_desc(void *descstart,unsigned int desclen,void *after,
			    u8 dsubtype,int iface,int altsetting)
{
  unsigned char *p;

  p = find_desc(descstart,desclen,after,USB_DT_CS_INTERFACE,iface,altsetting);
  while (p) {
    if (p[0] >= 3 && p[2] == dsubtype) return p;
    p = find_desc(descstart,desclen,p,USB_DT_CS_INTERFACE,iface,altsetting);
  }
  return NULL;
}

static void parse_streaming(struct extigy *s, unsigned char *buffer,
			    unsigned int buflen, int asifin, int asifout)
{
  struct usb_device *dev = s->usbdev;
  struct audiodev *as;
  struct usb_config_descriptor *config = dev->actconfig;
  struct usb_interface_descriptor *alts;
  struct usb_interface *iface;
  struct audioformat *fp;
  unsigned char *fmt, *csep;
  unsigned int i, j, k, format;

  if (!(as = kmalloc(sizeof(struct audiodev), GFP_KERNEL)))
    return;
  memset(as, 0, sizeof(struct audiodev));
  init_waitqueue_head(&as->usbin.dma.wait);
  init_waitqueue_head(&as->usbout.dma.wait);
  spin_lock_init(&as->lock);
  spin_lock_init(&as->usbin.durb[0].urb.lock);
  spin_lock_init(&as->usbin.durb[1].urb.lock);
  spin_lock_init(&as->usbin.surb[0].urb.lock);
  spin_lock_init(&as->usbin.surb[1].urb.lock);
  spin_lock_init(&as->usbout.durb[0].urb.lock);
  spin_lock_init(&as->usbout.durb[1].urb.lock);
  spin_lock_init(&as->usbout.surb[0].urb.lock);
  spin_lock_init(&as->usbout.surb[1].urb.lock);
  as->state = s;
  as->usbin.interface = asifin;
  as->usbout.interface = asifout;
  /* Search for input formats */
  if (asifin >= 0) {
    as->usbin.flags = FLG_CONNECTED;
    iface = &config->interface[asifin];
    for (i = 0; i < iface->num_altsetting; i++) {
      alts = &iface->altsetting[i];
      if (alts->bInterfaceClass != USB_CLASS_AUDIO
	  || alts->bInterfaceSubClass != 2)
	continue;
      if (alts->bNumEndpoints < 1) {
	//err("IF%u ALT%u does not have an endpoint",asifin,i);
	continue;
      }
      if ((alts->endpoint[0].bmAttributes & 0x03) != 0x01 ||
	  !(alts->endpoint[0].bEndpointAddress & 0x80)) {
	err("IF%u ALT%u first endpoint not isochronous in",asifin,i);
	continue;
      }
      fmt = find_csif_desc(buffer,buflen,NULL,AS_GENERAL,asifin,i);
      if (!fmt) {
	err("IF%u ALT%u FORMAT_TYPE descriptor not found",asifin,i);
	continue;
      }
      if (fmt[0] < 7 || fmt[6] != 0 || (fmt[5] != 1 && fmt[5] != 2)) {
	err("IF%u ALT%u format not supported",asifin,i);
	continue;
      }
      format = (fmt[5] == 2) ? (AFMT_U16_LE|AFMT_U8) : (AFMT_S16_LE|AFMT_S8);
      fmt = find_csif_desc(buffer,buflen,NULL,FORMAT_TYPE,asifin,i);
      if (!fmt) {
	err("IF%u ALT%u FORMAT_TYPE descriptor not found",asifin,i);
	continue;
      }
      if (fmt[0] < 8 + 3 * (fmt[7] ? fmt[7] : 2) || fmt[3] != 1) {
	err("IF%u ALT%u FORMAT_TYPE descriptor not supported",asifin,i);
	continue;
      }
      if (fmt[4] < 1 || fmt[4] > 2 || fmt[5] < 1 || fmt[5] > 2) {
	err("IF%u ALT%u unsupported channels %u framesize %u",
	    asifin,i,fmt[4],fmt[5]);
	continue;
      }
      csep = find_desc(buffer,buflen,NULL,USB_DT_CS_ENDPOINT,asifin,i);
      if (!csep || csep[0] < 7 || csep[2] != EP_GENERAL) {
	err("IF%u ALT%u no or invalid class specific endpoint descriptor",
	    asifin,i);
	continue;
      }
      if (as->numfmtin >= MAXFORMATS) continue;
      fp = &as->fmtin[as->numfmtin++];
      if (fmt[5] == 2) format &= (AFMT_U16_LE | AFMT_S16_LE);
      else             format &= (AFMT_U8 | AFMT_S8);
      if (fmt[4] == 2) format |= AFMT_STEREO;
      fp->format = format;
      fp->altsetting = i;
      fp->sratelo = fp->sratehi = fmt[8] | (fmt[9] << 8) | (fmt[10] << 16);
      for (j = fmt[7] ? (fmt[7] - 1) : 1; j > 0; j--) {
	k = fmt[8 + 3 * j] | (fmt[9 + 3 * j] << 8) | (fmt[10 + 3 * j] << 16);
	if (k > fp->sratehi)
	  fp->sratehi = k;
	if (k < fp->sratelo)
	  fp->sratelo = k;
      }
      fp->attributes = csep[3];
      dbg("IF%u ALT%u: format 0x%08x sample rate %u - %u",
	  asifin,i,fp->format,fp->sratelo,fp->sratehi);
    }
  }
  /* Search for output formats */
  if (asifout >= 0) {
    as->usbout.flags = FLG_CONNECTED;
    iface = &config->interface[asifout];
    for (i = 0; i < iface->num_altsetting; i++) {
      alts = &iface->altsetting[i];
      if (alts->bInterfaceClass != USB_CLASS_AUDIO
	  || alts->bInterfaceSubClass != 2)
	continue;
      if (alts->bNumEndpoints < 1) {
	//err("IF%u ALT%u does not have an endpoint",asifout,i);
	continue;
      }
      if ((alts->endpoint[0].bmAttributes & 0x03) != 0x01 ||
	  (alts->endpoint[0].bEndpointAddress & 0x80)) {
	err("IF%u ALT%u first endpoint not isochronous out",asifout,i);
	continue;
      }
      fmt = find_csif_desc(buffer,buflen,NULL,AS_GENERAL,asifout,i);
      if (!fmt) {
	err("IF%u ALT%u FORMAT_TYPE descriptor not found",asifout,i);
	continue;
      }
      if (fmt[0] < 7 || fmt[6] != 0 || (fmt[5] != 1 && fmt[5] != 2)) {
	dbg("IF%u ALT%u format not supported",asifout,i);
	continue;
      }
      format = (fmt[5] == 2) ? (AFMT_U16_LE|AFMT_U8) : (AFMT_S16_LE|AFMT_S8);
      fmt = find_csif_desc(buffer,buflen,NULL,FORMAT_TYPE,asifout,i);
      if (!fmt) {
	err("IF%u ALT%u FORMAT_TYPE descriptor not found",asifout,i);
	continue;
      }
      if (fmt[0] < 8 + 3 * (fmt[7] ? fmt[7] : 2) || fmt[3] != 1) {
	err("IF%u ALT%u FORMAT_TYPE descriptor not supported",asifout,i);
	continue;
      }
      if (fmt[4] < 1 || fmt[4] > 2 || fmt[5] < 1 || fmt[5] > 2) {
	err("IF%u ALT%u unsupported channels %u framesize %u",
	    asifout,i,fmt[4],fmt[5]);
	continue;
      }
      csep = find_desc(buffer,buflen,NULL,USB_DT_CS_ENDPOINT,asifout,i);
      if (!csep || csep[0] < 7 || csep[2] != EP_GENERAL) {
	err("IF%u ALT%u no or invalid class specific endpoint descriptor",
	    asifout,i);
	continue;
      }
      if (as->numfmtout >= MAXFORMATS) continue;
      fp = &as->fmtout[as->numfmtout++];
      if (fmt[5] == 2) format &= (AFMT_U16_LE | AFMT_S16_LE);
      else             format &= (AFMT_U8 | AFMT_S8);
      if (fmt[4] == 2) format |= AFMT_STEREO;
      fp->format = format;
      fp->altsetting = i; //TODO?
      fp->sratelo = fp->sratehi = fmt[8] | (fmt[9] << 8) | (fmt[10] << 16);
      for (j = fmt[7] ? (fmt[7] - 1) : 1; j > 0; j--) {
	k = fmt[8 + 3 * j] | (fmt[9 + 3 * j] << 8) | (fmt[10 + 3 * j] << 16);
	if (k > fp->sratehi)
	  fp->sratehi = k;
	if (k < fp->sratelo)
	  fp->sratelo = k;
      }
      fp->attributes = csep[3];
      dbg("IF%u ALT%u: format 0x%08x sample rate %u - %u",
	  asifout,i,fp->format,fp->sratelo,fp->sratehi);
    }
  }
  if (as->numfmtin == 0 && as->numfmtout == 0) {
    kfree(as);
    return;
  }
  as->ossdsp = register_sound_dsp(&exaudio_fops,-1);
  if (as->ossdsp < 0) {
    err("cannot register dsp");
    kfree(as);
    return;
  } else {
    INFO("registered dsp (major 14 minor %d)",as->ossdsp);
  }
  /* everything successful */
  list_add_tail(&as->list,&s->audiolist);
}

/*----------------------------------------------------------------------------*/
static int query_vol_range(struct usb_device* dev,
			   unsigned char unitid,unsigned int ctrlif,
			   struct mixerchannel* ch)
{ int err;
  short tmp;
  unsigned char buf[2];

  ch->min = ch->max = ch->step = 0;

  err = usb_control_msg(dev,usb_rcvctrlpipe(dev,0),GET_MIN,
			USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_IN,
			(VOLUME_CONTROL<<8)|0,
			(unitid<<8)|ctrlif,
			buf,2,HZ);
  if (err < 0) return(FALSE);
  tmp = buf[0] | (buf[1]<<8);
  dbg("unit%u min VOL 0x%04X %d",unitid,tmp&0xFFFF,tmp);
  ch->min = tmp;

  err = usb_control_msg(dev,usb_rcvctrlpipe(dev,0),GET_MAX,
			USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_IN,
			(VOLUME_CONTROL<<8)|0,
			(unitid<<8)|ctrlif,
			buf,2,HZ);
  if (err < 0) return(FALSE);
  tmp = buf[0] | (buf[1]<<8);
  dbg("unit%u max VOL 0x%04X %d",unitid,tmp&0xFFFF,tmp);
  ch->max = tmp;

  err = usb_control_msg(dev,usb_rcvctrlpipe(dev,0),GET_RES,
			USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_IN,
			(VOLUME_CONTROL<<8)|0,
			(unitid<<8)|ctrlif,
			buf,2,HZ);
  if (err < 0) return(FALSE);
  tmp = buf[0] | (buf[1]<<8);
  dbg("unit%u res VOL 0x%04X %d",unitid,tmp&0xFFFF,tmp);
  ch->step = tmp;

  return(TRUE);
}

/*----------------------------------------------------------------------------*/
static void init_mixch(struct mixerdev* ms,struct mixerchannel* ch,
		       __u16 osschan,
		       __u8 volctl,
		       __u8 mutectl)
{ struct usb_device* dev = ms->state->usbdev;

  ch->osschannel = osschan;
  ch->value      = 0;
  ch->volctl     = volctl;
  ch->mutectl    = mutectl;
  ch->leftch     = 0;
  ch->rightch    = 0;
  ch->min        = 0;
  ch->max        = 0;
  ch->step       = 1;
  if (volctl) {
    if (!query_vol_range(dev,volctl,ms->iface,ch)) {
      err("cannot query VOL ranges on u%u",volctl);
    }
  }
}

/*----------------------------------------------------------------------------*/
static void set_mixch_stereo(struct mixerchannel* ch,__u8 leftch,__u8 rightch)
{
  ch->leftch  = leftch;
  ch->rightch = rightch;
}

/*----------------------------------------------------------------------------*/

/* SET and GET, macros to quicky try command new ideas. */

#define SET(req,val,idx,d1,d2,len)					\
{									\
  unsigned char buf[2];							\
  int err;								\
  buf[0] = d1;								\
  buf[1] = d2;								\
  if (len < 3) {							\
    err = usb_control_msg(dev, usb_sndctrlpipe(dev,0), req,		\
			  USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_OUT,\
			  val, idx, buf, len, HZ);			\
    if (err < 0) err("cannot SET idx %hX : %d",idx,err);		\
  }									\
}

#define GET(req,val,idx,len)						\
{									\
  unsigned char buf[2];							\
  int err;								\
  buf[0] = buf[1] = 0;							\
  if (len < 3) {							\
    err = usb_control_msg(dev, usb_rcvctrlpipe(dev,0), req,		\
			  USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_IN,\
			  val, idx, buf, len, HZ);			\
    if (err < 0) err("cannot GET idx %hX : %d",idx,err);		\
    else err("GOT 0x%02hX%02hX from %hX",buf[1],buf[0],idx);		\
  }									\
}

static void construct_mixer(struct extigy *s,unsigned char *desc,
			    unsigned int len,unsigned int ctrlif)
{ int i, j;
  struct usb_device* dev = s->usbdev;
  struct mixerdev *ms;

#ifdef ORIG_FIRMWARE
#define N_AC_UNITS 31
  unsigned char uids[N_AC_UNITS] = {
    6,0,0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,
    27,28,19,20,21,22,23,24,25,26
  };
#else
#define N_AC_UNITS 32
  unsigned char uids[N_AC_UNITS] = {
    6,0,0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,
    27,28,19,20,21,29,22,23,24,25,26
  };
#endif

  /* Check that the Units match what we expect. */
  for (i = 0, j = 0; i < len && j < N_AC_UNITS; i += desc[i], j++) {
    if (uids[j] != desc[i+3]) {
      err("unexpected unit %u at decriptor %d",desc[i+3],j);
      return;
    }
  }

  SET(0x01,0x0200,0x1100,0x03,0x0,1);//mode 3
  SET(0x01,0x0100,0x1100,0x00,0x0,1);//enable 0
  SET(0x01,0x0100,0x1100,0x01,0x0,1);//enable 1
  SET(0x01,0x0100,0x1100,0x01,0x0,1);//enable 1

  /* Turn on the CMSS LED. */
  //SET(0x01,0x0100,0x1100,0x01,0x0,1);
  SET(0x01,0x0100,0x1b00,0x00,0x0,1);
  SET(0x05,0x0001,0x1b00,0x64,0x00,2);
  SET(0x05,0x0000,0x1b00,0x02,0x0,1);
  SET(0x01,0x0100,0x1500,0x00,0x0,1);

  SET(0x01,0x0200,0x1100,0x07,0x0,1);//mode 7
  SET(0x01,0x0200,0x1100,0x04,0x0,1);//mode 4

  /* Make the RECORD mixer. */
  dbg("constructing RECORD mixer");
  ms = kmalloc(sizeof(struct mixerdev) +
	       N_RECMIX_CH * sizeof(struct mixerchannel),GFP_KERNEL);
  if (!ms) {
    err("low memory");
    return;
  }
  memset(ms,0,sizeof(struct mixerdev));
  ms->state    = s;
  ms->type     = RECORD;
  ms->selector = REC_SRC_UNIT;
  ms->iface    = ctrlif;
  ms->numch    = N_RECMIX_CH;
  init_mixch(ms,&ms->ch[0],SOUND_MIXER_LINE,REC_LINE_UNIT,REC_LINE_UNIT);
  set_mixch_stereo(&ms->ch[0],1,2);
  init_mixch(ms,&ms->ch[1],SOUND_MIXER_MIC,REC_MIC_UNIT,REC_MIC_UNIT);
  init_mixch(ms,&ms->ch[2],SOUND_MIXER_IMIX,0,0); /* No vol adj "What U Hear"*/
  init_mixch(ms,&ms->ch[3],SOUND_MIXER_VOLUME,REC_MASTER_UNIT,REC_MASTER_UNIT);
  ms->ossmixer = register_sound_mixer(&exaudio_mixer_fops,-1);
  if (ms->ossmixer < 0) {
    err("cannot register mixer");
    kfree(ms);
    return;
  } else {
    INFO("registered mixer (major 14 minor %d)",ms->ossmixer);
  }
  list_add_tail(&ms->list,&s->mixerlist);

  /* Make the PLAYBACK mixer. */
  dbg("constructing PLAYBACK mixer");
  ms = kmalloc(sizeof(struct mixerdev) +
	       N_PLAYMIX_CH * sizeof(struct mixerchannel),GFP_KERNEL);
  if (!ms) {
    err("low memory");
    return;
  }
  memset(ms,0,sizeof(struct mixerdev));
  ms->state    = s;
  ms->type     = PLAYBACK;
  ms->selector = 0;
  ms->iface    = ctrlif;
  ms->numch    = N_PLAYMIX_CH;
  init_mixch(ms,&ms->ch[0],SOUND_MIXER_PCM,PLAY_PCM_UNIT,PLAY_PCM_UNIT);
  init_mixch(ms,&ms->ch[1],SOUND_MIXER_ALTPCM,PLAY_DIGIN_UNIT,PLAY_DIGIN_UNIT);
  init_mixch(ms,&ms->ch[2],SOUND_MIXER_VOLUME,
	     PLAY_MASTER_UNIT,PLAY_MASTER_UNIT);
  init_mixch(ms,&ms->ch[3],SOUND_MIXER_LINE1,PLAY_CPAIR1_UNIT,0);
  init_mixch(ms,&ms->ch[4],SOUND_MIXER_LINE2,PLAY_CPAIR2_UNIT,0);
  init_mixch(ms,&ms->ch[5],SOUND_MIXER_LINE3,PLAY_CPAIR3_UNIT,0);
  set_mixch_stereo(&ms->ch[3],1,2);
  set_mixch_stereo(&ms->ch[4],3,4);
  set_mixch_stereo(&ms->ch[5],5,6);
  init_mixch(ms,&ms->ch[6],SOUND_MIXER_SPEAKER,0,PLAY_SPEAKER_UNIT);
  init_mixch(ms,&ms->ch[7],SOUND_MIXER_DIGITAL1,0,PLAY_DIGOUT1_UNIT);
  init_mixch(ms,&ms->ch[8],SOUND_MIXER_DIGITAL2,0,PLAY_DIGOUT2_UNIT);
  init_mixch(ms,&ms->ch[9],SOUND_MIXER_DIGITAL3,0,PLAY_DIGOUTS_UNIT);
  ms->ossmixer = register_sound_mixer(&exaudio_mixer_fops,-1);
  if (ms->ossmixer < 0) {
    err("cannot register mixer");
    kfree(ms);
    return;
  } else {
    INFO("registered mixer (major 14 minor %d)",ms->ossmixer);
  }
  list_add_tail(&ms->list,&s->mixerlist);
}

/*----------------------------------------------------------------------------*/
#if 0
static int procfs_read(char* buf,char** start,off_t offset,int cnt,
		       int* eof,void* data)
{ unsigned int dlen, i, j;
  int len = 0;

  for (i = 0; i < s->desclen;) {
    dlen = s->desc[i];

    for (j = 0; j < dlen && i < s->desclen; j++, i++) {
      if (3 >= cnt - len)
	len += sprintf(buf+len,"%02X ",s->desc[i]);
      else
	goto end;
    }
    if (1 >= cnt - len)
      len += sprintf(buf+len,"\n");
    else
      goto end;
  }

 end:
  *eof = 1;
  return(len);
}
#endif

/*----------------------------------------------------------------------------*/
static void *exaudio_probe(struct usb_device *dev, unsigned int ifnum,
			   const struct usb_device_id *id)
{ struct usb_config_descriptor *config = dev->actconfig;
  unsigned char buf[8];
  int err, ret;

  struct extigy *s;
  struct usb_interface *iface;
  unsigned char ifin[USB_MAXINTERFACES], ifout[USB_MAXINTERFACES];
  unsigned char *p1;
  unsigned int i, j, k, numifin = 0, numifout = 0;

  if (0x041E != dev->descriptor.idVendor ||
      0x3000 != dev->descriptor.idProduct) {
    err("not an extigy");
    return(NULL);
  }

  if (ifnum != 0) {
    INFO("for the MIDI interface, use the usb-midi driver");
    return(NULL);
  }

  INFO("DEV%d idVendor=0x%04X idProduct=0x%04X wTotalLength=%u "
       "bNumInterfaces=%u"
       ,dev->devnum,dev->descriptor.idVendor,dev->descriptor.idProduct
       ,config->wTotalLength,config->bNumInterfaces);

  if (usb_set_configuration(dev,config->bConfigurationValue) < 0) {
    err("set_configuration failed (ConfigValue 0x%x)",
	config->bConfigurationValue);
    return(NULL);
  }

#ifdef ORIG_FIRMWARE
#define TOO_SHORT 794
#define BIGDESC   1701
#else
#define TOO_SHORT 483
#define BIGDESC   1709
#endif

  if (TOO_SHORT == config->wTotalLength) {

    /* Send message to force it to reconnect with full interface. */
    err = usb_control_msg(dev,usb_sndctrlpipe(dev,0),
			  0x10,0x43,0x0001,0x000a,NULL,0,HZ);
    if (err < 0) err("Error sending vendor msg: %d",err);
    err = usb_get_device_descriptor(dev);
    if (err < 0) ERR("good, usb_get_device_descriptor failed: %d",err);
    if (usb_set_configuration(dev,config->bConfigurationValue) < 0)
      ERR("good, usb_set_configuration failed");
    return(NULL);

  } else if (BIGDESC == config->wTotalLength) {

    /* That's better, a full set of descriptors! */
    s = kmalloc(sizeof(struct extigy),GFP_KERNEL);
    if (!s) return(NULL);
    memset(s,0,sizeof(struct extigy));
    INIT_LIST_HEAD(&s->audiolist);
    INIT_LIST_HEAD(&s->mixerlist);
    s->usbdev = dev;
    s->count = 1;

    ret = usb_get_descriptor(dev,USB_DT_CONFIG,0,buf,8);
    if (ret < 0) {
      err("cannot get first 8 bytes of config desc 0 of DEV%d",dev->devnum);
      kfree(s);
      return(NULL);
    }
    if (buf[1] != USB_DT_CONFIG || buf[0] < 9) {
      err("invalid config descriptor 0 of DEV%d",dev->devnum);
      kfree(s);
      return(NULL);
    }
    s->desclen = buf[2] | (buf[3] << 8);
    s->desc = kmalloc(s->desclen,GFP_KERNEL);
    if (!s->desc) {
      kfree(s);
      return(NULL);
    }
    ret = usb_get_descriptor(dev,USB_DT_CONFIG,0,s->desc,s->desclen);
    if (ret < 0) {
      err("cannot get config descriptor 0 of DEV%d",dev->devnum);
      kfree(s->desc);
      kfree(s);
      return(NULL);
    }
#if 0
    create_proc_read_entry("exaudio",0,NULL,procfs_read,NULL);
#endif

    /* parse_control, now inlined, follows... */
    /* Find audiocontrol interface */
    p1 = find_csif_desc(s->desc,s->desclen,NULL,HEADER,ifnum,-1);
    if (!p1) {
      err("DEV%d ACIF%u no HEADER found",dev->devnum,ifnum);
      goto ret;
    }
    if (p1[0] < 8 + p1[7]) {
      err("DEV%d ACIF%u HEADER error",dev->devnum,ifnum);
      goto ret;
    }
    if (!p1[7]) err("DEV%d ACIF%u has no AudioStreaming or MidiStreaming IFs",
		    dev->devnum,ifnum);

    for (i = 0; i < p1[7]; i++) {
      j = p1[8+i];
      if (j >= config->bNumInterfaces) {
	err("DEV%d ACIF%u IF%u does not exist",dev->devnum,ifnum,j);
	continue;
      }
      iface = &config->interface[j];
      if (iface->altsetting[0].bInterfaceClass != USB_CLASS_AUDIO) {
	err("DEV%d ACIF%u IF%u is not an AudioClass IF",dev->devnum,ifnum,j);
	continue;
      }
      if (iface->altsetting[0].bInterfaceSubClass == 3) {
	dbg("DEV%d ACIF%u IF%u MIDIStreaming not supported",
	    dev->devnum,ifnum,j);
	continue;
      }
      if (iface->altsetting[0].bInterfaceSubClass != 2) {
	err("DEV%d ACIF%u IF%u invalid AudioClass subtype",dev->devnum,ifnum,j);
	continue;
      }
      if (iface->num_altsetting == 0) {
	err("DEV%d ACIF%u has no working interface.",dev->devnum,ifnum);
	continue;
      }
      if (iface->num_altsetting == 1) {
	err("DEV%d ACIF%u has only 1 altsetting.",dev->devnum,ifnum);
	continue;
      }
      if (iface->altsetting[0].bNumEndpoints > 0) {
	/* Check all endpoints; should they all have a bandwidth of 0 ? */
	for (k = 0; k < iface->altsetting[0].bNumEndpoints; k++) {
	  if (iface->altsetting[0].endpoint[k].wMaxPacketSize > 0) {
	    err("DEV%d ACIF%u ENDP%d ALT0 has nonzero bandwidth",
		dev->devnum,ifnum,k);
	    break;
	  }
	}
	if (k < iface->altsetting[0].bNumEndpoints) continue;
      }
      if (iface->altsetting[1].bNumEndpoints < 1) {
	err("DEV%d ACIF%u IF%u has no endpoint",dev->devnum,ifnum,j);
	continue;
      }
      /* note: this requires the data endpoint to be ep0 and the optional sync
	 ep to be ep1, which seems to be the case */
      if (iface->altsetting[1].endpoint[0].bEndpointAddress & USB_DIR_IN) {
	if (numifin < USB_MAXINTERFACES) {
	  dbg("IF %d is INPUT  no %d",j,numifin);
	  ifin[numifin++] = j;
	  usb_driver_claim_interface(&exaudio_driver,iface,(void*)-1);
	}
      } else {
	if (numifout < USB_MAXINTERFACES) {
	  dbg("IF %d is OUTPUT no %d",j,numifout);
	  ifout[numifout++] = j;
	  usb_driver_claim_interface(&exaudio_driver,iface,(void*)-1);
	}
      }
    }
    INFO("DEV%d IF%u has %u input and %u output AudioStreaming IFs",
	 dev->devnum,ifnum,numifin,numifout);
    for (i = 0; i < numifin && i < numifout; i++)
      parse_streaming(s,s->desc,s->desclen,ifin[i],ifout[i]);
    for (j = i; j < numifin; j++)
      parse_streaming(s,s->desc,s->desclen,ifin[i],-1);
    for (j = i; j < numifout; j++)
      parse_streaming(s,s->desc,s->desclen,-1,ifout[i]);

    construct_mixer(s,s->desc,s->desclen,ifnum);

  ret:
    if (list_empty(&s->audiolist) && list_empty(&s->mixerlist)) {
      kfree(s->desc);
      kfree(s);
      return(NULL);
    }
    /* Ready the IR interupt handler. */
    setup_interrupt(s,dev);
    INFO("everything successful");
    down(&open_sem);
    list_add_tail(&s->audiodev,&audiodevs);
    up(&open_sem);
    return(s);

  } else {
    ERR("This extigy's layout is not known to me (firmware mismatch?)");
    return(NULL);
  }
}

/*----------------------------------------------------------------------------*/
static void exaudio_disconnect(struct usb_device *dev,void *ptr)
{
  struct extigy *s = (struct extigy*)ptr;
  struct list_head *list;
  struct audiodev *as;
  struct mixerdev *ms;

  /* we get called with -1 for every audiostreaming interface registered */
  if (s == (struct extigy*) -1) {
    dbg("exaudio_disconnect: called with -1");
    return;
  }
  if (!s->usbdev) {
    dbg("exaudio_disconnect: already called for %p!",s);
    return;
  }
  down(&open_sem);
  usb_unlink_urb(&s->irurb);
  list_del(&s->audiodev);
  INIT_LIST_HEAD(&s->audiodev);
  s->usbdev = NULL;
  /* Deregister all audio and mixer devices,
   * so no new processes can open this device. */
  for (list = s->audiolist.next; list != &s->audiolist; list = list->next) {
    as = list_entry(list,struct audiodev,list);
    usbin_disc(as);
    usbout_disc(as);
    wake_up(&as->usbin.dma.wait);
    wake_up(&as->usbout.dma.wait);
    if (as->ossdsp >= 0)
      unregister_sound_dsp(as->ossdsp);
    as->ossdsp = -1;
  }
  for (list = s->mixerlist.next; list != &s->mixerlist; list = list->next) {
    ms = list_entry(list,struct mixerdev,list);
    if (ms->ossmixer >= 0)
      unregister_sound_mixer(ms->ossmixer);
    ms->ossmixer = -1;
  }
  release(s);
  wake_up(&open_wait);
}

/*----------------------------------------------------------------------------*/
static int __init exaudio_init(void)
{
  usb_register(&exaudio_driver);
  INFO(DRIVER_VERSION);
  register_ir_device();
  return(0);
}

/*----------------------------------------------------------------------------*/
static void __exit exaudio_cleanup(void)
{
  usb_deregister(&exaudio_driver);
  unregister_ir_device();
}

module_init(exaudio_init);
module_exit(exaudio_cleanup);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
