/* $Id: remote.c,v 1.4 2003/01/29 13:55:10 cochran Exp $
 *
 *  Copyright (c) 2003 Richard Cochran (cochran@cs.umass.edu)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 */
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/module.h>
#include <linux/list.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>
#include <linux/usb.h>

#include "exaudio.h"
#include "remote.h"

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,19)
typedef struct devrequest {
  __u8 requesttype;
  __u8 request;
  __u16 value;
  __u16 index;
  __u16 length;
} __attribute__ ((packed)) devrequest;
#endif

extern struct semaphore open_sem;
extern struct list_head audiodevs;

static int irmajor = -1;

static int ir_open(struct inode *inode, struct file *filp);
static ssize_t ir_read(struct file *filp,char *buf,size_t count,loff_t *f_pos);
static int ir_release(struct inode *inode, struct file *filp);

static struct file_operations ir_fops = {
  owner:   THIS_MODULE,
  //ioctl: ir_ioctl,
  open:    ir_open,
  read:    ir_read,
  release: ir_release,
  //write: ir_write,
};

/*----------------------------------------------------------------------------*/
static int ctlmsg(struct usb_device* dev,
		  __u8 request,
		  __u8 requesttype,
		  __u16 value,
		  __u16 index,
		  void* inbuf,
		  __u16 len,
		  usb_complete_t complete,
		  void* context)
{
  struct urb* new;
  devrequest* dr;
  unsigned char* buf;

  new = usb_alloc_urb(0);
  if (!new) {
    err("low memory");
    return(-1);
  }

  dr = kmalloc(sizeof(devrequest),GFP_ATOMIC);
  if (!dr) {
    err("low memory");
    kfree(new);
    return(-1);
  }

  buf = kmalloc(len,GFP_ATOMIC);
  if (!buf) {
    err("low memory");
    kfree(new);
    kfree(dr);
    return(-1);
  }

  if (inbuf) {
    /* Sending a command, copy */
    //TODO
  } else {
    /* Receiving into buf, so do nothing */
  }

  dr->request     = request;
  dr->requesttype = requesttype;
  dr->value       = cpu_to_le16p(&value);
  dr->index       = cpu_to_le16p(&index);
  dr->length      = cpu_to_le16p(&len);
  
  usb_fill_control_urb(new,dev,usb_rcvctrlpipe(dev,0),
		       (char*)dr,buf,len,complete,context);

  return(usb_submit_urb(new));
}

/*----------------------------------------------------------------------------*/
static void fetch_ir_code(struct urb* u)
{
  unsigned char* buf = u->transfer_buffer;
  struct extigy* exy = (struct extigy*)u->context;
  if (u->status < 0) {
    err("control request with bad status=%d",u->status);
  } else {
    exy->ircnt++;
    exy->irval = (exy->ircnt<<16)|(buf[1]<<8)|buf[0];
    //err("IR event %08X",exy->irval);
  }
  kfree(u->setup_packet);
  kfree(u->transfer_buffer);
  usb_free_urb(u);
}

/*----------------------------------------------------------------------------*/
static unsigned int last = 0;

static void handle_interrupt(struct urb* u)
{
  struct extigy* exy = (struct extigy*)u->context;

  if (u->status < 0) {
    err("interrupt with bad status=%d",u->status);
    return;
  }

  //err("%d IN BITs %08X last %d",exy->ircnt,exy->usbdev->toggle[0],last);

  if (last == usb_gettoggle(exy->usbdev,1,0)) {
    //err("missed a toggle - fix it");
    usb_dotoggle(exy->usbdev,1,0);
  }

  last = usb_gettoggle(exy->usbdev,1,0);

  if (exy->irbuf[1]) { /* A jack (un)plugged event. */
    return;
  }

  /* Fetch the code from memory. */
  if (ctlmsg(exy->usbdev,GET_MEM,0xA1,0,0,NULL,2,fetch_ir_code,exy)) {
    err("cannot send control message to get IR code");
  }
}

/*----------------------------------------------------------------------------*/
static int ir_open(struct inode *inode, struct file *filp)
{
  struct list_head* devs;
  struct extigy* exy;
  if (down_interruptible(&open_sem)) return(-ERESTARTSYS);
  /* TODO - find the particular extigy, by minor number.
   * For now just use the first one.
   * for (devs = audiodevs.next; devs != &audiodevs; devs = devs->next)*/
  devs = audiodevs.next;
  exy = list_entry(devs,struct extigy,audiodev);
  if (!exy->usbdev) {
    up(&open_sem);
    return(-ENODEV);
  }
  filp->private_data = exy;
  exy->count++;
  up(&open_sem);
  return(0);
}

/*----------------------------------------------------------------------------*/
static ssize_t ir_read(struct file *filp,char *buf,size_t count,loff_t *f_pos)
{
  struct extigy* exy = (struct extigy*)filp->private_data;
  __u32 result;

  if (4 != count) return(-EINVAL);
  /* TODO - sleep for a blocking read */
  result = exy->irval;
  if (copy_to_user(buf,&result,count)) return(-EFAULT);
  *f_pos += count;
  return(count);
}

/*----------------------------------------------------------------------------*/
static int ir_release(struct inode *inode, struct file *filp)
{
  struct extigy* exy = (struct extigy*)filp->private_data;
  lock_kernel();
  down(&open_sem);
  release(exy); /* This decrements exy->count, and calls up() */
  unlock_kernel();
  return(0);
}

/*----------------------------------------------------------------------------*/
int register_ir_device(void)
{
  int num;
  num = register_chrdev(0,"ir",&ir_fops);
  if (num < 0) {
    err("cannot get a major number");
    return(num);
  }
  irmajor = num;
  return(0);
}

/*----------------------------------------------------------------------------*/
void setup_interrupt(struct extigy* exy,struct usb_device *dev)
{
  int err;
  unsigned int pipe;
  pipe = usb_rcvintpipe(dev,1);
  FILL_INT_URB(&exy->irurb,dev,pipe,exy->irbuf,8,handle_interrupt,exy,1);
  err = usb_submit_urb(&exy->irurb);
  if (err) {
    err("cannot submit interrupt urb");
  }
}

/*----------------------------------------------------------------------------*/
void unregister_ir_device(void)
{
  unregister_chrdev(irmajor,"ir");
}
