/* 
 *    linux/drivers/video/fbcon-splash.c - splash screen handling functions.
 *	
 *	(w) 2001 by Volker Poplawski, <volker@suse.de>
 * 		    Stefan Reinauer, <stepan@suse.de>
 * 		    Steffen Winterfeldt, <snwint@suse.de>
 * 		    
 * 	       SuSE screen work by Ken Wimer, <wimer@suse.de>
 */

#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/fb.h>
#include <linux/vt_kern.h>
#include <linux/vmalloc.h>

#include <asm/irq.h>
#include <asm/system.h>

#include <video/fbcon.h>
#include <video/font.h>
#include <video/fbcon-cfb16.h>  /* for fbcon_cfb16 */

#include "fbcon-splash.h"
#include "fbcon-jpegdec.h"

#define SPLASH_VERSION "2.0.2"

#ifdef CONFIG_BLK_DEV_INITRD
unsigned char signature[] = "BOOTSPL1SPL2";
#endif

/* from drivers/char/console.c */
extern void con_remap_def_color(int currcons, int new_color);

/* internal control states and data pointers */
struct splash_data splash_data;

unsigned char *linux_splash;	/* decoded picture */
int linux_splash_size = 0;

int splash_bytes;		/* bytes per line in linux_splash */
int splash_shown = 0;		/* is the splash onscreen? */

int splash_default = 0xf01;

static int splash_status(struct display *p);
static int splash_recolor(struct display *p);
static int splash_check_jpeg(unsigned char *jpeg, int width, int height, int depth);

int __init splash_init(char *options)
{
	splash_default = simple_strtoul(options, NULL, 0);
	return 0;
}

__setup("splash=", splash_init);


void penguin_off(struct display *p, int update)
{
	unsigned int i, j, x, y, width, height;
	unsigned short bg;

	if (fg_console != p->conp->vc_num)
	    return;
	if (!linux_splash)
	    return;
        if (!p->splash_data || p->splash_data->splash_penguin_wi <= 0 || p->splash_data->splash_penguin_he <= 0)
	    return;
	if (p->var.bits_per_pixel != 16)
	    return;
	bg = (p->splash_data->splash_penguin_r & 0xf8) << 8 |
	     (p->splash_data->splash_penguin_g & 0xfc) << 3 |
	     (p->splash_data->splash_penguin_b & 0xf8) >> 3;
	x = p->splash_data->splash_penguin_xo;
	y = p->splash_data->splash_penguin_yo;
	width = p->splash_data->splash_penguin_wi;
	height = p->splash_data->splash_penguin_he;
	for (i = y; i < y + height; i++)
	    for (j = x; j < x + width; j++)
		*(unsigned short *)(linux_splash + j * 2 + i * splash_bytes) = bg;

	if (update) {
	    update_region (fg_console, p->conp->vc_origin +
			p->conp->vc_size_row * p->conp->vc_top,
			p->conp->vc_size_row * (p->conp->vc_bottom -
			p->conp->vc_top) / 2 );
	    if (p->dispsw->clear_margins)
		p->dispsw->clear_margins(p->conp, p, 0);
	}
}

static int splash_check_jpeg(unsigned char *jpeg, int width, int height, int depth)
{
    int size, err;
    unsigned char *mem;

    size = ((width + 15) & ~15) * ((height + 15) & ~15) * (depth >> 3);
    mem = vmalloc(size);
    if (!mem) {
	printk(KERN_INFO "No memory for decoded picture!\n");
	return -1;
    }
    if ((err = jpeg_decode(jpeg, mem, ((width + 15) & ~15), ((height + 15) & ~15), depth))) {
	    printk(KERN_INFO "error %d while decompressing picture.\n",err);
    }
    vfree(mem);
    return err ? -1 : 0;
}

int splash_getraw(unsigned char *start, unsigned char *end)
{
	unsigned char *ndata;
	int found = 0;
	int splash_size = 0;
	void *splash_start = 0;
        int unit = 0;
        int width = 0, height = 0;
	struct display *p;

	printk(KERN_INFO "Looking for splash picture...");

	p = &fb_display[0];
	
	for (ndata = start; ndata < end; ndata++) {
		if (*((unsigned int *)ndata) != *((unsigned int *)signature))
			continue;
		if (*((unsigned int *)(ndata+4))==*((unsigned int *)(signature+4))) {
			printk(".");
			unit = 0;
			p = &fb_display[0];
			width = p->var.xres;
			height = p->var.yres;

			splash_size = ndata[16] + (ndata[17] << 8) + (ndata[18] << 16) + (ndata[19] << 24);
			if (ndata + 20 + splash_size > end) {
				printk(" found, but truncated!\n");
				return -1;
			}
			if (!jpeg_check_size(ndata + 20, width, height)) { 
				ndata += 20 - 1 + splash_size;
				continue;
			}
			if (splash_check_jpeg(ndata + 20, width, height, p->var.bits_per_pixel))
				return -1;
			if (p->splash_data)
				vfree(p->splash_data);
			p->splash_data = vmalloc(sizeof(*p->splash_data) + splash_size);
			if (!p->splash_data)
				break;
			p->splash_data->splash_text_xo = ndata[ 8] + (ndata[ 9] << 8);
			p->splash_data->splash_text_yo = ndata[10] + (ndata[11] << 8);
			p->splash_data->splash_text_wi = ndata[12] + (ndata[13] << 8);
			p->splash_data->splash_text_he = ndata[14] + (ndata[15] << 8);
			/* use 8x16 font... */
			p->splash_data->splash_text_xo *= 8;
			p->splash_data->splash_text_wi *= 8;
			p->splash_data->splash_text_yo *= 16;
			p->splash_data->splash_text_he *= 16;
			p->splash_data->splash_color = (splash_default >> 8) & 0x0f;
			p->splash_data->splash_fg_color = (splash_default >> 4) & 0x0f;
			p->splash_data->splash_state = splash_default & 1;
			p->splash_data->splash_penguin_xo = p->splash_data->splash_text_xo + 10;
			p->splash_data->splash_penguin_yo = p->splash_data->splash_text_yo + 10;
			p->splash_data->splash_penguin_wi = p->splash_data->splash_text_wi - 20;
			p->splash_data->splash_penguin_he = p->splash_data->splash_text_he - 20;
			p->splash_data->splash_penguin_r = 0xf0;
			p->splash_data->splash_penguin_g = 0xf0;
			p->splash_data->splash_penguin_b = 0xf0;
			p->splash_data->splash_no_penguin = 0;
			splash_start = ndata + 20;
			found = 1;
			break;
		}
		if (*((unsigned int *)(ndata + 4)) == *((unsigned int *)(signature+8))) {
			printk(".");
			unit = ndata[8];
			if (unit >= MAX_NR_CONSOLES)
			    continue;
			if (unit)
			    vc_allocate(unit);
			p = &fb_display[unit];
			width = fb_display[unit].var.xres;
			height = fb_display[unit].var.yres;
			splash_size = ndata[12] + (ndata[13] << 8) + (ndata[14] << 16) + (ndata[15] << 24);
			if (splash_size == -1) {
			    printk(" found, updating values.\n");
			    if (p->splash_data) {
				if (ndata[9] != 255)
				    p->splash_data->splash_state = ndata[9];
				if (ndata[10] != 255)
				    p->splash_data->splash_fg_color = ndata[10];
				if (ndata[11] != 255)
				    p->splash_data->splash_color = ndata[11];
			    }
			    return unit;
			}
			if (splash_size == 0) {
				printk(" found, freeing memory.\n");
				if (p->splash_data)
					vfree(p->splash_data);
				p->splash_data = 0;
				return unit;
			}
			if (ndata + 35 + splash_size > end) {
				printk(" found, but truncated!\n");
				return -1;
			}
			if (!jpeg_check_size(ndata + 35, width, height)) {
				ndata += 35 - 1 + splash_size;
				continue;
			}
			if (splash_check_jpeg(ndata + 35, width, height, p->var.bits_per_pixel))
				return -1;
			if (p->splash_data)
				vfree(p->splash_data);
			p->splash_data = vmalloc(sizeof(*p->splash_data) + splash_size);
			if (!p->splash_data)
				break;
			p->splash_data->splash_state = ndata[9];
			p->splash_data->splash_fg_color = ndata[10];
			p->splash_data->splash_color = ndata[11];
			p->splash_data->splash_text_xo = ndata[16] + (ndata[17] << 8);
			p->splash_data->splash_text_yo = ndata[18] + (ndata[19] << 8);
			p->splash_data->splash_text_wi = ndata[20] + (ndata[21] << 8);
			p->splash_data->splash_text_he = ndata[22] + (ndata[23] << 8);
			p->splash_data->splash_penguin_xo = ndata[24] + (ndata[25] << 8);
			p->splash_data->splash_penguin_yo = ndata[26] + (ndata[27] << 8);
			p->splash_data->splash_penguin_wi = ndata[28] + (ndata[29] << 8);
			p->splash_data->splash_penguin_he = ndata[30] + (ndata[31] << 8);
			p->splash_data->splash_penguin_r = ndata[32];
			p->splash_data->splash_penguin_g = ndata[33];
			p->splash_data->splash_penguin_b = ndata[34];
			p->splash_data->splash_no_penguin = 0;
			splash_start = ndata + 35;
			found = 1;
			break;
		}
	}

	if (!found) {
		printk(" no good signature found.\n");
		return -1;
	}
	if (p->splash_data->splash_text_xo + p->splash_data->splash_text_wi > width || p->splash_data->splash_text_yo + p->splash_data->splash_text_he > height) {
		vfree(p->splash_data);
		p->splash_data = 0;
		printk(" found, but has oversized text area!\n");
		return -1;
	}
        if (p->dispsw->setup != fbcon_cfb16.setup) {
		vfree(p->splash_data);
		p->splash_data = 0;
		printk(" found, but framebuffer can't handle it!\n");
		return -1;
	}
	printk (" found (%dx%d, %d bytes).\n", width, height, splash_size);
	/* copy data so that initrd memory can be freed. */
	memcpy(p->splash_data->splash_jpeg, splash_start, splash_size);
	return unit;
}

int splash_prepare(p)
struct display *p;
{
	int err;
        int width, height, depth, size;

	if (!p->splash_data || !p->splash_data->splash_state) {
		if (linux_splash)
			vfree(linux_splash);
		linux_splash = 0;
		linux_splash_size = 0;
		splash_shown = 0;
		return -1;
	}
        width = p->var.xres;
        height = p->var.yres;
        depth = p->var.bits_per_pixel;
	if (depth != 16) {	/* Other targets might need fixing */
		splash_shown = 0;
		return -2;
	}

	size = ((width + 15) & ~15) * ((height + 15) & ~15) * (depth>>3);
	if (size != linux_splash_size) {
		if (linux_splash)
			vfree(linux_splash);
		linux_splash_size = 0;
		linux_splash = 0;
		splash_shown = 0;
	}
	if (!linux_splash)
		linux_splash = vmalloc(size);

	if (!linux_splash) {
		linux_splash_size = 0;
		printk(KERN_INFO "Not enough memory for splash screen.\n");
		splash_shown = 0;
		return -3;
	}

	if ((err = jpeg_decode(p->splash_data->splash_jpeg, linux_splash, 
		 ((width + 15) & ~15), ((height + 15) & ~15), depth))) {
		printk(KERN_INFO "Error %d while decompressing splash screen.\n", err);
		vfree(linux_splash);
		linux_splash = 0;
		linux_splash_size = 0;
		splash_shown = 0;
		return -4;
	}
	linux_splash_size = size;
	splash_bytes = ((width + 15) & ~15) * (depth >> 3);
	splash_data = *p->splash_data;
	splash_shown = p->splash_data->splash_state;
	if (p->splash_data->splash_no_penguin)
	    penguin_off(p, 0);
	return 0;
}

#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>

struct proc_dir_entry *proc_splash;
int splash_read_proc(char *buffer, char **start, off_t offset, int size,
			int *eof, void *data);
int splash_write_proc(struct file *file, const char *buffer,
			unsigned long count, void *data);

int splash_read_proc(char *buffer, char **start, off_t offset, int size,
			int *eof, void *data)
{
	int len = 0;
	off_t begin = 0;
	struct display *p = &fb_display[0];

	int color = p->splash_data ? p->splash_data->splash_color << 4 | p->splash_data->splash_fg_color : splash_default >> 4;
	int status = p->splash_data ? p->splash_data->splash_state & 1 : 0;
	len += sprintf(buffer + len, "Splash screen v%s (0x%02x, %dx%d): %s\n",
		    SPLASH_VERSION, color, p->var.xres, p->var.yres, status ? "on" : "off");
	if (offset >= begin + len)
		return 0;

	*start = buffer + (begin - offset);

	return (size < begin + len - offset ? size : begin + len - offset);
}

static int splash_recolor(struct display *p)
{
	if (!p->splash_data)
	    return -1;
	if (!p->splash_data->splash_state)
	    return 0;
	con_remap_def_color(p->conp->vc_num, p->splash_data->splash_color << 4 | p->splash_data->splash_fg_color);
	if (fg_console == p->conp->vc_num) {
	        splash_data = *p->splash_data;
		update_region(fg_console,
			      p->conp->vc_origin +
			      p->conp->vc_size_row *
			      p->conp->vc_top,
			      p->conp->vc_size_row *
			      (p->conp->vc_bottom -
			       p->conp->vc_top) / 2);
	}
	return 0;
}

static int splash_status(struct display *p)
{
	printk(KERN_INFO "Splash status on console %d changed to %s\n", p->conp->vc_num, p->splash_data && p->splash_data->splash_state ? "on" : "off");

	if (fg_console == p->conp->vc_num)
		splash_prepare(p);
	if (p->splash_data && p->splash_data->splash_state) {
		con_remap_def_color(p->conp->vc_num, p->splash_data->splash_color << 4 | p->splash_data->splash_fg_color);
		/* resize_con also calls con_switch which resets yscroll */
		vc_resize_con(p->splash_data->splash_text_he / fontheight(p),
                              p->splash_data->splash_text_wi / fontwidth(p),
                              p->conp->vc_num);
		if (fg_console == p->conp->vc_num) {
			update_region(fg_console,
				      p->conp->vc_origin +
				      p->conp->vc_size_row *
				      p->conp->vc_top,
				      p->conp->vc_size_row *
				      (p->conp->vc_bottom -
				       p->conp->vc_top) / 2);
			if (p->dispsw->clear_margins)
			    p->dispsw->clear_margins(p->conp, p, 0);
		}
	} else {
	  	/* Switch bootsplash off */
		con_remap_def_color(p->conp->vc_num, 0x07);
		vc_resize_con(p->var.yres / fontheight(p),
			      p->var.xres / fontwidth(p),
			      p->conp->vc_num);
	}
	return 0;
}

int splash_write_proc(struct file *file, const char *buffer,
		      unsigned long count, void *data)
{
        int new, unit;
	struct display *p;
	
	if (!buffer || !splash_default)
		return count;

	if (!strcmp(buffer,"hide penguin\n") || !strcmp(buffer,"show penguin\n")) {
		p = &fb_display[0];
		if (p->splash_data) {
		    if (*buffer == 'h') {
			printk(KERN_INFO "Switching off penguin.\n");
			p->splash_data->splash_no_penguin = 1;
			penguin_off(p, 1);
		    } else {
			printk(KERN_INFO "Switching on penguin.\n");
			p->splash_data->splash_no_penguin = 0;
			splash_status(p);
		    }
		}
		return count;
	} 
	if (!strncmp(buffer, signature, 7)) {
	    unit = splash_getraw((unsigned char *)buffer, (unsigned char *)buffer + count);
	    if (unit >= 0) {
		p = &fb_display[unit];
		splash_status(p);
	    }
	    return count;
	}
	p = &fb_display[0];
	if (!p->splash_data)
		return count;
	if (buffer[0] == 't') {
	        p->splash_data->splash_state ^= 1;
		splash_status(p);
		return count;
	}
	new = simple_strtoul(buffer, NULL, 0);
	if (new > 1) {
		/* expert user */
		p->splash_data->splash_color    = new >> 8 & 0xff;
		p->splash_data->splash_fg_color = new >> 4 & 0x0f;
	}
	if ((new & 1) == p->splash_data->splash_state)
		splash_recolor(p);
	else {
	        p->splash_data->splash_state = new & 1;
		splash_status(p);
	}
	return count;
}

int splash_proc_register(void)
{
	if ((proc_splash = create_proc_entry("splash", 0, 0))) {
		proc_splash->read_proc = splash_read_proc;
		proc_splash->write_proc = splash_write_proc;
		return 0;
	}
	return 1;
}

int splash_proc_unregister(void)
{
	if (proc_splash)
		remove_proc_entry("splash", 0);
	return 0;
}
#endif
