/*
 *  linux/fs/supermount/super.c
 *
 *  Original version:
 *      Copyright (C) 1995, 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/inode.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/super.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2, 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *
 */

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/list.h>
#include <linux/locks.h>

#include "supermount_i.h"

struct inode *
supermount_create_inode(struct super_block *s)
{
	struct inode *inode = new_inode(s);
	struct supermount_sb_info *sbi = supermount_sbi(s);

	if (!inode)
		return NULL;

	inode->i_mode = sbi->s_default_mode;
	if (s->s_root) {
		inode->i_uid = s->s_root->d_inode->i_uid;
		inode->i_gid = s->s_root->d_inode->i_gid;
	} else {
		inode->i_uid = 0;
		inode->i_gid = 0;
	}
	inode->i_blksize = inode->i_sb->s_blocksize;
	inode->i_rdev = 0;
	inode->i_atime = inode->i_mtime = inode->i_ctime = CURRENT_TIME;
	supermount_i(inode)->i_sb_version = sbi->s_version;
	list_add(&(supermount_i(inode)->i_fs), &sbi->s_inodes);

	return inode;
}

static void
supermount_clean_inodes(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct list_head *ptr;
	struct supermount_inode_info *info;

	list_for_each(ptr, &sbi->s_inodes) {
		info = list_entry(ptr, struct supermount_inode_info, i_fs);
		if (!info)
			BUG();
		if (!info->i_dentry)
			continue;
		dput(info->i_dentry);
		info->i_dentry = NULL;
	}
}

/*
 * Cut off busy dentries unless mounted over. They go away when
 * their owners die
 * We run under mutex or sb lock so nobody can change directory structure
 * d_drop holds dcache lock for us
 * FIXME: it probably should unmount submounts
 */
static void
supermount_drop_dentries(struct super_block *sb)
{
	struct dentry *d;
	struct list_head *ptr;

	if (!sb->s_root)
		return;

	list_for_each(ptr, &sb->s_root->d_subdirs) {
		d = list_entry(ptr, struct dentry, d_child);

		if (!have_submounts(d))
			d_drop(d);
        }
}

void
subfs_umount(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct vfsmount *mnt = sbi->s_undermount;
	struct inode *root_inode;

	if (sb->s_root) {
		root_inode = sb->s_root->d_inode;
		if (!supermount_i(root_inode)->i_dentry)
			BUG();
		dput(supermount_i(root_inode)->i_dentry);
		supermount_i(root_inode)->i_dentry = NULL;
	}
	shrink_dcache_sb(sb);
	supermount_clean_inodes(sb);
	supermount_drop_dentries(sb);
	BUG_ON(!mnt);
	mntput(mnt);
	sbi->s_state = SUPERMOUNT_UNMOUNTED;
	sbi->s_undermount = NULL;
	sbi->s_version++;
	sb->s_flags = sbi->s_mflags;
}

static int
subfs_remount(struct super_block *sb, int flags)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct super_block *subsb = supermount_subfs_sb(sbi);
	int real_flags = (sb->s_flags | flags) & MS_RMT_MASK;
	int retval;

	retval = do_remount_sb(subsb, real_flags, NULL);
	if (flags & MS_RDONLY)
		fsync_dev(subsb->s_dev);
	return retval;
}

static inline int
subfs_remount_ro(struct super_block *sb)
{
	int retval = subfs_remount(sb, MS_RDONLY);

	if (retval)
		return retval;
	unmark_subfs_rw(sb);
	return 0;
}

static inline int
subfs_remount_rw(struct super_block *sb)
{
	int retval = subfs_remount(sb, 0);

	if (retval)
		return retval;
	mark_subfs_rw(sb);
	return 0;
}

static int
subfs_real_mount(struct super_block *sb, char *type)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct vfsmount *mnt;
	int retval;

	if (sbi->s_data_copy)
		strcpy(sbi->s_data, sbi->s_data_copy);

	mnt = do_kern_mount(type,
			    sb->s_flags | MS_MGC_VAL | MS_SUPERMOUNTED,
			    sbi->s_devname, sbi->s_data);

	if (IS_ERR(mnt)) {
		retval = PTR_ERR(mnt);
	} else {
		retval = 0;
		mnt->mnt_sb->s_media_changed = 0;
		sbi->s_undermount = mnt;
	}
	return retval;
}

static int
subfs_real_mount2(struct super_block *sb, char *type)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct inode *root_inode;
	int retval = 0;

	sb->s_flags = sbi->s_mflags;

	retval = subfs_real_mount(sb, type);

	if (retval == -EROFS) {
		sb->s_flags |= MS_RDONLY;
		retval = subfs_real_mount(sb, type);
		if (!retval)
			goto assign_root_inode;
	}
	if (retval)
		goto out;

	if (!(sb->s_flags & MS_RDONLY)) {
		retval = subfs_remount_ro(sb);
		if (retval)
			goto out;
	}
assign_root_inode:
	root_inode = sb->s_root->d_inode;
	if (supermount_i(root_inode)->i_dentry)
		BUG();
	supermount_i(root_inode)->i_dentry = dget(sbi->s_undermount->mnt_root);
out:
	return retval;
}

static char *types_in_order[] = {
	"udf",
	"iso9660",
	"vfat",
	"msdos",
	"ext2",
	NULL
};

/*
 * -ENXIO
 *
 */

static int
subfs_mount(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	int retval = -ENODEV;
	char **type = types_in_order;

	if (sbi->s_type && strcmp(sbi->s_type, "auto"))
		return subfs_real_mount2(sb, sbi->s_type);

	while (*type && retval
	       && retval != -ENXIO
	       && retval != -ENOMEDIUM
	       && retval != -ENOENT) {
		retval = subfs_real_mount2(sb, *type);
		type++;
	}

	return retval;
}

/* We used not to check the medium if we was already mounted, this
 * provoked problems in floppies & things that don't have a working
 * locking mechanism, now we _always_ check if the medium has changed.
 */

/* Return 0 (OK) if medium is valid. */
int
subfs_go_online(struct super_block *sb)
{
	int retval = 0;
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	down(&sbi->sem);

	if (subfs_is_mounted(sbi)) {
		struct super_block *subsb = supermount_subfs_sb(sbi);
		kdev_t dev = subsb->s_dev;

		/*
		 * check for disk change or disk not present
		 *
		 * FIXME there is a race here; check must be protected
		 * with mutex; currently it is possible that media changed
		 * flag is stolen
		 */

		if (!(subsb->s_media_changed || query_disk_change(dev)))
			goto mounted;

		/* We have a disk change!  Unmount the subfs */
		subfs_umount(sb);
		/* The call to disk_changed may clear the media-
		   changed flag on the device, so we need to force a
		   media invalidation.  We don't want to do this
		   before unmounting the subfs, naturally! */
		invalidate_media(dev);
	}
	/* OK, we're unmounted now --- can we remount?  Please? */
	if (sbi->s_undermount)
		BUG();

	retval = subfs_mount(sb);
	if (retval) {
		retval = 1;
		goto out;
	}
	sbi->s_version++;
	supermount_i(sb->s_root->d_inode)->i_sb_version = sbi->s_version;
	unmark_subfs_rw(sb);
	sbi->s_state = SUPERMOUNT_SUSPENDED;
mounted:
	switch (sbi->s_state) {
	case SUPERMOUNT_UNMOUNTED:
		BUG();
	case SUPERMOUNT_SUSPENDED:
		supermount_lock_door(sb);
		sbi->s_state = SUPERMOUNT_ONLINE;
		break;
	case SUPERMOUNT_ONLINE:
		/* No need to do anything */ ;
	}
out:
	atomic_inc(&sbi->s_opencount);
	up(&sbi->sem);
	return retval;
}

/* 
 * Check the state of the supermount filesystem after it has been
 * released by some inode.  If it is no longer active, do the necessary
 * cleanup: fsync the disk, remount it readonly if appropriate, and
 * unlock the drive door if possible.
 */

void
subfs_go_inactive(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	down(&sbi->sem);

	/* there was a problem mounting */
	if (!subfs_is_mounted(sbi))
		goto out;

	if ((atomic_read(&sbi->s_opencount) > 1))
		goto out;

	/*
	 * Do an initial fsync_dev.  We do this before locking the 
	 * superblock to minimise the duration of the lock. 
	 */
	if (subfs_is_dirty(sbi)) {
		fsync_dev(sb->s_dev);
		mark_subfs_clean(sb);
	}

	/*
	 * First thing is, has the filesystem been made
	 * read-write since it became active?  If so, now is the
	 * time to retire it back to read-only state.
	 */

	if (subfs_is_rw(sbi)) {
		int retval = subfs_remount_ro(sb);
		BUG_ON(retval);
	}
	supermount_unlock_door(sb);
	sbi->s_state = SUPERMOUNT_SUSPENDED;
out:
	atomic_dec(&sbi->s_opencount);
	up(&sbi->sem);
}

/*
 * Try to put the filesystem into writable mode for the duration of
 * an open(O_RDWR) or an inode operation requiring write access.  
 */

int
subfs_get_write_access(struct super_block *sb)
{
	int retval;
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	down(&sbi->sem);
	if (!subfs_is_online(sbi))
		BUG();

	retval = 0;
	if (atomic_read(&sbi->s_writecount) == 0) {
		/*
		 * We may already have write access to the filesystem
		 */
		if (!subfs_is_rw(sbi)) {
			if (sb->s_flags & MS_RDONLY)
				retval = -EROFS;
			else
				retval = subfs_remount_rw(sb);
		}
	}

	if (!retval)
		atomic_inc(&sbi->s_writecount);
	up(&sbi->sem);

	return retval;
}

/*
 * Once we're done, release write access.
 */
void
subfs_put_write_access(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	down(&sbi->sem);
	if (!subfs_is_rw(sbi))
		BUG();

	BUG_ON(atomic_read(&sbi->s_writecount) == 0);
	atomic_dec(&sbi->s_writecount);
	up(&sbi->sem);
}

/* Supermount open/close inode.  These functions maintain the
   superblock reference counts of active inodes; the subfs is
   suspended when that count reaches zero.

   Files are opened on any significant reference, and are not closed
   until they become fully dereferenced (during the last iput).
   Directories, on the other hand, are always closed unless they are
   active and opened; a directory referenced as CWD is not open.
*/

void
subfs_get_read_access(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	down(&sbi->sem);
	BUG_ON(!subfs_is_online(sbi));
	atomic_inc(&sbi->s_opencount);
	up(&sbi->sem);
}

/* we can't check here for subfs_is_online, because in the case of one
 * error, system will call ->release() on inodes, where subfs is not
 * online.
 */

void
subfs_put_read_access(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	down(&sbi->sem);
	atomic_dec(&sbi->s_opencount);
	up(&sbi->sem);
}
