/*
 *  linux/fs/supermount/super_operations.c
 *
 *  Original version:
 *      Copyright (C) 1995, 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/inode.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/super.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2, 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *
 */

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/list.h>
#include <linux/slab.h>

#include "supermount_i.h"

static struct super_operations supermount_sops;

static void
supermount_write_inode(struct inode *inode, int sync)
{
	struct supermount_sb_info *sbi = supermount_sbi(inode->i_sb);
	struct super_block *subsb;
	struct dentry *dent;
	struct inode *subi;
	int rc = -EIO;

	if (subfs_go_online(inode->i_sb))
		goto out;

	rc = 0;
	dent = get_subfs_dentry_by_inode(inode);
	if (IS_ERR(dent))
		goto out;
	subi = dent->d_inode;
	if (!subi)
		goto dput_ret;
	subsb = supermount_subfs_sb(sbi);
	if (subi && subsb && subsb->s_op && subsb->s_op->write_inode) {
		subi->i_mode = inode->i_mode;
		subi->i_uid = inode->i_uid;
		subi->i_gid = inode->i_gid;
		subi->i_size = inode->i_size;
		subi->i_atime = inode->i_atime;
		subi->i_mtime = inode->i_mtime;
		subi->i_ctime = inode->i_ctime;
		rc = subfs_get_write_access(inode->i_sb);
		if (rc)
			goto dput_ret;;
		subi->i_state |= I_LOCK;
		subsb->s_op->write_inode(subi, sync);
		subi->i_state &= ~I_LOCK;
		mark_subfs_dirty(inode->i_sb);
		subfs_put_write_access(inode->i_sb);
	}
dput_ret:
	dput(dent);
out:
	subfs_go_inactive(inode->i_sb);
}

static void
supermount_clear_inode(struct inode *inode)
{
	dput(supermount_i(inode)->i_dentry);
	list_del(&supermount_i(inode)->i_fs);
}

static struct supermount_sb_info *
create_sbi(struct super_block *sb)
{
	struct supermount_sb_info *sbi = kmalloc(sizeof (*sbi), GFP_KERNEL);

	if (!sbi)
		return NULL;

	memset(sbi, 0, sizeof (*sbi));

	sbi->s_state = SUPERMOUNT_UNMOUNTED;
	sbi->s_default_mode = 0777 | S_IFDIR;
	atomic_set(&sbi->s_opencount, 0);
	atomic_set(&sbi->s_writecount, 0);
	sbi->s_mflags = sb->s_flags & MS_RMT_MASK;
	INIT_LIST_HEAD(&sbi->s_inodes);
	sema_init(&sbi->sem, 1);
	return sbi;
}

static void
free_sbi(struct supermount_sb_info *sbi)
{
	if (sbi->s_type)
		kfree(sbi->s_type);
	if (sbi->s_devname)
		kfree(sbi->s_devname);
	if (sbi->s_data)
		kfree(sbi->s_data);
	if (sbi->s_data_copy)
		kfree(sbi->s_data_copy);
	kfree(sbi);
	
}

/* Release the superblock and any resources it has reserved 
 * it is called with the sb locked
 */

static void
supermount_put_super(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	if (subfs_is_mounted(sbi))
		subfs_umount(sb);

	sb->u.generic_sbp = 0;
	free_sbi(sbi);
	sb->s_dev = 0;
}

static int
copy_option(char **option, const char *val)
{
	char *tmp;
	if (!val || !*val)
		return -EINVAL;
	tmp = kmalloc(1 + strlen(val), GFP_KERNEL);
	if (!tmp)
		return -ENOMEM;
	strcpy(tmp, val);
	*option = tmp;
	return 0;
}

static int
parse_options(char *options, struct supermount_sb_info *sbi)
{
	char *this_char;
	char *value;
	int rc;

	if (!options)
		return 0;
	while ((this_char = options)) {
		if (!strncmp(this_char, "--", 2))
			this_char += 2;
		if (!*this_char)
			break;
		if (*this_char == ',') {
			/* An empty option, or the option "--",
			   introduces options to be passed through to
			   the subfs */
			return copy_option(&sbi->s_data, ++this_char);
		}
		if ((options = strchr(this_char, ',')))
			*options++ = 0;

		if ((value = strchr(this_char, '=')))
			*value++ = 0;

		if (!strcmp(this_char, "fs")) {
			rc = copy_option(&sbi->s_type, value);
			if (rc)
				return rc;
		} else if (!strcmp(this_char, "dev")) {
			rc = copy_option(&sbi->s_devname, value);
			if (rc)
				return rc;
		} else {
			printk(KERN_DEBUG "Unrecognized mount option %s",
					 this_char);
			return -EINVAL;
		}
	}
	return 0;
}

static struct inode *
supermount_root_inode(struct super_block *s)
{
	struct inode *inode = supermount_create_inode(s);

	if (!inode)
		return NULL;

	inode->i_version = 0;
	inode->i_op = &supermount_root_iops;
	inode->i_fop = &supermount_dir_operations;

	return inode;
}

/* read_super: the main mount() entry point into the VFS layer. */
struct super_block *
supermount_read_super(struct super_block *s, void *data, int silent)
{
	struct inode *root_inode;
	struct dentry *root;
	struct supermount_sb_info *sbi = create_sbi(s);

	if (!sbi)
		goto fail_no_memory;

	if (parse_options((char *) data, sbi))
		goto fail_free_sbi;

	if (!sbi->s_devname) {
		printk(KERN_ERR "%s: no dev=<filesystem> option", __FUNCTION__);
		goto fail_free_sbi;
	}
	if (sbi->s_data) {
		sbi->s_data_copy = kmalloc(1 + strlen(sbi->s_data), GFP_KERNEL);
		if (!sbi->s_data_copy) {
			printk(KERN_ERR "%s: not enough memory", __FUNCTION__);
			goto fail_free_sbi;
		}
		strcpy(sbi->s_data_copy, sbi->s_data);
	} else
		sbi->s_data_copy = NULL;

	s->s_blocksize = 1024;
	s->s_blocksize_bits = 10;
	s->s_magic = SUPERMOUNT_SUPER_MAGIC;
	s->s_op = &supermount_sops;
	s->u.generic_sbp = sbi;

	root_inode = supermount_root_inode(s);
	if (!root_inode)
		goto fail_free_sbi;

	root = d_alloc_root(root_inode);
	if (!root)
		goto fail_free_root_inode;

	s->s_root = root;
	s->s_root->d_op = &supermount_dir_dops;
	dget(root);

	return s;

fail_free_root_inode:
	iput(root_inode);
fail_free_sbi:
	s->s_dev = 0;
	free_sbi(sbi);
fail_no_memory:
	return NULL;
}

/*
 * called with lock_super held
 */

static void
supermount_write_super(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	if (subfs_is_online(sbi) && subfs_is_rw(sbi)) {
		struct super_block *subsb = supermount_subfs_sb(sbi);
		if (subsb && subsb->s_op && subsb->s_op->write_super)
			subsb->s_op->write_super(subsb);
		mark_subfs_clean(sb);
	}
}

static int
supermount_statfs(struct super_block *sb, struct statfs *buf)
{
	int rc = 0;
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	if (subfs_is_online(sbi)) {
		struct super_block *subsb = supermount_subfs_sb(sbi);
		if (subsb->s_op && subsb->s_op->statfs)
			rc = subsb->s_op->statfs(subsb, buf);
	}
	buf->f_type = SUPERMOUNT_SUPER_MAGIC;
	return rc;
}

static struct super_operations supermount_sops = {
	.write_inode	= supermount_write_inode,
	.clear_inode	= supermount_clear_inode,
	.put_super	= supermount_put_super,
	.write_super	= supermount_write_super,
	.statfs		= supermount_statfs,
};
