#ifndef _SUPERMOUNT_I_H
#define _SUPERMOUNT_I_H

#define SUPERMOUNT_DEBUG

/*
 * The supermount superblock magic number
 */

#define SUPERMOUNT_SUPER_MAGIC	0x9fa1

static inline int
is_inode_supermounted(struct inode *inode)
{
	return (inode->i_sb &&
		(inode->i_sb->s_magic == SUPERMOUNT_SUPER_MAGIC));
}

static inline int
is_dentry_supermounted(struct dentry *dentry)
{
	return (dentry->d_inode && is_inode_supermounted(dentry->d_inode));
}

#define SUPERMOUNT_VERSION		"0.7"

/*
 * The difference between SUSPENDED and ONLINE is that in SUSPENDED
 * the door is unlocked, and in ONLINE it is locked.
 */

typedef enum {
	SUPERMOUNT_UNMOUNTED,	/* No media mounted */
	SUPERMOUNT_SUSPENDED,	/* Mounted but suspended because 
				   no files open */
	SUPERMOUNT_ONLINE,	/* Mounted and active */
} sm_state_t;

/*
 * supermount super-block data in memory
 */
struct supermount_sb_info {
	sm_state_t s_state;
	int s_version;		/* Used to indicate obsolete inodes */
	mode_t s_default_mode;	/* Default mode for supermount root */

	char *s_type;		/* Type of fs to be sub-mounted */
	char *s_devname;	/* Where to mount the subfs from */
	/* unused */
	int s_mflags;		/* Flags to pass when mounting subfs */
	char *s_data;		/* Data to pass when mounting subfs */
	char *s_data_copy;	/* copy of previous data.  FS
				   destroy that data, and we
				   need to mount several
				   times.  We could allocate
				   the memory in
				   supermount_mount, but then
				   we have more possible
				   errors */
	struct vfsmount *s_undermount;	/* Mount point for subfs */
	atomic_t s_opencount;	/* Refcount of opened inodes
				   (an inode in use as cwd
				   does not count as open) */
	atomic_t s_writecount;	/* Refcount of write accesses
				   on the filesystem */
	/* Flags */
	int s_dirty:1;		/* Do we need to fsync() the subfs? */
	int s_rw:1;		/* Is the subfs mounted read/write? */
	struct list_head s_inodes;
	struct semaphore sem;
};

static inline struct supermount_inode_info *
supermount_i(struct inode *inode)
{
	return (&inode->u.supermount_i);
}

static inline struct supermount_sb_info *
supermount_sbi(struct super_block *sb)
{
	if (!sb)
		BUG();
	return (struct supermount_sb_info *) (sb->u.generic_sbp);
}

static inline struct super_block *
supermount_subfs_sb(struct supermount_sb_info *sbi)
{
	return sbi->s_undermount->mnt_sb;
}

/*
 * Function prototypes
 */

static inline struct file *
get_subfile(struct file *file)
{
	struct file *subfile = (struct file *) file->private_data;
	if (!subfile)
		BUG();
	return subfile;
}

static inline void
put_subfile(struct file *file, struct file *subfile)
{
	file->private_data = (void *) subfile;
}

/* How to test if an inode is obsolete: */
static inline int
is_inode_obsolete(struct inode *inode)
{
	struct supermount_sb_info *sbi = supermount_sbi(inode->i_sb);

	return (supermount_i(inode)->i_sb_version != sbi->s_version);
}

/* Manage the subfs dirty flag */
static inline void
mark_subfs_dirty(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	sbi->s_dirty = 1;
}
static inline void
mark_subfs_clean(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	sbi->s_dirty = 0;
}
static inline int
subfs_is_dirty(struct supermount_sb_info *sbi)
{
	return (sbi->s_dirty);
}

/* Manage the subfs rw flag */
static inline void
mark_subfs_rw(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	sbi->s_rw = 1;
}
static inline void
unmark_subfs_rw(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	sbi->s_rw = 0;
}
static inline int
subfs_is_rw(struct supermount_sb_info *sbi)
{
	return (sbi->s_rw);
}

static inline int
subfs_is_online(struct supermount_sb_info *sbi)
{
	return (sbi->s_state == SUPERMOUNT_ONLINE);
}

static inline int
subfs_is_mounted(struct supermount_sb_info *sbi)
{
	return (sbi->s_state != SUPERMOUNT_UNMOUNTED);
}

/* super.c */
extern struct inode *supermount_create_inode(struct super_block *s);
extern void subfs_umount(struct super_block *sb);
extern int subfs_go_online(struct super_block *);
extern void subfs_go_inactive(struct super_block *);
extern int subfs_get_write_access(struct super_block *);
extern void subfs_put_write_access(struct super_block *);
extern void subfs_get_read_access(struct super_block *);
extern void subfs_put_read_access(struct super_block *);
/* translate.c */
extern struct dentry *get_subfs_dentry(struct dentry *dentry);
extern struct dentry *get_subfs_dentry_by_inode(struct inode *inode);
/* mediactl.c */
extern void supermount_handle_door(struct super_block *, int);
static inline void
supermount_lock_door(struct super_block *sb)
{
	supermount_handle_door(sb, MEDIA_LOCK);
}
static inline void
supermount_unlock_door(struct super_block *sb)
{
	supermount_handle_door(sb, MEDIA_UNLOCK);
}

/* inode_operations.c */
extern struct inode_operations supermount_root_iops;
extern struct inode_operations supermount_dir_iops;
extern struct inode_operations supermount_file_iops;
extern struct inode_operations supermount_symlink_iops;
/* super_operations.c */
extern struct super_block *supermount_read_super(struct super_block *,
						 void *, int);
/* file_operations.c */
extern struct file_operations supermount_dir_operations;
extern struct file_operations supermount_file_operations;
/* dentry_operations.c */
extern struct dentry_operations supermount_file_dops;
extern struct dentry_operations supermount_dir_dops;

#endif				/* _SUPERMOUNT_I_H */
