/*
 *  linux/fs/supermount/translate.c
 *
 *  Original version:
 *      Copyright (C) 1995, 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/dir.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/dir.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2 & 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *  Rewriten for kernel 2.4. (C) 2001 MandrakeSoft Inc.
 *                                    Juan Quintela (quintela@mandrakesoft.com)
 *
 */

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/string.h>

#include "supermount_i.h"

static int
unwalk_dentry(struct dentry *dentry, char *name)
{
	char *tmp;
	char *path;
	int number = 0;

	path = name + PATH_MAX;
	*path = 0;

	spin_lock(&dcache_lock);
	for (;;) {
		if (dentry->d_parent == dentry)
			break;
		path -= dentry->d_name.len + 1;
		if (path < name)
			goto err_out;
		memcpy(path + 1, dentry->d_name.name, dentry->d_name.len);
		*path = 0;
		number++;
		dentry = dentry->d_parent;
	}
	spin_unlock(&dcache_lock);
	if (number)
		/* we need to removing leading '\0' */
		path++;

	tmp = name;
	while (path <= name + PATH_MAX)
		*tmp++ = *path++;

	return number;

err_out:
	spin_unlock(&dcache_lock);
	return -EINVAL;
}

static struct dentry *
walk_path(struct dentry *root, char *path, int number)
{
	struct dentry *base = root;

	while (number-- && !IS_ERR(base)) {
		struct dentry *subdent;
		int len = strlen(path);
		subdent = lookup_one_len(path, base, len);
		path += len + 1;
		dput(base);
		base = subdent;
	}
	return base;
}

static struct dentry *
__get_subfs_dentry(struct dentry *dentry)
{
	struct supermount_sb_info *sbi;
	struct dentry *subdent = ERR_PTR(-EINVAL);
	struct dentry *root;
	int number;
	char *name;
	struct vfsmount *mnt;

	name = __getname();
	if (!name)
		return ERR_PTR(-ENOMEM);

	sbi = supermount_sbi(dentry->d_sb);
	number = unwalk_dentry(dentry, name);
	if (number < 0) {
		subdent = ERR_PTR(number);
		goto out;
	}
	mnt = mntget(sbi->s_undermount);
	if (!mnt)
		goto out;

	root = dget(mnt->mnt_sb->s_root);
	subdent = walk_path(root, name, number);
	mntput(mnt);
out:
	putname(name);
	return subdent;
}

struct dentry *
get_subfs_dentry_by_inode(struct inode *inode)
{
	struct dentry *dentry;
	struct supermount_sb_info *sbi;

	BUG_ON(!inode);
	BUG_ON(!is_inode_supermounted(inode));

	if (is_inode_obsolete(inode))
		return ERR_PTR(-ESTALE);

	sbi = supermount_sbi(inode->i_sb);
	if (!subfs_is_mounted(sbi))
		return ERR_PTR(-EINVAL);

	dentry = dget(supermount_i(inode)->i_dentry);
	if (!dentry)
		BUG();

	return dentry;
}

struct dentry *
get_subfs_dentry(struct dentry *dentry)
{
	struct dentry *subdent;

	BUG_ON(!dentry);
	BUG_ON(dentry->d_inode && !is_dentry_supermounted(dentry));

	if (dentry->d_inode && !is_inode_obsolete(dentry->d_inode))
		subdent = get_subfs_dentry_by_inode(dentry->d_inode);
	else
		subdent = __get_subfs_dentry(dentry);

	return subdent;
}
