/* 
 * grsecurity/gracl.c
 * Copyright Michael Dalton, Brad Spengler 2000, 2001, 2002
 *
 */

#include <linux/kernel.h>
#include <linux/mm.h>
#include <asm/uaccess.h>
#include <asm/errno.h>
#include <asm/mman.h>
#include <linux/file.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/types.h>
#include <linux/capability.h>
#include <linux/sched.h>
#include <linux/gracl.h>
#include <linux/gralloc.h>
#include <linux/sysctl.h>

static struct acl_subject_db acl_subj_set;
static struct acl_object_db acl_obj_set;
static struct name_db name_set;
static struct name_db inodev_set;

static struct admin_pw pwent; /* password entry */

static spinlock_t gr_proc_lock = SPIN_LOCK_UNLOCKED;
rwlock_t gr_inode_lock = RW_LOCK_UNLOCKED;

/* give each cpu its own shared page for performance reasons */
static char *gr_shared_page[NR_CPUS];

static unsigned long gr_status = GR_STATUS_INIT;

extern int chkpw(struct gr_arg *entry, struct admin_pw * pwent);

#ifdef CONFIG_GRKERNSEC_RESLOG
extern __inline__ void gr_log_resource(const struct task_struct *task,
				const int res, const unsigned long wanted);
#endif

#ifdef CONFIG_GRKERNSEC
int gr_disable = 0;
#else
int gr_disable = 1;
#endif

/* The special mode : admin(called god_label for historical reasons) mode */

static struct acl_subject_label *god_label = NULL;
static struct acl_subject_label *kernel_label = NULL;
static struct acl_subject_label *root_label = NULL;

/* The following are used to keep a place held in the hash table when we move
   entries around.  They can be replaced during insert. */

static struct acl_subject_label *deleted_subject;
static struct acl_object_label *deleted_object;
static struct name_entry *deleted_inodev;

extern int gr_init_uidset(void);
extern void gr_free_uidset(void);
extern void gr_remove_uid(uid_t uid);
extern int gr_find_uid(uid_t uid);

#ifdef CONFIG_GRKERNSEC_ACL_SUPERDEBUG
static void list_procs(void);
static void list_names(void);
#endif

__inline__ int gr_acl_is_enabled(void)
{
	return (gr_status & GR_READY);
}

static __inline__ char * d_real_path(const struct dentry * dentry, const struct vfsmount * vfsmnt,
				     char * buf, int buflen)
{
	char *res;
	struct dentry *our_dentry;
	struct vfsmount *our_mount;
	struct vfsmount *rootmnt;
	struct dentry *root;

	our_dentry  = (struct dentry *)dentry;
	our_mount = (struct vfsmount *)vfsmnt;

	read_lock(&child_reaper->fs->lock);
	rootmnt = mntget(child_reaper->fs->rootmnt);
	root = dget(child_reaper->fs->root);
	read_unlock(&child_reaper->fs->lock);

	spin_lock(&dcache_lock);
	res = __d_path(our_dentry, our_mount, root, rootmnt, buf, buflen);
	spin_unlock(&dcache_lock);
	dput(root);
	mntput(rootmnt);
	return res;
}

__inline__ struct acl_subject_label *
lookup_acl_subj_label(const ino_t ino, const kdev_t dev)
{
	unsigned long index = fhash(ino, dev, acl_subj_set.s_size);
	struct acl_subject_label *match;
	__u8 i = 0;

	match = acl_subj_set.s_hash[index];

	while (match && (match->inode != ino || match->device != dev)) {
		index = (index + (1 << i)) % acl_subj_set.s_size;
		match = acl_subj_set.s_hash[index];
		i = (i + 1) % 32;
	}

	if (unlikely(match && (match != deleted_subject) &&
		     (match->inode == ino) && (match->device == dev)))
		return match;
	else
		return NULL;
}

static __inline__ struct acl_object_label *
lookup_acl_obj_label(const ino_t ino, const kdev_t dev,
		     const unsigned long parent_pos)
{
	unsigned long index = phash(ino, dev, parent_pos, acl_obj_set.o_size);
	struct acl_object_label *match;
	__u8 i = 0;

	match = acl_obj_set.o_hash[index];

	while (match && (match->inode != ino || match->device != dev ||
	       match->parent_pos != parent_pos)) {
		index = (index + (1 << i)) % acl_obj_set.o_size;
		match = acl_obj_set.o_hash[index];
		i = (i + 1) % 32;
	}
	
	if (unlikely(match && (match != deleted_object) &&
		     (match->inode == ino) && (match->device == dev) &&
		     (match->parent_pos == parent_pos)))
		return match;
	else
		return NULL;
}

static __inline__ struct name_entry *
lookup_name_entry(const char *name)
{
	unsigned long index = nhash(name, name_set.n_size);
	struct name_entry *match;
	__u8 i = 0;

	match = name_set.n_hash[index];

	while (match && (strcmp(match->name, name)) != 0) {
		index = (index + (1 << i)) % name_set.n_size;
		match = name_set.n_hash[index];
		i = (i + 1) % 32;
	}

	if (unlikely(match && !strcmp(name, match->name)))
		return match;
	else
		return NULL;
}


static __inline__ struct name_entry *
lookup_inodev_entry(const ino_t ino, const kdev_t dev)
{
	unsigned long index = fhash(ino, dev, inodev_set.n_size);
	struct name_entry *match;
	__u8 i = 0;

	match = inodev_set.n_hash[index];

	while (match && (match->inode != ino || match->device != dev)) {
		index = (index + (1 << i)) % inodev_set.n_size;
		match = inodev_set.n_hash[index];
		i = (i + 1) % 32;
	}

	if (unlikely(match && (match != deleted_inodev) &&
		     (match->inode == ino) && (match->device == dev)))
		return match;
	else
		return NULL;
}

/* The following routines perform insertion into the tables after
 * the add_{type} process the config file */

static void
insert_inodev_entry(struct name_entry *nentry)
{
	unsigned long index = fhash(nentry->inode, nentry->device, 
					inodev_set.n_size);
	struct name_entry **curr;
	__u8 i = 0;

	curr = &inodev_set.n_hash[index];

	while (*curr && *curr != deleted_inodev) {
		index = (index + (1 << i)) % inodev_set.n_size;
		curr = &inodev_set.n_hash[index];
		i = (i + 1) % 32;
	}

	*curr = nentry;

	return;
}

static int
insert_name_entry(char *name, const ino_t inode, const kdev_t device)
{
	unsigned long index = nhash(name, name_set.n_size);
	struct name_entry **curr;
	__u8 i = 0;

	curr = &name_set.n_hash[index];

	while (*curr && strcmp(name, (*curr)->name)) {
		index = (index + (1 << i)) % name_set.n_size;
		curr = &name_set.n_hash[index];
		i = (i + 1) % 32;
	}

	if (!(*curr)) {
		struct name_entry *nentry = acl_alloc(sizeof(struct name_entry));
		if (!nentry)
			return 0;
		nentry->name = name;
		nentry->inode = inode;
		nentry->device = device;
		*curr = nentry;
		/* insert us into the table searchable by inode/dev */
		insert_inodev_entry(nentry);
	}

	return 1;
}

static void
insert_acl_obj_label(struct acl_object_label *obj)
{
	unsigned long index = phash(obj->inode, obj->device, obj->parent_pos,
					acl_obj_set.o_size);
	struct acl_object_label **curr;
	__u8 i = 0;

	curr = &acl_obj_set.o_hash[index];

	while (*curr && *curr != deleted_object) {
		index = (index + (1 << i)) % acl_obj_set.o_size;
		curr = &acl_obj_set.o_hash[index];
		i = (i + 1) % 32;
	}

	*curr = obj;
	acl_obj_set.o_used++;

	return;
}

static void
insert_acl_subj_label(struct acl_subject_label *obj)
{
	unsigned long index = fhash(obj->inode, obj->device, acl_subj_set.s_size);
	struct acl_subject_label **curr;
	__u8 i = 0;

	curr = &acl_subj_set.s_hash[index];

	while (*curr && *curr != deleted_subject) {
		index = (index + (1 << i)) % acl_subj_set.s_size;
		curr = &acl_subj_set.s_hash[index];
		i = (i + 1) % 32;
	}

	*curr = obj;
	acl_subj_set.s_used++;

	return;
}

/* Create a table(requesting size close to 2^pwr entrie of member size)
 * and replace the current table size with the new one.
 * We use prime table sizes until we reach the bounds of 32-bit word machines 
 * Once we reach there we simply double the table size */

static void **
create_table(__u32 *len)
{
	unsigned long table_sizes[] = {
		7, 13, 31, 61, 127, 251, 509, 1021, 2039, 4093, 8191, 16381, 
		32749, 65521, 131071, 262139, 524287, 1048573, 2097143,
		4194301, 8388593, 16777213, 33554393, 67108859, 134217689, 
		268435399, 536870909, 1073741789, 2147483647
	};
	void * newtable = NULL;
	unsigned int pwr = 0;

	while ((pwr < ((sizeof(table_sizes)/sizeof(table_sizes[0])) - 1)) && 
			table_sizes[pwr] <= (2 * (*len)))
		pwr++;

	if(table_sizes[pwr] <= (2 * (*len)))
		return newtable;

	if((table_sizes[pwr] * sizeof(void *)) <= PAGE_SIZE)
		newtable = kmalloc(table_sizes[pwr] * sizeof(void *), GFP_KERNEL);
	else
		newtable = vmalloc(table_sizes[pwr] * sizeof(void *));

	*len = table_sizes[pwr];

	return newtable;
}


#ifdef CONFIG_GRKERNSEC_ACL_SUPERDEBUG
static void
list_names(void)
{
	unsigned long i;
	struct name_entry *curr;

	for (i = 0; i < name_set.n_size; i++) {
		curr = name_set.n_hash[i];
		if(!curr) continue;

		security_debug(GR_LISTNAMES_DEBUG_MSG,
				i, curr->inode,
				kdevname(curr->device),
				curr->name);
	}

	return;
}

static void
list_procs(void)
{
	unsigned long i;
	struct acl_subject_label *curr;
	struct acl_object_label *curr2;
	for (i = 0; i < acl_subj_set.s_size; i++) {
		curr = acl_subj_set.s_hash[i];
		if (!curr) continue;
		security_debug(GR_LISTPROCSS_DEBUG_MSG,
				i, curr->inode,
				kdevname(curr->device),
				curr->mode, curr->pos);
	}
	for (i = 0; i < acl_obj_set.o_size; i++) {
		curr2 = acl_obj_set.o_hash[i];
		if (!curr2) continue;
		security_debug(GR_LISTPROCSO_DEBUG_MSG, i,
			       curr2->inode, kdevname(curr2->device),
			       curr2->mode, curr2->parent_pos);
	}

	return;
}

#endif

static int
init_variables(const unsigned long acl_obj_size, const unsigned long acl_subj_size,
	       const unsigned long acl_ip_size)
{
	unsigned long stacksize;
	unsigned int i;

	acl_subj_set.s_size = acl_subj_size;
	acl_obj_set.o_size = acl_obj_size;
	name_set.n_size = (acl_obj_size + acl_subj_size);
	inodev_set.n_size = (acl_obj_size + acl_subj_size);

	if(!gr_init_uidset()) return 1;

	/* set up the stack that holds allocation info */

	stacksize = (3 * acl_obj_size) + (4 * acl_subj_size) + acl_ip_size + 3;

	if (!acl_alloc_stack_init(stacksize))
		return 1;

	/* create our shared page */
	
	for(i = 0; i < NR_CPUS; i++) {
		gr_shared_page[i] = (char *)__get_free_page(GFP_KERNEL);
		if (!gr_shared_page[i])
			return 1;
	}

	/* create our empty, fake deleted acls */
	deleted_subject = (struct acl_subject_label *)acl_alloc(sizeof(struct acl_subject_label));
	deleted_object = (struct acl_object_label *)acl_alloc(sizeof(struct acl_object_label));
	deleted_inodev = (struct name_entry *)acl_alloc(sizeof(struct name_entry));
	
	if (!deleted_subject || !deleted_object || !deleted_inodev)
		return 1;

	memset(deleted_subject, 0, sizeof(struct acl_subject_label));
	memset(deleted_object, 0, sizeof(struct acl_object_label));
	memset(deleted_inodev, 0, sizeof(struct name_entry));

	/* We only want 50% full tables for now */

	acl_subj_set.s_hash = (struct acl_subject_label **)create_table(&acl_subj_set.s_size);
	acl_obj_set.o_hash = (struct acl_object_label **)create_table(&acl_obj_set.o_size);
	name_set.n_hash = (struct name_entry **)create_table(&name_set.n_size);
	inodev_set.n_hash = (struct name_entry **)create_table(&inodev_set.n_size);

	if (!acl_subj_set.s_hash || !acl_obj_set.o_hash || !name_set.n_hash ||
	    !inodev_set.n_hash)
		return 1;
	memset(acl_subj_set.s_hash, 0,
	       sizeof (struct acl_subject_label *) * acl_subj_set.s_size);
	memset(acl_obj_set.o_hash, 0,
	       sizeof (struct acl_object_label *) * acl_obj_set.o_size);
	memset(name_set.n_hash, 0,
	       sizeof (struct name_entry *) * name_set.n_size);
	memset(inodev_set.n_hash, 0,
	       sizeof (struct name_entry *) * inodev_set.n_size);

	acl_subj_set.s_used = acl_obj_set.o_used = 0;

	return 0;
}

static void
free_variables(void)
{
	unsigned int i;

	acl_free_all();

	for (i = 0; i < NR_CPUS; i++) {
		if (!gr_shared_page[i])
			break;
		else
			free_page((unsigned long)gr_shared_page[i]);
	}

	if(acl_subj_set.s_hash) {
		if((acl_subj_set.s_size * sizeof(struct acl_subject_label *)) <= PAGE_SIZE)
			kfree(acl_subj_set.s_hash);
		else
			vfree(acl_subj_set.s_hash);
	}
	if(acl_obj_set.o_hash) {
		if((acl_obj_set.o_size * sizeof(struct acl_object_label *)) <= PAGE_SIZE)
			kfree(acl_obj_set.o_hash);
		else
			vfree(acl_obj_set.o_hash);
	}
	if(name_set.n_hash) {
		if((name_set.n_size * sizeof(struct name_entry *)) <= PAGE_SIZE)
			kfree(name_set.n_hash);
		else
			vfree(name_set.n_hash);
	}

	if(inodev_set.n_hash) {
		if((inodev_set.n_size * sizeof(struct name_entry *)) <= PAGE_SIZE)
			kfree(inodev_set.n_hash);
		else
			vfree(inodev_set.n_hash);
	}

	gr_free_uidset();	

	return;
}

static void
grsecurity_disable(void)
{
	struct task_struct * task;

	read_lock(&tasklist_lock);
	for_each_task(task) {
		task->acl_admin = 0;
		task->acl = NULL;
	}
	read_unlock(&tasklist_lock);

	free_variables();

	memset(&name_set, 0, sizeof(struct name_db));
	memset(&inodev_set, 0, sizeof(struct name_db));
	memset(&acl_obj_set, 0, sizeof(struct acl_object_db));
	memset(&acl_subj_set, 0, sizeof(struct acl_subject_db));

	kernel_label = NULL;
	god_label = NULL;
	root_label = NULL;
		
	return;
}

static int
copy_user_acl(struct gr_arg *arg)
{
	struct acl_object_label *o_tmp, **o_utmp, *o_utmp2;
	struct acl_subject_label *s_tmp, **s_utmp, *s_utmp2;
	struct acl_ip_label **i_tmp, *i_utmp2;
	unsigned long s_num, o_num, i_num;
	char *tmp;
	int err = 0;

	s_utmp = (struct acl_subject_label **)arg->subj_db.s_table;
	o_utmp = (struct acl_object_label **)arg->obj_db.o_table;

	for(o_num = 0;o_num < arg->obj_db.o_entries;o_num++) {
		if ((o_tmp = (struct acl_object_label *)
			acl_alloc(sizeof (struct acl_object_label))) == NULL) {
			err = -ENOMEM;
			goto cleanup;
		}
		if (copy_from_user(&o_utmp2, o_utmp + o_num, 
	     		sizeof (struct acl_object_label *))) {
			err = -EFAULT;
			goto cleanup;
		}
		if (copy_from_user(o_tmp, o_utmp2, 
	     		sizeof (struct acl_object_label))) {
			err = -EFAULT;
			goto cleanup;
		}
		if ((tmp = (char *)acl_alloc
			(o_tmp->filename_len)) == NULL) {
			err = -ENOMEM;
			goto cleanup;
		}
		if (copy_from_user(tmp, 
			o_tmp->filename, o_tmp->filename_len)) {
			err = -EFAULT;
			goto cleanup;
		}
		o_tmp->filename = tmp;

		insert_acl_obj_label(o_tmp);
		if(!insert_name_entry(o_tmp->filename, o_tmp->inode,
				      o_tmp->device)) {
			err = -ENOMEM;
			goto cleanup;
		}
	}
	for(s_num = 0;s_num < arg->subj_db.s_entries;s_num++) {
		if ((s_tmp = acl_alloc(sizeof (struct acl_subject_label))) == NULL) {
			err = -ENOMEM;
			goto cleanup;
		}
		if (copy_from_user(&s_utmp2, s_utmp + s_num,
		 		sizeof (struct acl_subject_label *))) {
			err = -EFAULT;
			goto cleanup;
		}
		if (copy_from_user(s_tmp, s_utmp2,
		 		sizeof (struct acl_subject_label))) {
			err = -EFAULT;
			goto cleanup;
		}
 		if ((tmp = (char *)acl_alloc(s_tmp->filename_len)) == NULL) {
			err = -ENOMEM;
			goto cleanup;
		}
		if (copy_from_user(tmp, s_tmp->filename, 
	     			s_tmp->filename_len)) {
			err = -EFAULT;
			goto cleanup;
		}
		s_tmp->filename = tmp;

		if (!(strcmp(s_tmp->filename, "god")))
			god_label = s_tmp;
		if (!(strcmp(s_tmp->filename, "kernel")))
			kernel_label = s_tmp;
		else if (!(strcmp(s_tmp->filename, "/")))
			root_label = s_tmp;

		/* add in ip acls */

		if(!s_tmp->ip_num) {
			s_tmp->ips = NULL;
			goto insert;
		}

		if ((i_tmp = (struct acl_ip_label **) acl_alloc(s_tmp->ip_num * sizeof (struct acl_ip_label *))) == NULL) {
			err = -ENOMEM;
			goto cleanup;
		}

		for(i_num = 0; i_num < s_tmp->ip_num; i_num++) {
	 		if ((*(i_tmp + i_num) = (struct acl_ip_label *) acl_alloc(sizeof(struct acl_ip_label))) == NULL) {
				err = -ENOMEM;
				goto cleanup;
			}
			if (copy_from_user(&i_utmp2, s_tmp->ips + i_num, sizeof(struct acl_ip_label *))) {
				err = -EFAULT;
				goto cleanup;
			}
			if (copy_from_user(*(i_tmp + i_num), i_utmp2, sizeof(struct acl_ip_label))) {
				err = -EFAULT;
				goto cleanup;
			}
		}	

		s_tmp->ips = i_tmp;

insert:
		insert_acl_subj_label(s_tmp);
		if(!insert_name_entry(s_tmp->filename, s_tmp->inode,
				      s_tmp->device)) {
			err = -ENOMEM;
			goto cleanup;
		}
	}

	goto return_err;
cleanup:
	free_variables();
return_err:	
	return err;

}

static int
gracl_init(struct gr_arg *args)
{
	int error = 0;
	
	memcpy(pwent.salt, args->salt, sizeof(pwent.salt));
	memcpy(pwent.sum, args->sum, sizeof(pwent.sum));	

	if (args->subj_db.s_entries < 3)
		return -EINVAL;

	if (init_variables(args->obj_db.o_entries, args->subj_db.s_entries,
			   args->subj_db.i_entries)) {
		security_alert_good(GR_INITF_ACL_MSG, GR_INITF_ACL_FLD, GR_VERSION);
		error = -1;
		free_variables();
	}

	error = copy_user_acl(args);
	if(error) {
		free_variables();
		goto out;
	}

#ifdef CONFIG_GRKERNSEC_ACL_SUPERDEBUG
		list_procs();
		list_names();
#endif

	if((error = gr_set_acls(0))) {
		free_variables();
		goto out;
	}

	gr_status |= GR_READY;
out:
	return error;
}

/* * * * * * * * * * * * * * * * * * * * * * *
 * Begin Misc Section 
 * * * * * * * * * * * * * * * * * * * * * * */

static __u32
chk_obj_label(const struct dentry * l_dentry, const struct vfsmount * l_mnt,
		const __u32 reqmode, const struct acl_subject_label *acl)
{
	struct dentry * dentry = (struct dentry *)l_dentry;
	struct vfsmount * mnt = (struct vfsmount *)l_mnt;
	struct dentry *root;
	struct vfsmount *rootmnt;
	struct acl_object_label *retval;

	read_lock(&child_reaper->fs->lock);
	rootmnt = mntget(child_reaper->fs->rootmnt);
	root = dget(child_reaper->fs->root);
	read_unlock(&child_reaper->fs->lock);
	spin_lock(&dcache_lock);

	for (;;) {
		if (unlikely(dentry == root && mnt == rootmnt))
			break;
		if (unlikely(dentry == mnt->mnt_root || IS_ROOT(dentry))) {
			if (mnt->mnt_parent == mnt)
				break;
			
			read_lock(&gr_inode_lock);
			retval = lookup_acl_obj_label(dentry->d_inode->i_ino, dentry->d_inode->i_dev, acl->pos);
			read_unlock(&gr_inode_lock);
			if (unlikely(retval && !(retval->mode & GR_DELETED)))
				goto out;

			dentry = mnt->mnt_mountpoint;
			mnt = mnt->mnt_parent;
			continue;
		}

		read_lock(&gr_inode_lock);
		retval = lookup_acl_obj_label(dentry->d_inode->i_ino, dentry->d_inode->i_dev, acl->pos);
		read_unlock(&gr_inode_lock);
		if (unlikely(retval && !(retval->mode & GR_DELETED)))
				goto out;

		dentry = dentry->d_parent;
	}

	read_lock(&gr_inode_lock);
	retval = lookup_acl_obj_label(dentry->d_inode->i_ino, dentry->d_inode->i_dev, acl->pos);
	read_unlock(&gr_inode_lock);

	if (unlikely(!retval || (retval->mode & GR_DELETED))) {
		read_lock(&gr_inode_lock);
		retval = lookup_acl_obj_label(root->d_inode->i_ino, root->d_inode->i_dev, acl->pos);
		read_unlock(&gr_inode_lock);
	}
out:
	spin_unlock(&dcache_lock);
	dput(root);
	mntput(rootmnt);	

	return (retval->mode & reqmode);
}

static struct acl_subject_label *
chk_subj_label(const struct dentry * l_dentry, const struct vfsmount * l_mnt)
{
	struct dentry * dentry = (struct dentry *)l_dentry;
	struct vfsmount * mnt = (struct vfsmount *)l_mnt;
	struct dentry *root;
	struct vfsmount *rootmnt;
	struct acl_subject_label *retval;

	read_lock(&child_reaper->fs->lock);
	rootmnt = mntget(child_reaper->fs->rootmnt);
	root = dget(child_reaper->fs->root);
	read_unlock(&child_reaper->fs->lock);
	spin_lock(&dcache_lock);

	for (;;) {
		if (unlikely(dentry == root && mnt == rootmnt))
			break;
		if (unlikely(dentry == mnt->mnt_root || IS_ROOT(dentry))) {
			if (mnt->mnt_parent == mnt)
				break;
			
			read_lock(&gr_inode_lock);
			retval = lookup_acl_subj_label(dentry->d_inode->i_ino, dentry->d_inode->i_dev);
			read_unlock(&gr_inode_lock);
			if (unlikely(retval && !(retval->mode & GR_DELETED)))
				goto out;

			dentry = mnt->mnt_mountpoint;
			mnt = mnt->mnt_parent;
			continue;
		}

		read_lock(&gr_inode_lock);
		retval = lookup_acl_subj_label(dentry->d_inode->i_ino, dentry->d_inode->i_dev);
		read_unlock(&gr_inode_lock);
		if (unlikely(retval && !(retval->mode & GR_DELETED)))
			goto out;

		dentry = dentry->d_parent;
	}

	read_lock(&gr_inode_lock);
	retval = lookup_acl_subj_label(dentry->d_inode->i_ino, dentry->d_inode->i_dev);
	read_unlock(&gr_inode_lock);

	if (unlikely(!retval || (retval->mode & GR_DELETED))) {
		read_lock(&gr_inode_lock);
		retval = lookup_acl_subj_label(root->d_inode->i_ino, root->d_inode->i_dev);
		read_unlock(&gr_inode_lock);
	}
out:
	spin_unlock(&dcache_lock);
	dput(root);
	mntput(rootmnt);	

	return retval;
}

static __inline__ void
gr_log_learn(const struct acl_subject_label *curracl, const struct dentry *dentry,
		const struct vfsmount *mnt, const char *pathname, const __u32 mode)
{
	security_learn(GR_LEARN_AUDIT_MSG, curracl->device, curracl->inode,
			dentry->d_inode ? (unsigned long)dentry->d_inode->i_dev : 0UL,
			dentry->d_inode ? dentry->d_inode->i_ino : 0UL,
			pathname, (unsigned long)mode);

	return;
}

__u32
gr_search_file(const struct dentry *dentry, const __u32 mode, const struct vfsmount *mnt)
{
	__u32 retval = mode;
	struct acl_subject_label * curracl;

	if (unlikely(!(gr_status & GR_READY))) return retval;

	curracl = current->acl;

	retval = chk_obj_label(dentry, mnt, mode, curracl);

	if (unlikely((curracl->mode & GR_LEARN) && (retval != mode))) {
		char * pathname;

		retval = mode;

		pathname = d_real_path(dentry, mnt, gr_shared_page[current->processor], PAGE_SIZE);
		/* if someone needs ptrace, only learn read-only ptrace */
		if (mode == (GR_WRITE | GR_PTRACERD))
			gr_log_learn(curracl, dentry, mnt, pathname, mode & ~GR_WRITE);
		else
			gr_log_learn(curracl, dentry, mnt, pathname, mode);
	}

	return retval;
}

__u32
gr_check_create(const struct dentry *new_dentry, const struct dentry *parent,
		const struct vfsmount *mnt, const __u32 mode)
{
	struct name_entry *match;
	struct acl_object_label *matchpo;
	struct acl_subject_label *curracl;
	char *pathname;
	__u32 retval;

	if (unlikely(!(gr_status & GR_READY)))
		return mode;

	pathname = d_real_path(new_dentry, mnt, gr_shared_page[current->processor], PAGE_SIZE);
	match = lookup_name_entry(pathname);

	if (!match)
		goto check_parent;

	curracl = current->acl;

	read_lock(&gr_inode_lock);
	matchpo = lookup_acl_obj_label(match->inode, match->device,
				       curracl->pos);
	read_unlock(&gr_inode_lock);

	if (matchpo) {
		if ((matchpo->mode & mode) != mode) {
			if(curracl->mode & GR_LEARN) {
				gr_log_learn(curracl, new_dentry, mnt, pathname, mode);

				return mode;
			}
			return 0;
		}
		return mode;
	}		

check_parent:
	curracl = current->acl;

	retval = chk_obj_label(parent, mnt, mode, curracl);

	if((retval != mode) && (curracl->mode & GR_LEARN)) {
		gr_log_learn(curracl, new_dentry, mnt, pathname, mode);
		return mode;
	}

	return retval;
}

int
gr_check_hidden_task(const struct task_struct *task)
{
	if (unlikely(!(gr_status & GR_READY)))
		return 0;

	if (!(task->acl->mode & GR_FIND) &&
	    !(current->acl->mode & GR_VIEW)) 
			return 1;

	return 0;
}

int
gr_check_protected_task(const struct task_struct *task)
{
	if (unlikely(!(gr_status & GR_READY) || !task))
		return 0;

	if ((task->acl->mode & GR_PROTECTED) &&
	    !(current->acl->mode & GR_KILL))
		return 1;

	return 0;
}

__inline__ void gr_copy_label(struct task_struct *tsk)
{
	/* We don't need to lock task...because task is being created
	 * in do_fork() so it can't be modified anywhere else as it has
	 * yet to be a true process. see kernel/fork.c */

	tsk->used_accept = 0;
	tsk->acl_admin = 0;
	tsk->acl = current->acl;
	tsk->curr_ip = current->curr_ip;
	if (unlikely(current->used_accept))
		current->curr_ip = 0;

	return;
}

static __inline__ void gr_set_proc_res(void)
{
	struct acl_subject_label * proc;
	unsigned short i;

	proc = current->acl;

	if (proc->mode & GR_LEARN)
		return;

	for (i = 0; i < RLIM_NLIMITS; i++) {
		if (!(proc->resmask & (1 << i)))
			continue;

		current->rlim[i].rlim_cur = proc->res[i].rlim_cur;
		current->rlim[i].rlim_max = proc->res[i].rlim_max;
	}

	return;
}

void gr_set_pax_flags(struct task_struct *task)
{
	struct acl_subject_label * proc;

	if (unlikely(!(gr_status & GR_READY)))
		return;

	proc = task->acl;

	if (proc->mode & GR_PAXPAGE)
		task->flags &= ~PF_PAX_PAGEEXEC;
	if (proc->mode & GR_PAXSEGM)
		task->flags &= ~PF_PAX_SEGMEXEC;
	if (proc->mode & GR_PAXGCC)
		task->flags |= PF_PAX_EMUTRAMP;
	if (proc->mode & GR_PAXMPROTECT)
		task->flags &= ~PF_PAX_MPROTECT;
	if (proc->mode & GR_PAXRANDMMAP)
		task->flags &= ~PF_PAX_RANDMMAP;
	if (proc->mode & GR_PAXRANDEXEC)
		task->flags |= PF_PAX_RANDEXEC;

	return;
}

void
gr_set_proc_label(const struct dentry *dentry, const struct vfsmount *mnt)
{
	struct acl_subject_label *newacl;

	if (unlikely(!(gr_status & GR_READY)))
		return;

	newacl = chk_subj_label(dentry, mnt);

	if ((newacl->mode & GR_LEARN) || !chk_obj_label(dentry, mnt, GR_INHERIT, current->acl))
		current->acl = newacl;

	gr_set_proc_res();

	return;
}

static __inline__ void
do_handle_delete(const ino_t ino, const kdev_t dev)
{
	struct acl_object_label *matchpo;
	struct acl_subject_label *matchps;
	unsigned long i;

	for (i = 0; i < acl_subj_set.s_used; i++) {
		if ((matchpo = lookup_acl_obj_label(ino, dev, i)) != NULL) {
#ifdef CONFIG_GRKERNSEC_ACL_DEBUG
			security_debug(GR_OBJDEL_DEBUG_MSG,
					kdevname(dev), ino, i);
#endif
			matchpo->mode |= GR_DELETED;
		}
	}

	if ((matchps = lookup_acl_subj_label(ino, dev)) != NULL) {
#ifdef CONFIG_GRKERNSEC_ACL_DEBUG
		security_debug(GR_SUBJDEL_DEBUG_MSG,
			       kdevname(dev), ino);
#endif
		matchps->mode |= GR_DELETED;
	}

	return;
}

void
gr_handle_delete(const ino_t ino, const kdev_t dev)
{
	if (unlikely(!(gr_status & GR_READY))) return;

	write_lock(&gr_inode_lock);
	if (lookup_inodev_entry(ino, dev))
		do_handle_delete(ino, dev);
	write_unlock(&gr_inode_lock);

	return;
}

static __inline__ void
update_acl_obj_label(const ino_t oldinode, const kdev_t olddevice, const ino_t newinode,
		      const kdev_t newdevice, const unsigned long parent_pos)
{
	unsigned long index = phash(oldinode, olddevice, parent_pos, acl_obj_set.o_size);
	struct acl_object_label **match;
	struct acl_object_label *tmp;
	__u8 i = 0;

	match = &acl_obj_set.o_hash[index];

	while (*match && ((*match)->inode != oldinode || (*match)->device != olddevice ||
	       (*match)->parent_pos != parent_pos)) {
		index = (index + (1 << i)) % acl_obj_set.o_size;
		match = &acl_obj_set.o_hash[index];
		i = (i + 1) % 32;
	}
	
	if (*match && ((*match) != deleted_object) && ((*match)->inode == oldinode) && 
	    ((*match)->device == olddevice) && ((*match)->parent_pos == parent_pos)) {
		tmp = *match;
		tmp->inode = newinode;
		tmp->device = newdevice;
		tmp->mode &= ~GR_DELETED;
		acl_obj_set.o_used--;
	
		*match = deleted_object;

		insert_acl_obj_label(tmp);
	}

	return;
}

static __inline__ void
update_acl_subj_label(const ino_t oldinode, const kdev_t olddevice, const ino_t newinode,
		      const kdev_t newdevice)
{
	unsigned long index = fhash(oldinode, olddevice, acl_subj_set.s_size);
	struct acl_subject_label **match;
	struct acl_subject_label *tmp;
	__u8 i = 0;

	match = &acl_subj_set.s_hash[index];

	while (*match && ((*match)->inode != oldinode || (*match)->device != olddevice)) {
		index = (index + (1 << i)) % acl_subj_set.s_size;
		i = (i + 1) % 32;
		match = &acl_subj_set.s_hash[index];
	}

	if (*match && (*match != deleted_subject) && ((*match)->inode == oldinode) && 
	    ((*match)->device == olddevice)) {
		tmp = *match;

		tmp->inode = newinode;
		tmp->device = newdevice;
		tmp->mode &= ~GR_DELETED;

		*match = deleted_subject;
		acl_subj_set.s_used--;
	
		insert_acl_subj_label(tmp);
	}

	return;
}

static __inline__ void
update_inodev_entry(const ino_t oldinode, const kdev_t olddevice, const ino_t newinode,
		      const kdev_t newdevice)
{
	unsigned long index = fhash(oldinode, olddevice, inodev_set.n_size);
	struct name_entry **match;
	struct name_entry *tmp;
	__u8 i = 0;

	match = &inodev_set.n_hash[index];

	while (*match && ((*match)->inode != oldinode || (*match)->device != olddevice)) {
		index = (index + (1 << i)) % inodev_set.n_size;
		i = (i + 1) % 32;
		match = &inodev_set.n_hash[index];
	}

	if (*match && (*match != deleted_inodev) && ((*match)->inode == oldinode) && 
	    ((*match)->device == olddevice)) {
		tmp = *match;

		tmp->inode = newinode;
		tmp->device = newdevice;

		*match = deleted_inodev;
	
		insert_inodev_entry(tmp);
	}

	return;
}

static __inline__ void
do_handle_create(const struct name_entry *matchn, const struct dentry * dentry,
		 const struct vfsmount * mnt)
{
	unsigned long i;

	update_acl_subj_label(matchn->inode, matchn->device,
			      dentry->d_inode->i_ino,
			      dentry->d_inode->i_dev);

	for (i = 0; i < acl_subj_set.s_used; i++) {
		update_acl_obj_label(matchn->inode, matchn->device,
				      dentry->d_inode->i_ino,
				      dentry->d_inode->i_dev, i);
	}

	update_inodev_entry(matchn->inode, matchn->device, dentry->d_inode->i_ino,
			    dentry->d_inode->i_dev);

	return;
}

void
gr_handle_create(const struct dentry * dentry, const struct vfsmount * mnt)
{
	struct name_entry *matchn;
	char *pathname;

	if (unlikely(!(gr_status & GR_READY))) return;

	pathname = d_real_path(dentry, mnt, gr_shared_page[current->processor], PAGE_SIZE);
	matchn = lookup_name_entry(pathname);

	if (matchn) {
		write_lock(&gr_inode_lock);
		do_handle_create(matchn, dentry, mnt);
		write_unlock(&gr_inode_lock);
	}

	return;
}

int
gr_handle_rename(struct inode * old_dir, struct inode * new_dir,
		 struct dentry * old_dentry, 
		 struct dentry * new_dentry,
		 struct vfsmount * mnt, const __u8 noreplace)
{
	struct name_entry *matchn;
	char *pathname;
	int error = 0;

	pathname = d_real_path(new_dentry, mnt, gr_shared_page[current->processor], PAGE_SIZE);
	matchn = lookup_name_entry(pathname);

	lock_kernel();
	error = vfs_rename(old_dir, old_dentry, new_dir, new_dentry);
	unlock_kernel();

	if(error)
		return error;

	write_lock(&gr_inode_lock);
	if (unlikely(!noreplace)) {
		if (lookup_inodev_entry(new_dentry->d_inode->i_ino,
				          new_dentry->d_inode->i_dev))
			do_handle_delete(new_dentry->d_inode->i_ino,
					 new_dentry->d_inode->i_dev);
	}

	if (lookup_inodev_entry(old_dentry->d_inode->i_ino,
			          old_dentry->d_inode->i_dev))
		do_handle_delete(old_dentry->d_inode->i_ino,
				 old_dentry->d_inode->i_dev);

	if (matchn)
		do_handle_create(matchn, old_dentry, mnt);
	write_unlock(&gr_inode_lock);

	return error;
}	

/* The following variables are needed for timer manipulation */
static struct timer_list gr_badpw;
static int failures = 0;
static int during_wait = 0;

static void
gr_timer(unsigned long ignored)
{
	failures = 0;
	during_wait = 0;
	del_timer(&gr_badpw);

	return;
}

int
gr_proc_handler(ctl_table * table, int write, struct file *filp, void *buffer,
		size_t * lenp)
{
	struct gr_arg *arg;
	struct gr_arg usermode;
	int error = sizeof (struct gr_arg);

	if (!write) return -EPERM;

	spin_lock(&gr_proc_lock);

	arg = (struct gr_arg *) buffer;

	if (*lenp != sizeof (struct gr_arg)) {
		security_alert_good(GR_PROC_ACL_MSG, GR_PROC_ACL_FLD,
			       (int) *lenp, (int) sizeof (struct gr_arg));
		error = -EINVAL;
		goto out;
	}
	if (during_wait) {
		error = -EPERM;
		goto out;
	}

	if (copy_from_user(&usermode, arg, sizeof (struct gr_arg))) {
		error = -EFAULT;
		goto out;
	}

	/* ensure pw is null terminated */

	usermode.pw[GR_PW_LEN - 1] = '\0';

	/* Okay. 
	 * We have our enough of the argument structure..(we have yet
	 * to copy_from_user the tables themselves) . Copy the tables
	 * only if we need them, i.e. for loading operations. */

	switch (usermode.mode) {
	case SHUTDOWN:
		if ((gr_status & GR_READY) && !(chkpw(&usermode,&pwent))) {
			gr_status &= ~GR_READY;
			security_alert_good(GR_SHUTS_ACL_MSG, GR_SHUTS_ACL_FLD,
				       DEFAULTSECARGS);
			grsecurity_disable();
			memset(&usermode, 0, sizeof(struct gr_arg));
			memset(&pwent, 0, sizeof(struct admin_pw)); 
		} else if(gr_status & GR_READY) {
			security_alert(GR_SHUTF_ACL_MSG, GR_SHUTF_ACL_FLD,
					DEFAULTSECARGS);
			error = -EPERM;
		} else
			security_alert_good(GR_SHUTI_ACL_MSG, GR_SHUTI_ACL_FLD,
					DEFAULTSECARGS);
		break;
	case ENABLE:
		if (gr_disable)
			security_alert_good(GR_ENABLEI_ACL_MSG, GR_ENABLEI_ACL_FLD,
				       GR_VERSION);
		else if (!(gr_status & GR_READY) && 
			!gracl_init(&usermode))
			security_alert_good(GR_ENABLE_ACL_MSG, GR_ENABLE_ACL_FLD,
					GR_VERSION);
		else
			security_alert(GR_ENABLEF_ACL_MSG, GR_ENABLEF_ACL_FLD,
					GR_VERSION, DEFAULTSECARGS);
		break;
	case RELOAD:
		if (gr_disable || !(gr_status & GR_READY))
			security_alert_good(GR_RELOADI_ACL_MSG, GR_RELOADI_ACL_FLD);
		else if (!(chkpw(&usermode, &pwent))) {
			gr_status &= ~GR_READY;
			grsecurity_disable();
			if(!gracl_init(&usermode))
				security_alert_good(GR_RELOAD_ACL_MSG, 
						GR_RELOAD_ACL_FLD, GR_VERSION);
			else
				security_alert(GR_RELOADF_ACL_MSG, 
						GR_RELOADF_ACL_FLD, GR_VERSION,
						DEFAULTSECARGS);
		} else {
			security_alert(GR_RELOADF_ACL_MSG, 
					GR_RELOADF_ACL_FLD, GR_VERSION,
					DEFAULTSECARGS);
			error = -EPERM;
		}
		break;
	case SEGVMOD:
		if (unlikely(!(gr_status & GR_READY))) {
			security_alert_good(GR_SEGVMODI_ACL_MSG, GR_SEGVMODI_ACL_FLD,
					DEFAULTSECARGS);
			break;
		}

		if (!(chkpw(&usermode, &pwent))) {
			security_alert_good(GR_SEGVMODS_ACL_MSG, GR_SEGVMODS_ACL_FLD,
					DEFAULTSECARGS);
			if(usermode.segv_device && usermode.segv_inode) {
				struct acl_subject_label * segvacl; 
				segvacl = lookup_acl_subj_label(usermode.segv_inode, usermode.segv_device);
				if(segvacl) {
					segvacl->crashes = 0;
					segvacl->expires = 0;
				}
			} else if (gr_find_uid(usermode.segv_uid) >= 0) {
				gr_remove_uid(usermode.segv_uid);
			}
		} else {
			security_alert(GR_SEGVMODF_ACL_MSG, GR_SEGVMODF_ACL_FLD,
					DEFAULTSECARGS);
			error = -EPERM;
		}
		break;
	case GOD:
		if (unlikely(!(gr_status & GR_READY))) {
			security_alert_good(GR_ADMINI_ACL_MSG, GR_ADMINI_ACL_FLD,
					DEFAULTSECARGS);
			break;
		}
		
		if (!(chkpw(&usermode, &pwent))) {
			security_alert_good(GR_ADMINS_ACL_MSG, GR_ADMINS_ACL_FLD,
				       DEFAULTSECARGS);
			read_lock(&tasklist_lock);
			if (current->p_pptr) {
				current->p_pptr->acl = (struct acl_subject_label *) god_label;
				current->p_pptr->acl_admin = 1;
			}
			read_unlock(&tasklist_lock);
		} else {
			security_alert(GR_ADMINF_ACL_MSG, GR_ADMINF_ACL_FLD,
					DEFAULTSECARGS);
			error = -EPERM;
		}

		break;
	default:
		security_alert(GR_INVMODE_ACL_MSG, GR_INVMODE_ACL_FLD,
			       usermode.mode, DEFAULTSECARGS);
		break;
	}

	if (error != -EPERM)
		goto out;

	failures++;

	if (failures > CONFIG_GRKERNSEC_ACL_MAXTRIES) {
		security_alert(GR_MAXPW_ACL_MSG, GR_MAXPW_ACL_FLD,
				CONFIG_GRKERNSEC_ACL_MAXTRIES);
		init_timer(&gr_badpw);
		gr_badpw.data = 0;
		gr_badpw.function = gr_timer;
		gr_badpw.expires = jiffies + CONFIG_GRKERNSEC_ACL_TIMEOUT * HZ;
		add_timer(&gr_badpw);
		during_wait = 1;
	}


out:
	spin_unlock(&gr_proc_lock);
	return error;
}

static __inline__ struct file * gr_find_proc_file(struct task_struct * task)
{
	struct mm_struct * mm;
	struct vm_area_struct * vma;
	struct file * result = NULL;

	task_lock(task);
	mm = task->mm;
	if (mm)
		atomic_inc(&mm->mm_users);
	task_unlock(task);
	if (!mm)
		goto out;
	down_read(&mm->mmap_sem);

	for(vma = mm->mmap; vma; vma = vma->vm_next) {
		if ((vma->vm_flags & VM_EXECUTABLE) && vma->vm_file) {
			result = vma->vm_file;
			break;
		}
	}
	up_read(&mm->mmap_sem);
	mmput(mm);
out:
	return result;
}

int gr_set_acls(const int type) {
	struct task_struct *task;
	struct file * filp;
	unsigned short i;

	read_lock(&tasklist_lock);
	for_each_task(task) {
		/* check to see if we're called from the exit handler,
		   if so, only replace ACLs that have inherited the admin
		   ACL */
		if (type && task->acl != god_label) continue;

		if ((filp = gr_find_proc_file(task))) {
			task->acl = chk_subj_label(filp->f_dentry, filp->f_vfsmnt);
			if (task->acl) {
				struct acl_subject_label *curr;
				curr = task->acl;
				
				if(!(curr->mode & GR_LEARN))
				for(i = 0; i < RLIM_NLIMITS; i++) {
					if(!(curr->resmask & (1 << i)))
						continue;

					task->rlim[i].rlim_cur = curr->res[i].rlim_cur;
					task->rlim[i].rlim_max = curr->res[i].rlim_max;
				}
			} else {
				read_unlock(&tasklist_lock);
				security_alert_good(GR_DEFACL_MSG, GR_DEFACL_FLD,
					task->comm, task->pid);
				return 1;
			}
		} else {
			// it's a kernel process
#ifdef CONFIG_GRKERNSEC_ACL_HIDEKERN
			task->acl = kernel_label;
#else
			read_lock(&child_reaper->fs->lock);
			task->acl = lookup_acl_subj_label(child_reaper->fs->root->d_inode->i_ino,
					child_reaper->fs->root->d_inode->i_dev);
			read_unlock(&child_reaper->fs->lock);
#endif
		}
	}
	read_unlock(&tasklist_lock);
	return 0;
}

void gr_learn_resource(const struct task_struct * task,
		       const int res, const unsigned long wanted)
{
	struct acl_subject_label *acl;

	if (unlikely((gr_status & GR_READY) && 
		     task->acl && (task->acl->mode & GR_LEARN)))
		goto skip_reslog;

#ifdef CONFIG_GRKERNSEC_RESLOG
	gr_log_resource(task, res, wanted);
#endif
skip_reslog:

	if(unlikely(!(gr_status & GR_READY) || !wanted)) return;

	acl = task->acl;
	
	if(likely(!acl || !(acl->mode & GR_LEARN) ||
		  !(acl->resmask & (1 << (unsigned short)res))))
		return;

	if(wanted >= acl->res[res].rlim_cur)
	{
		unsigned long res_add;

		res_add = wanted;
		switch(res) {
		case RLIMIT_CPU: res_add += GR_RLIM_CPU_BUMP; break;
		case RLIMIT_FSIZE: res_add += GR_RLIM_FSIZE_BUMP; break;
		case RLIMIT_DATA: res_add += GR_RLIM_DATA_BUMP; break;
		case RLIMIT_STACK: res_add += GR_RLIM_STACK_BUMP; break;
		case RLIMIT_CORE: res_add += GR_RLIM_CORE_BUMP; break;
		case RLIMIT_RSS: res_add += GR_RLIM_RSS_BUMP; break;
		case RLIMIT_NPROC: res_add += GR_RLIM_NPROC_BUMP; break;
		case RLIMIT_NOFILE: res_add += GR_RLIM_NOFILE_BUMP; break;
		case RLIMIT_MEMLOCK: res_add += GR_RLIM_MEMLOCK_BUMP; break;
		case RLIMIT_AS: res_add += GR_RLIM_AS_BUMP; break;
		case RLIMIT_LOCKS: res_add += GR_RLIM_LOCKS_BUMP; break;
		}

		acl->res[res].rlim_cur = res_add;
	
		if(wanted > acl->res[res].rlim_max)
			acl->res[res].rlim_max = res_add;

		security_learn(GR_LEARN_AUDIT_MSG, acl->device,
				acl->inode, acl->res[res].rlim_cur,
				acl->res[res].rlim_max, "", (unsigned long)res);
	}

	return;
}

#ifdef CONFIG_SYSCTL
extern struct proc_dir_entry * proc_sys_root;

__u32 gr_handle_sysctl(const struct ctl_table * table, __u32 mode)
{
	struct proc_dir_entry * tmp;
	struct nameidata nd;
	const char * proc_sys = "/proc/sys";
	char * path;	
	unsigned short len = 0, pos = 0, depth = 0, i;
	__u32 err = 0;

	if (unlikely(!(gr_status & GR_READY))) return 1;

	/* convert the requested sysctl entry into a pathname */

	for(tmp = table->de; tmp != proc_sys_root; tmp = tmp->parent) {
		len += strlen(tmp->name);
		len++;
		depth++;
	}	

	if((len + depth + strlen(proc_sys) + 1) > PATH_MAX)
		return 0; // deny

	path = kmalloc(PATH_MAX, GFP_KERNEL);

	if(!path) return 0;

	memset(path, 0, PATH_MAX);

	memcpy(path, proc_sys, strlen(proc_sys));

	pos += strlen(proc_sys);

	for(; depth > 0; depth--) {
		path[pos] = '/';
		pos++;
		for(i = 1, tmp = table->de; tmp != proc_sys_root; tmp = tmp->parent) {
			if(depth == i) {
				memcpy(path + pos, tmp->name, strlen(tmp->name));
				pos += strlen(tmp->name);
			}
			i++;
		}
	}

	if (path_init(path, LOOKUP_FOLLOW, &nd))
		err = path_walk(path, &nd);

	if(err) 
		goto out;
	
	err = chk_obj_label(nd.dentry, nd.mnt, mode, current->acl);

	if (unlikely((current->acl->mode & GR_LEARN) && (err != mode))) {
		err = mode;
		gr_log_learn(current->acl, nd.dentry, nd.mnt, path, mode);
	}

	path_release(&nd);

out:
	kfree(path);
	return err;
}
#endif

int gr_handle_ptrace(struct task_struct * task, const long request)
{
	struct file * filp;
	__u32 retmode;

	if(unlikely(!(gr_status & GR_READY))) return 0;

	filp = gr_find_proc_file(task);

	if (unlikely(!filp))
		return 0;

	retmode = gr_search_file(filp->f_dentry, GR_WRITE | GR_PTRACERD,
				 filp->f_vfsmnt);
	
	if (retmode & GR_PTRACERD) {
		switch (request) {
			case PTRACE_POKETEXT:
			case PTRACE_POKEDATA:
			case PTRACE_POKEUSR:
			case PTRACE_SETREGS:
			case PTRACE_SETFPREGS:
#ifdef CONFIG_X86
			case PTRACE_SETFPXREGS:
#endif
				return 1;
			default:
				return 0;
		}
	} else if(!(current->acl->mode & GR_OVERRIDE) &&
	   (permission(filp->f_dentry->d_inode, MAY_WRITE) ||
	   !(retmode & GR_WRITE))) {
		security_alert(GR_PTRACE_ACL_MSG, GR_PTRACE_ACL_FLD,	
				task->comm, task->pid, DEFAULTSECARGS);
		return 1;
	}

	return 0;
}

int gr_handle_ptrace_exec(const struct dentry *dentry, const struct vfsmount *mnt,
			  const char *filename)
{
	__u32 retmode;

	if(unlikely(!(gr_status & GR_READY))) return 0;

	if (unlikely((current->ptrace & PT_PTRACED) && !(current->acl->mode & GR_OVERRIDE)))
		retmode = gr_search_file(dentry, GR_WRITE | GR_PTRACERD, mnt);
	else
		return 0;

	if (!(retmode & GR_PTRACERD) && 
	    (permission(dentry->d_inode, MAY_WRITE) ||
	    !(retmode & GR_WRITE))) {
		security_alert(GR_PTRACE_EXEC_ACL_MSG, GR_PTRACE_EXEC_ACL_FLD,
			kdevname(dentry->d_inode->i_dev),
			dentry->d_inode->i_ino, filename, DEFAULTSECARGS);
		return 1;
	}

	return 0;
}

int gr_handle_mmap(const struct file * filp, const unsigned long prot)
{
	if(unlikely(!(gr_status & GR_READY))) return 0;

	if(!(current->acl->mode & GR_OVERRIDE) && filp && (prot & PROT_EXEC) &&
	   chk_obj_label(filp->f_dentry, filp->f_vfsmnt, GR_WRITE, root_label)) {
		security_alert(GR_WRITLIB_ACL_MSG, GR_WRITLIB_ACL_FLD,
				kdevname(filp->f_dentry->d_inode->i_dev),
				filp->f_dentry->d_inode->i_ino,
				DEFAULTSECARGS);
		return 1;
	}

	return 0;
}
